#include <fstream>
#include <iomanip>
#include <cmath>
#include <string>
#include <string.h>
#include <stdio.h>

#include "NavierStokes.h"
#include "Thompson2D.h"
#include "kinCond1D.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;

#include "field/field_base.h"

int main(void) {
  char buf[256];
  
  // Reading and renumbering of the mesh:
  
  Mesh MESH;
  MESH.read("square.msh","gmsh2dMesh");
  string sortType="position";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ_de_coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",
                 COMPONENT(0)=Double(1.0),COMPONENT(1)=Double(1.0e-4)));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd=MESH.getDomain(2);
    MESH=renumberDefDomainFrontal(MESH,Bnd);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }

  // Domains:

  Domain Dom,Bnd1,Bnd2,Bnd3;
  Dom=MESH.getDomainFromIntId(16);
  Dom.setName("main domain");
  Bnd1=MESH.getDomainFromIntId(17);
  Bnd1.setName("left boundary");
  Bnd2=MESH.getDomainFromIntId(18);
  Bnd2.setName("lower boundary");
  Bnd3=MESH.getDomainFromIntId(19);
  Bnd3.setName("upper boundary");
  
  Domain PointA,PointB,PointC;
  {
    Domain Corners1,Corners2;
    
    Corners1=Bnd3.getBoundary();
    Corners2=Bnd1.getBoundary();
    PointA=Corners1*Corners2;
    
    Corners1=Bnd2.getBoundary();
    Corners2=Bnd1.getBoundary();
    PointB=Corners1*Corners2;
    
    Corners1=Bnd3.getBoundary();
    Corners2=Bnd2.getBoundary();
    PointC=Corners1*Corners2;
  }
  
  // Coordinate systems:
  
  string coordInterp="quadratic";
  Field coord; 
  {
    Field tmpCoord=MESH.getCoordinateField(Dom);
    tmpCoord.setName("tmpCoord");
    coord=Field((DOM=Dom,INTERPOLATION=coordInterp,NBRCOMPONENTS=2,
                 COMPONENTNAME(0)="X",COMPONENTNAME(1)="Y",
                 NAME="coordinates")); 
    coord.interpolate(tmpCoord,Dom,tmpCoord);
  }
    
  // Other Fields:
  
  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=3,
		  COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		  COMPONENTNAME(2)="w",MINREF=1.0,
		  NAME="velocities"));
      
  // Remeshing:
  
  Thompson2D Prob1((param("type","2D2D - coordinates - "),
		    param("domain",Dom),
		    param("coord",coord),
		    param("integRule","order 5"),
		    param("name","2D Thompson equations")));
  Prob1.setLayer(-100);
  
  kinCond1D Prob2((param("type","3D1D - Coord V - "),
		   param("domain",Bnd2),
		   param("coord",coord),
		   param("velo",velocities),
		   param("axisym",true),
		   param("surfaceTension",2e-3),
		 param("coupledNormals",false),
                   param("fixBegin",true),
                   param("fixEnd",true),
		   param("integRule","order 5"),
		   param("name","FS on Bnd2") ));
  kinCond1D Prob3((param("type","3D1D - Coord V - "),
		   param("domain",Bnd3),
		   param("coord",coord),
		   param("velo",velocities),
		   param("axisym",true),
		   param("surfaceTension",4e-3),
		 param("coupledNormals",false),
                   param("fixBegin",true),
                   param("fixEnd",true),
		   param("integRule","order 5"),
		   param("name","FS on Bnd3") ));
		   
  // Various quantities from Prob1 and Prob3:
  
  Array X=Prob1.X;
  Array Y=Prob1.Y;
  Array testX=Prob1.testX;
  Array testY=Prob1.testY;
  Array refX=Prob1.refX;
  Array refY=Prob1.refY;
  
  Array ACoordXY=Prob1.coordinates;
  TensorCoordSys TCSXY=Prob1.tcs;
  
  GlobLoc GL=Prob3.gl;
  
  // Circle on Domain Bnd2:
  
  Array R((TYPE="array",ORDER=0,COMPONENT=0.1,NAME="R"));
  Array circleEquation=(X*X+Y*Y-R*R);
  Array Prob2bis=integrate(Prob2.testDisplN*circleEquation,Prob2.dom,Prob2.itg);
  Prob2bis.setLayer(100);
  Prob2bis.setName("Prob2bis (circle equation)");
  
  // Boundary conditions for Thompson remshing:
  
  Constr Prob1BC2((NAME="Thompson left boundary",TYPE="usual",DOM=Bnd1,
		   TEST=testX,EQUATION=X-refX,GLOBLOC=GL));
  
  // Navier-Stokes equations:
  
  Array volForce((TYPE="array",ORDER=1,DIM(0)=3,NAME="volumic force",
                  COMPONENT(0)=0.0,
		  COMPONENT(1)=-10000.0,
		  COMPONENT(2)=0.0));
		  
  NavierStokes Prob4((param("type","3D2D - V P - F GPn"),
		        param("domain",Dom),
		        param("coord",coord),
		        param("coordMobil",true),
		        param("velo",velocities),
		        param("pressure",pressure),
		        param("axisym",true),
		        param("integRule","order 5"),
		        param("name","2D Stokes equations"),
			param("force",volForce),
		        param("rho",10.0),
		        param("viscosity",1.0) ));
  
  // Boundary conditions for NS equations:
    
  Array Avelo=Prob4.velocities;
  Array Apressure=Prob4.pressure;
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Equations de mouvement"));  
  
  Array U((TYPE="arrayComp",ORDER=0,NAME="U",COMPONENT()=Avelo.component(0)));
  Array testU((TYPE="arrayComp",ORDER=0,NAME="testU",COMPONENT()=testMvt.component(0)));
  Array W((TYPE="arrayComp",ORDER=0,NAME="W",COMPONENT()=Avelo.component(2)));
  Array testW((TYPE="arrayComp",ORDER=0,NAME="testW",COMPONENT()=testMvt.component(2)));

  Constr Prob4BC1((NAME="U=0 on boundary 1",TYPE="usual",DOM=Bnd1,
		   TEST=testU,EQUATION=U,GLOBLOC=GL));
  Constr Prob4BC2((NAME="velo=0 on boundary 2",TYPE="usual",DOM=Bnd2,
		   TEST=testMvt,EQUATION=Avelo,GLOBLOC=GL));

  Array Omega((TYPE="array",ORDER=0,NAME="Omega",COMPONENT()=0.0));
  Constr Prob4BC3((NAME="W=w*X on boundary 2",TYPE="usual",DOM=Bnd2,LAYER=100,
		   TEST=testW,EQUATION=W-Omega*X,GLOBLOC=GL));
		   
  // Contact Point :
  
  GlobLoc GL3D0D((COORD=Prob2.coordinates,GLTYPE="3D0D"));
  GlobLoc GL0D((COORD=ACoordXY,GLTYPE="0D"));
  
  Array Avelo2D((TYPE="arrayComp",ORDER=1,DIM(0)=2,NAME="Avelo2D",
                 COMPONENT(0)=Avelo.component(0),
		 COMPONENT(1)=Avelo.component(1)));
  Array testMvt2D((TYPE="arrayComp",ORDER=1,DIM(0)=2,NAME="testMvt2D",
                   COMPONENT(0)=testMvt.component(0),
		   COMPONENT(1)=testMvt.component(1)));

  Field displTField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
		   COMPONENTNAME(0)="dT",MINREF=1.0,NAME="displTField"));
  Field veloTField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="vT",MINREF=1.0,NAME="veloTField"));
  Array AdT((TYPE="arrayField",ORDER=0,NAME="AdT",
  	     DERIV=true,TENSORCOORDSYS=Prob2.tcs,
  	     COMPONENT()=displTField.component(0)));
  Array AvT((TYPE="arrayField",ORDER=0,NAME="AvT",
  	     DERIV=true,TENSORCOORDSYS=Prob2.tcs,
  	     COMPONENT()=veloTField.component(0)));
  Array Tvect=Prob2.Tvect;

  Integ ITG0D((GLOBLOC=GL3D0D,TENSORCOORDSYS=Prob2.tcs,RULE=1));

  Array Eq5=Prob2.testDisplT*AvT;
  Array Prob5=integrate(Eq5,PointC,ITG0D);
  Prob5.setLayer(+100);
  Prob5.setName("Prob5: contact point");
  Constr Prob5BC1((NAME="T displ",TYPE="usual",DOM=PointC,
		   TEST=Prob2.testDisplT,EQUATION=AdT-Prob2.displT,
		   GLOBLOC=GL0D));
  Prob5BC1.setLayer(+100);
  Constr Prob5BC2((NAME="Vt Point C",TYPE="usual",DOM=PointC,
		   TEST=testMvt2D,EQUATION=Avelo2D-Tvect*AvT,
		   GLOBLOC=GL0D));
  Prob5BC2.setLayer(+100);
  
  // Volume equation :
  
  double bubbleVolume=2.0*3.1415926535*0.1*0.1*0.1/3.0;
  Integ INTG((GLOBLOC=Prob4.gl,TENSORCOORDSYS=Prob4.tcs,RULE=3));
  Field attachPos((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                   MINREF=1.0,
	           COMPONENTNAME(0)="attachPos",NAME="attachPos"));
  Array AattachPos((TYPE="arrayField",ORDER=0,NAME="AattachPos",DERIV=true,
                    TENSORCOORDSYS=Prob4.tcs,
		    COMPONENT()=attachPos.component(0)));
  Array AVolume((TYPE="array",ORDER=0,COMPONENT()=bubbleVolume,
		 NAME="imposed bubble volume"));
  Array testAttachPos((TYPE="arrayTest",TEST=AattachPos,
		       NAME="Test for attachment position equation"));  
  Array Prob6=integrate(testAttachPos,Dom,INTG)-testAttachPos*AVolume;
  Prob6.setName("equation for the attachment point");
  
  Array testDisplY=Prob1.testY;
  Array displY=Prob1.Y;
  Constr Prob1BC5((NAME="Free surface attachement BC",TYPE="usual",
                   DOM=PointA,LAYER=200,
		   TEST=testDisplY,EQUATION=displY-AattachPos,GLOBLOC=GL0D));

  // Resolution:
  
  Solver total((PROBLEM=Prob1,CONSTR=Prob1BC2,
 	        PROBLEM=Prob4,CONSTR=Prob4BC1,CONSTR=Prob4BC2,CONSTR=Prob4BC3,
	        PROBLEM=Prob2,PROBLEM=Prob2bis,
	        PROBLEM=Prob3,
	        PROBLEM=Prob5,CONSTR=Prob5BC1,CONSTR=Prob5BC2,
		PROBLEM=Prob6,
		CONSTR=Prob1BC5,
	        CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=8,ZERO=1.0e-12));
  
  Plot output((DOM=Dom,COORD=ACoordXY,RESULT=Avelo,RESULT=Apressure,
               RESULT=volForce));
  
  // Initial solution:
  
  ofstream f("data.out");
  f << setw(6) << "i"
    << setw(16) << "capillarity"
    << setw(16) << "wetFactor"
    << setw(16) << "omega"
    << endl ;
    
  int i;
  double capillarity,omega,wetFactor;
  string filename;
  
  i=0;
  wetFactor=0.0;
  capillarity=0.0;
  omega=0.0;
  {
    cerr  << "step " << i << ": " 
          << "NEW VALUE OF SURFACE TENSION: " << capillarity << endl ;
	  
    f << setw(6) << i
      << setw(16) << capillarity
      << setw(16) << wetFactor
      << setw(16) << omega
      << endl ;
      
    Prob2.surfaceTension.component()=capillarity*wetFactor;
    Prob3.surfaceTension.component()=capillarity;
    Omega.component()=omega;
    total.NewtonRaphson();
    Prob2.updateNormalDispl();
    Prob3.updateNormalDispl();
    if (total.getLastState()==1) {
      sprintf(buf,"%d",i);
      filename="cup_";
      filename+=buf;
      output.exec(filename);
    }
    else 
      exit(0);
    i++;
  }
  
  capillarity=2.5;
  {
    cerr  << "step " << i << ": " 
          << "NEW VALUE OF SURFACE TENSION: " << capillarity << endl ;
	  
    f << setw(6) << i
      << setw(16) << capillarity
      << setw(16) << wetFactor
      << setw(16) << omega
      << endl ;
      
    Prob2.surfaceTension.component()=capillarity*wetFactor;
    Prob3.surfaceTension.component()=capillarity;
    Omega.component()=omega;
    total.NewtonRaphson();
    Prob2.updateNormalDispl();
    Prob3.updateNormalDispl();
    if (total.getLastState()==1) {
      sprintf(buf,"%d",i);
      filename="cup_";
      filename+=buf;
      output.exec(filename);
    }
    else 
      exit(0);
    i++;
  }

  wetFactor=0.1;
  while (wetFactor<0.9) {
    wetFactor+=(1.0-wetFactor)/20.0;
    if (wetFactor>0.9)
      wetFactor=0.9;
//  wetFactor=-0.01;
//  double term=0.1;
//  while (-wetFactor<0.5) {
//    term=(1.0-abs(wetFactor))/10.0;
//    wetFactor-=term;
//    if (-wetFactor>0.5)
//      wetFactor=-0.5;
    
    cerr  << "step " << i << ": " 
          << "NEW VALUE OF WET FACTOR: " << wetFactor << endl ;
	  
    f << setw(6) << i
      << setw(16) << capillarity
      << setw(16) << wetFactor
      << setw(16) << omega
      << endl ;
      
    Prob2.surfaceTension.component()=capillarity*wetFactor;
    Prob3.surfaceTension.component()=capillarity;
    Omega.component()=omega;
    total.NewtonRaphson();
    Prob2.updateNormalDispl();
    Prob3.updateNormalDispl();
    if (total.getLastState()==1) {
      sprintf(buf,"%d",i);
      filename="cup_";
      filename+=buf;
      output.exec(filename);
    }
    else 
      exit(0);
    i++;
  }
  
  omega=0.1;
  while (omega<10.0) {
    omega+=0.1;
    if (omega>10.0)
      omega=10.0;
    
    cerr  << "step " << i << ": " 
          << "NEW VALUE OF OMEGA: " << omega << endl ;
	  
    f << setw(6) << i
      << setw(16) << capillarity
      << setw(16) << wetFactor
      << setw(16) << omega
      << endl ;
      
    Prob2.surfaceTension.component()=capillarity*wetFactor;
    Prob3.surfaceTension.component()=capillarity;
    Omega.component()=omega;
    total.NewtonRaphson();
    Prob2.updateNormalDispl();
    Prob3.updateNormalDispl();
    if (total.getLastState()==1) {
      sprintf(buf,"%d",i);
      filename="cup_";
      filename+=buf;
      output.exec(filename);
    }
    else 
      exit(0);
    i++;
  }

}
  
  
  
  
  
  
  
  
