#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "ReinerRivlin2D.h"
#include "BirdCarreau.h"
#include "streamFunction2D.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



int main(void) {
  
  // Reading and renumbering of the mesh:
  
  Mesh MESH;
  MESH.read("square.msh","rsMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ de coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
		 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd1=MESH.getDomain(1);
    MESH=renumberDefDomainFrontal(MESH,-Bnd1);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("square.msh2","rsMesh");
  
  // Domains:
  
  Domain Dom,Bnd,Bnd1,Bnd2,Bnd3,Bnd4,Corners1,Corners2,Corner;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("left boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("lower boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("right boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("upper boundary");
  Corners1=Bnd1.getBoundary();
  Corners2=Bnd4.getBoundary();
  Corner=Corners1*Corners2;
  Bnd=Bnd1+Bnd2+Bnd3+Bnd4;
  
  // Fields and Arrays:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=3,
		    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
                    COMPONENTNAME(2)="w",NAME="velocities"));
  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="p",NAME="pressure"));
  Field gradpn((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
	        COMPONENTNAME(0)="gradpn",NAME="gradpn"));
  Array AFlowRate((TYPE="array",ORDER=0,COMPONENT()=100.0,
		   NAME="imposed flow rate"));
  
  // Navier-Stokes equations:
  
  BirdCarreau fViscosity(10.0,5.7939,0.8985);
  BirdCarreau fRRViscosity(-8.53,3.9212,-0.8252);
  
  ReinerRivlin2D Prob1((param("type","3D2D - V P GPn - F Q"),
		        param("domain",Dom),
		        param("coord",coord),
		        param("axisym",false),
		        param("velo",velocities),
		        param("pressure",pressure),
		        param("integRule","order 5"),
		        param("name","2D ReinerRivlin equations"),
		        param("viscosity",fViscosity),
		        param("RRviscosity",fRRViscosity),
		        param("rho",0.0),
		        param("gradPn",gradpn),
		        param("flowRate",AFlowRate) ));

  Array ACoord=Prob1.coordinates;
  TensorCoordSys TCS=Prob1.tcs;
  GlobLoc GL=Prob1.gl;
  
  GlobLoc GL0((COORD=ACoord,GLTYPE="0D"));
  GlobLoc GL2((COORD=ACoord,GLTYPE="3D1D"));
  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  Integ INTG2((GLOBLOC=GL2,TENSORCOORDSYS=TCS,RULE=3));  
  
  GlobLoc::setDefaultGL(GL);

  // Needed quantities:
  
  Array Avelo=Prob1.velocities;
  Array Apressure=Prob1.pressure;
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,
	         NAME="test for momentum equations")); 
  Array testDiv((TYPE="arrayTest",TEST=Apressure,
		 NAME="test for continuity equations"));
    
  // RR boundary conditions:
  
  Array AveloZero((TYPE="array",ORDER=1,DIM=3,NAME="vitessesZero",
		   TENSORCOORDSYS=TCS));
  Constr Prob1BC1((TYPE="usual",NAME="V=0 on boundary",DOM=Bnd,
		   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  Array ApressureZero((TYPE="array",ORDER=0,NAME="pressureZero",
		       TENSORCOORDSYS=TCS));
  Constr Prob1BC2((TYPE="usual",NAME="P=0 on a corner",DOM=Corner,
		   TEST=testDiv,EQUATION=Apressure-ApressureZero,GLOBLOC=GL0,
		   TENSORCOORDSYS=TCS));
  
  // Equations for stream function:
  
  Field psi((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	     COMPONENTNAME(0)="psi",NAME="psi"));
  streamFunction2D Prob2((param("type","3D2D - psi - V"),
		          param("psi",psi),
		          param("coord",coord),
		          param("velo",velocities),
		          param("domain",Dom),
		          param("integRule","order 5"),
		          param("name","stream function equations") ));
  
  Array Astream=Prob2.psi;
  Array testStream=test(Astream);
  Field streamBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
		  NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
		  NAME="value of stream function for BC"));
  Array AstreamBC((TYPE="arrayField",ORDER=0,NAME="streamBC",
		   TENSORCOORDSYS=TCS,COMPONENT()=streamBC.component(0)));
    
  Array APsiZero((TYPE="array",ORDER=0,NAME="psi = 0",TENSORCOORDSYS=TCS));
  Constr Prob2BC1((NAME="Psi=0 on boundary",TYPE="usual",DOM=Bnd,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  
  // Resolution:
  
  Solver NS((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,
             CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=8));
  Solver STREAM((PROBLEM=Prob2,CONSTR=Prob2BC1,CONVERGENCE=1.0e-4,
		 DIVERGENCE=1.0e4,NITMAX=1));
  
  Plot output((DOM=Dom,COORD=ACoord,RESULT=Avelo,RESULT=Apressure,
	       RESULT=Astream));
  
  NS.NewtonRaphson();
  STREAM.NewtonRaphson();

  string filename="square";
  output.exec(filename);
}


