#include <cmath>
#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "FENE.h"
#include "streamFunction2D.h"
#include "Thompson2D.h"
#include "kinCond1D.h"
#include "kinCondXY.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Filter.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



int main(void) {
  char buf[256];
  Double Eplate=1.6;
  Double L0=0.002;
  
  // Reading and renumbering of the mesh:
  
  Mesh MESH;
  MESH.read("stretch.mesh","emc2Mesh");
  string sortType="frontal";
  if (sortType=="position") {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ_de_coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",
		 COMPONENT(0)=Double(1.0e-4),COMPONENT(1)=Double(1.0)));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd=MESH.getDomain(2);
    MESH=renumberDefDomainFrontal(MESH,Bnd);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  
  // Domains:
  
  Domain Dom,Bnd1,Bnd2,Bnd3,Bnd4;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("left boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("lower boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("right boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("upper boundary");
  
  // Coordinate systems:
  
  Field coord; 
  {
    Field tmpCoord=MESH.getCoordinateField(Dom);
    coord=Field((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
                 COMPONENTNAME(0)="X",COMPONENTNAME(1)="Y",
                 NAME="coordinates")); 
    coord.interpolate(tmpCoord,Dom,tmpCoord);
  }
  
  // Other Fields:
  
  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
		    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		    NAME="velocities"));
  Field configuration((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=4,
		       COMPONENTNAME(0)="Arr",COMPONENTNAME(1)="Arz",
		       COMPONENTNAME(2)="Azz",COMPONENTNAME(3)="Att",
		       NAME="configuration"));
  Field psi((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	     COMPONENTNAME(0)="psi",NAME="psi"));
  
  // Initialisation of configuration Field: 
  
  {
    Array AConfig((TYPE="array",ORDER=1,DIM=4,COMPONENT(0)=1.0,COMPONENT(1)=0.0,
		   COMPONENT(2)=1.0,COMPONENT(3)=1.0));
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ_de_coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    configuration.interpolate(AConfig,Dom,ACoord);
  }
      
  // TimeDep:
  
  TimeDep TD((NAME="transient",SCHEME="Galerkin",
	      MAXSTEPSNBR=7000,DTMIN=0.000001,DTMAX=1.0,DTINIT=0.001,TINIT=0.0,
	      EPSILON=0.001,TFINAL=5.0,
	      INFLU=coord,INFLU=velocities,INFLU=configuration,ACTIVE=pressure));
  
  // Remeshing:
  
  Thompson2D Prob1((param("type","2D2D - coordinates - "),
		    param("domain",Dom),
		    param("coord",coord),
		    param("integRule","order 5"),
		    param("name","2D Thompson equations")));
  Prob1.setLayer(-100);
  
  kinCond1D Prob3((param("type","3D1D - Coord V - "),
		   param("domain",Bnd3),
		   param("coord",coord),
		   param("velo",velocities),
		   param("axisym",true),
		   param("fixEnd",true),
		   param("surfaceTension",28.9e-3),
		   param("integRule","order 5"),
		   param("timeDep",TD),
		   param("name","All for free surface") ));
  GlobLoc GL=Prob3.gl;
    
  // Various quantities from Prob1 and Prob3:
  
  Array X=Prob1.X;
  Array Y=Prob1.Y;
  Array testX=Prob1.testX;
  Array testY=Prob1.testY;
  Array refX=Prob1.refX;
  Array refY=Prob1.refY;
  
  Array ACoordXY=Prob1.coordinates;
  TensorCoordSys TCSXY=Prob1.tcs;
  
  // Boundary conditions for Thompson remshing:
  
  Array Yplate((TYPE="array",ORDER=0,COMPONENT()=L0/2.0,NAME="Yplate",
	        TENSORCOORDSYS=TCSXY));
	     
  Constr Prob1BC1((NAME="Thompson left boundary",TYPE="usual",DOM=Bnd1,
		   TEST=testX,EQUATION=X-refX,GLOBLOC=GL,LAYER=100));
  Constr Prob1BC2((NAME="Thompson lower boundary",TYPE="usual",DOM=Bnd2,
		   TEST=testY,EQUATION=Y-refY,GLOBLOC=GL,LAYER=100));
  Constr Prob1BC3((NAME="Thompson upper boundary Y",TYPE="usual",DOM=Bnd4,
		   TEST=testY,EQUATION=Y-Yplate,GLOBLOC=GL,LAYER=100));
  Constr Prob1BC4((NAME="Thompson upper boundary X",TYPE="usual",DOM=Bnd4,
		   TEST=testX,EQUATION=X-refX,GLOBLOC=GL,LAYER=100));
		   
  // All FENE-P equations:
  
  FENE Prob2((param("type","3D2D - A V P - F GPn"),
	      param("domain",Dom),
	      param("timeDep",TD),
	      param("coord",coord),
	      param("coordMobil",true),
	      param("config",configuration),
	      param("velo",velocities),
	      param("pressure",pressure),
	      param("axisym",true),
	      param("integRule","order 5"),
	      param("name","2D FENE-P equations"),
	      param("rho",0.0),
	      param("newtVisc",35.7),
	      param("relaxTime",2.0),
	      param("lSquare",4325.0),
	      param("elastVisc",69.3) ));
  
  // Boundary conditions for FENE-P equations:
  
  Array Aconfig=Prob2.configuration;
  Array Avelo=Prob2.velocities;
  Array Avelo2D((TYPE="arrayField",ORDER=1,DIM=2,NAME="velocities",
	         COMPONENT(0)=velocities.component(0),
	         COMPONENT(1)=velocities.component(1)));
  Array Apressure=Prob2.pressure;
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Equations de mouvement"));  
  
  Array U((TYPE="arrayComp",ORDER=0,NAME="u",COMPONENT()=Avelo.component(0)));
  Array V((TYPE="arrayComp",ORDER=0,NAME="v",COMPONENT()=Avelo.component(1)));
  Array testU((TYPE="arrayComp",ORDER=0,NAME="u",COMPONENT()=testMvt.component(0)));
  Array testV((TYPE="arrayComp",ORDER=0,NAME="v",COMPONENT()=testMvt.component(1)));
  
  Array Vplate((TYPE="array",ORDER=0,COMPONENT()=L0*Eplate/2.0,NAME="Vplate",
	        TENSORCOORDSYS=TCSXY));
		
  Constr Prob2BC1((NAME="U=0 on boundary 1",TYPE="usual",DOM=Bnd1,
		   TEST=testU,EQUATION=U,GLOBLOC=GL,LAYER=100));
  Constr Prob2BC2((NAME="V=0 on boundary 2",TYPE="usual",DOM=Bnd2,
		   TEST=testV,EQUATION=V,GLOBLOC=GL,LAYER=100));
  Constr Prob2BC3((NAME="U=0 on boundary 4",TYPE="usual",DOM=Bnd4,
		   TEST=testU,EQUATION=U,GLOBLOC=GL,LAYER=100));
  Constr Prob2BC4((NAME="V=Vplate on boundary 4",TYPE="usual",DOM=Bnd4,
		   TEST=testV,EQUATION=V-Vplate,GLOBLOC=GL,LAYER=100));
  
  // Equations for stream function:
  
  streamFunction2D Prob4((param("type","3D2D - psi - V"),
			  param("axisym",true),
		          param("psi",psi),
		          param("coord",coord),
		          param("velo",velocities),
		          param("domain",Dom),
		          param("integRule","order 5"),
		          param("name","stream function equations") ));

  // Boundary conditions for stream function:
  
  Array Astream=Prob4.psi;
  Array testStream=test(Astream);
  Field streamBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
		  NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
		  NAME="value of stream function for BC"));
  Array AstreamBC((TYPE="arrayField",ORDER=0,NAME="streamBC",
		   TENSORCOORDSYS=TCSXY,COMPONENT()=streamBC.component(0)));
  
  Array APsiZero((TYPE="array",ORDER=0,NAME="psi = 0",TENSORCOORDSYS=TCSXY));
  Constr Prob4BC1((NAME="Psi=0 on boundary 1",TYPE="usual",DOM=Bnd1,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL,
		   TENSORCOORDSYS=TCSXY));
  Constr Prob4BC2((NAME="Psi=0 on boundary 2",TYPE="usual",DOM=Bnd2,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL,
		   TENSORCOORDSYS=TCSXY));
  
  // Resolution:
  
  Solver totalTransi((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,CONSTR=Prob1BC3,
                      CONSTR=Prob1BC4,
		      PROBLEM=Prob2,CONSTR=Prob2BC1,CONSTR=Prob2BC2,CONSTR=Prob2BC3,
		      CONSTR=Prob2BC4,
		      PROBLEM=Prob3,
		      CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  totalTransi.setZero(1.0e-50);
  
  Solver STREAM((PROBLEM=Prob4,CONSTR=Prob4BC1,CONSTR=Prob4BC2,
		 CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  
  Plot output((DOM=Dom,COORD=ACoordXY,RESULT=Avelo2D,RESULT=Apressure,
	       RESULT=Astream,RESULT=Aconfig,
	       RESULT=Prob2.coordinatesTimeDeriv,
	       RESULT=Prob2.CauchyStressTensor,
	       RESULT=Prob2.extraStressTensor));
  Plot outputDX((DOM=Dom,COORD=ACoordXY,RESULT=Avelo2D,RESULT=Apressure,
	         RESULT=Astream,RESULT=Aconfig,
	         RESULT=Prob2.coordinatesTimeDeriv,
	         RESULT=Prob2.CauchyStressTensor,
	         RESULT=Prob2.extraStressTensor,
		 TYPE="openDX"));
  
  // Time-dependent solution:
  
  {
    TD.addInflu(totalTransi);  
    while(TD.isToBeContinued()) {
      Yplate.component()=L0/2*exp(Eplate*TD.getCurrentTime());
      Vplate.component()=L0/2*Eplate*exp(Eplate*TD.getCurrentTime());
      totalTransi.NewtonRaphson();
      if (TD.getCurrentStepIndex()==1||(TD.getCurrentStepIndex()%5==0)) {
        STREAM.NewtonRaphson();
	sprintf(buf,"%d",TD.getCurrentStepIndex());
	string filename="stretch_";
	filename+=buf;
	output.exec(filename);
	outputDX.exec(filename);
      }
    }
  }
  
}
