#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "NavierStokes.h"
#include "Thompson2D.h"
#include "kinCondXY.h"
#include "streamFunction2D.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Filter.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



int main(void) {
  char buf[256];
  
  // Reading and renumbering of the mesh:
  
  Mesh MESH;
  MESH.read("threadR.mesh","emc2Mesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ_de_coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",
                 COMPONENT(0)=Double(1.0),COMPONENT(1)=Double(1.0e-4)));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd=MESH.getDomain(2);
    MESH=renumberDefDomainFrontal(MESH,Bnd);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }

  // Domains:

  Domain Dom,Bnd1,Bnd2,Bnd3,Bnd4;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("lower boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("right boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("upper boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("left boundary");
  
  // Coordinate systems:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  
  // Other Fields:
  
  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
		    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		    NAME="velocities"));
  Field psi((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	     COMPONENTNAME(0)="psi",NAME="psi"));
      
  // TimeDep:
  
  TimeDep TD((NAME="transient",SCHEME="Galerkin",MAXSTEPSNBR=700,
	      DTMIN=0.000001,DTMAX=0.1,DTINIT=0.01,TINIT=0.0,EPSILON=0.001,
	      TFINAL=2000.0,INFLU=velocities,ACTIVE=pressure));

  // Remeshing:
  
  Thompson2D Prob1((param("type","2D2D - coordinates - "),
		    param("domain",Dom),
		    param("coord",coord),
		    param("integRule","order 5"),
		    param("filter",Bnd3),
		    param("name","2D Thompson equations")));
  
  kinCondXY Prob3((param("type","3D1D - Coord V - "),
		   param("domain",Bnd3),
		   param("timeDep",TD),
		   param("coord",coord),
		   param("velo",velocities),
		   param("axisym",true),
		   param("surfaceTension",1.0),
		   param("integRule","order 5"),
		   param("name","All for free surface") ));
  
  // Various quantities from Prob1 and Prob3:
  
  Array X=Prob1.X;
  Array Y=Prob1.Y;
  Array testX=Prob1.testX;
  Array testY=Prob1.testY;
  Array refX=Prob1.refX;
  Array refY=Prob1.refY;
  
  Array ACoordXY=Prob1.coordinates;
  TensorCoordSys TCSXY=Prob1.tcs;
  
  GlobLoc GL=Prob3.gl;
  
  // Boundary conditions for Thompson remshing:
  
  Array One((TYPE="array",ORDER=0,COMPONENT()=1.0,NAME="1.0",
	     TENSORCOORDSYS=TCSXY));
  Array Ampli((TYPE="array",ORDER=0,NAME="amplitude",TENSORCOORDSYS=TCSXY));
  
  Constr Prob1BC1((NAME="Thompson lower boundary",TYPE="usual",DOM=Bnd1,
		   TEST=testX,EQUATION=X-refX,GLOBLOC=GL));
  Constr Prob1BC2((NAME="Thompson right boundary",TYPE="usual",DOM=Bnd2,
		   TEST=testY,EQUATION=Y-refY,GLOBLOC=GL));
  Constr Prob1BC3((NAME="Thompson upper boundary",TYPE="usual",DOM=Bnd3,
		   TEST=testX,
		   EQUATION=X-refX*(One+Ampli*(cos(refY*(3.1415926535/5.0)))),
		   GLOBLOC=GL));
  Constr Prob1BC4((NAME="Thompson left boundary",TYPE="usual",DOM=Bnd4,
		   TEST=testY,EQUATION=Y-refY,GLOBLOC=GL));
  
  // Navier-Stokes equations:
  
  NavierStokes Prob2((param("type","3D2D - V P - F GPn"),
		        param("domain",Dom),
		        param("timeDep",TD),
		        param("coord",coord),
		        param("coordMobil",true),
		        param("velo",velocities),
		        param("pressure",pressure),
		        param("axisym",true),
		        param("integRule","order 5"),
		        param("name","2D Stokes equations"),
		        param("rho",0.0),
		        param("viscosity",0.5) ));
  
  // Boundary conditions for NS equations:
    
  Array Avelo=Prob2.velocities;
  Array Avelo2D((TYPE="arrayField",ORDER=1,DIM=2,NAME="velocities",
	         COMPONENT(0)=velocities.component(0),
	         COMPONENT(1)=velocities.component(1)));
  Array Apressure=Prob2.pressure;
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Equations de mouvement"));  
  
  Array U((TYPE="arrayComp",ORDER=0,NAME="u",COMPONENT()=Avelo.component(0)));
  Array V((TYPE="arrayComp",ORDER=0,NAME="v",COMPONENT()=Avelo.component(1)));
  Array testU((TYPE="arrayComp",ORDER=0,NAME="u",COMPONENT()=testMvt.component(0)));
  Array testV((TYPE="arrayComp",ORDER=0,NAME="v",COMPONENT()=testMvt.component(1)));

  Constr Prob2BC1((NAME="U=0 on boundary 1",TYPE="usual",DOM=Bnd1,
		   TEST=testU,EQUATION=U,GLOBLOC=GL));
  Constr Prob2BC2((NAME="V=0 on boundary 2",TYPE="usual",DOM=Bnd2,
		   TEST=testV,EQUATION=V,GLOBLOC=GL));
  Constr Prob2BC3((NAME="V=0 on boundary 4",TYPE="usual",DOM=Bnd4,
		   TEST=testV,EQUATION=V,GLOBLOC=GL));
  
  // Equations for stream function:
  
  streamFunction2D Prob4((param("type","3D2D - psi - V"),
			  param("axisym",true),
		          param("psi",psi),
		          param("coord",coord),
		          param("velo",velocities),
		          param("domain",Dom),
		          param("integRule","order 5"),
		          param("name","stream function equations") ));
  
  // Boundary conditions for stream function:
  
  Array Astream=Prob4.psi;
  Array testStream=test(Astream);
  Field streamBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
		  NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
		  NAME="value of stream function for BC"));
  Array AstreamBC((TYPE="arrayField",ORDER=0,NAME="streamBC",
		   TENSORCOORDSYS=TCSXY,COMPONENT()=streamBC.component(0)));
  
  Array APsiZero((TYPE="array",ORDER=0,NAME="psi = 0",TENSORCOORDSYS=TCSXY));
  Constr Prob4BC1((NAME="Psi=0 on boundary 1",TYPE="usual",DOM=Bnd1,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL,
		   TENSORCOORDSYS=TCSXY));
  Constr Prob4BC2((NAME="Psi=0 on boundary 2",TYPE="usual",DOM=Bnd2,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL,
		   TENSORCOORDSYS=TCSXY));
  Constr Prob4BC3((NAME="Psi=0 on boundary 4",TYPE="usual",DOM=Bnd4,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL,
		   TENSORCOORDSYS=TCSXY));
  
  // Resolution:
  
  Solver totalInit((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,CONSTR=Prob1BC3,
		    CONSTR=Prob1BC4,
		    PROBLEM=Prob2,CONSTR=Prob2BC1,CONSTR=Prob2BC2,CONSTR=Prob2BC3,
		    PROBLEM=Prob3,
		    CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6));
  Solver totalTransi((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,CONSTR=Prob1BC4,
		      PROBLEM=Prob2,CONSTR=Prob2BC1,CONSTR=Prob2BC2,CONSTR=Prob2BC3,
		      PROBLEM=Prob3,
		      CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  Solver STREAM((PROBLEM=Prob4,CONSTR=Prob4BC1,CONSTR=Prob4BC2,CONSTR=Prob4BC3,
		 CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  
  Plot output((DOM=Dom,COORD=ACoordXY,RESULT=Avelo2D,RESULT=Apressure,
	       RESULT=Astream));
  
  // Initial solution:
  
  Ampli.component()=0.05;
  totalInit.NewtonRaphson();
  STREAM.NewtonRaphson();
  
  output.exec("thread");
  
  // Time-dependent solution:
  
  {
    TD.addInflu(totalTransi);  
    while(TD.isToBeContinued()) {
      totalTransi.NewtonRaphson();
      if (TD.getCurrentStepIndex()==1
	  ||Int(TD.getCurrentTime()/0.5)!=Int(TD.getPreviousTime()/0.5)
	  ||TD.getCurrentTime()>13.0) {
        STREAM.NewtonRaphson();
	sprintf(buf,"%d",TD.getCurrentStepIndex());
	string filename="thread_";
	filename+=buf;
	output.exec(filename);
      }
    }
  }

}
  
  
  
  
  
  
  
  
