
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "ReinerRivlin2D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



ReinerRivlin2D::ReinerRivlin2D(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1,eqDim=-1;
    bool coordFieldNeeded=false,veloFieldNeeded=false,
	 pressureFieldNeeded=false,gradPnFieldNeeded=false;
    bool forceArrayExpected=false,forceArrayFound=false,
         gradPnArrayExpected=false,gradPnArrayFound=false,
	 flowRateArrayExpected=false,flowRateArrayFound=false,
         gravityArrayExpected=false,gravityArrayFound=false;
    Array testFlowRate;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName=="2D2D"||globLocTypeName=="3D2D") {
      globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
      locDim=GlobLoc::getLocDimFromName(globLocTypeName);
      if (locDim!=2)
	errThrow("ReinerRivlin2D::ReinerRivlin2D",
		 "invalide GlobLoc type name: "+globLocTypeName+".");	
    }
    else
      errThrow("ReinerRivlin2D::ReinerRivlin2D",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    
    // Time Dependence:
    
    bool boolTransient=false;
    try {
      if (PL.getNbrParamsNamed("timeDep")>0) {
	localTimeDep=PL.extractTimeDepData("timeDep");
        boolTransient=true;
      }
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of TimeDep recovery.");
    
    // Field type names:
    
    if (fieldTypeNames=="V P") {
      veloFieldNeeded=true;
      pressureFieldNeeded=true;
    }
    else if (fieldTypeNames=="V P GPn") {
      veloFieldNeeded=true;
      pressureFieldNeeded=true;
      gradPnFieldNeeded=true;
    }
    else
      errThrow("ReinerRivlin2D::ReinerRivlin2D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames=="GPn")
      gradPnArrayExpected=true;
    else if (arrayTypeNames=="Q")
      flowRateArrayExpected=true;
    else if (arrayTypeNames=="F Q") {
      forceArrayExpected=true;
      gravityArrayExpected=true;
      flowRateArrayExpected=true;
    }
    else if (arrayTypeNames=="F GPn") {
      forceArrayExpected=true;
      gravityArrayExpected=true;
      gradPnArrayExpected=true;
    }
    
    if (flowRateArrayExpected&&(!gradPnFieldNeeded))
      errThrow("ReinerRivlin2D::ReinerRivlin2D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Boolean parameters:
    
    bool boolCoordMobil=false;
    bool boolAxisym=false;
    try {    
      if (PL.getNbrParamsNamed("coordMobil")>0)
	boolCoordMobil=PL.extractBoolData("coordMobil");
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end bools recovery.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of strings recovery.");
    
    // Material parameters:
    
    Double dRho=0.0;
    
    Double dViscosity=0.0;
    bool boolFViscosity=false;
    function1 *fViscosity=NULL;
    
    Double dRR=0.0;
    bool boolFRR=false;
    function1 *fRR=NULL;
    
    try {    
      param P;
      
      P=PL.extractParamNamed("rho",true);
      if (P.getTypeName()=="Double")
	dRho=P.getDoubleData();
      else
	errThrow("ReinerRivlin2D::ReinerRivlin2D","invalid type name for rho: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("viscosity",true);
      if (P.getTypeName()=="function1") {
	boolFViscosity=true;
	fViscosity=(P.getFunction1Data());
      }
      else if (P.getTypeName()=="Double") {
	dViscosity=P.getDoubleData();
      }
      else
	errThrow("ReinerRivlin2D::ReinerRivlin2D","invalid type name for viscosity: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("RRviscosity",true);
      if (P.getTypeName()=="function1") {
	boolFRR=true;
	fRR=(P.getFunction1Data());
      }
      else if (P.getTypeName()=="Double") {
	dRR=P.getDoubleData();
      }
      else
	errThrow("ReinerRivlin2D::ReinerRivlin2D","invalid type name for RRviscosity: "
		 +P.getTypeName()+".");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of material parameters recovery.");
    
    // Domain:
    
    Domain dom;
    try {  
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("ReinerRivlin2D::ReinerRivlin2D","invalid domain dimension.");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D","at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,veloField,pressureField,tempField,gradPnField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (veloFieldNeeded)
	veloField=PL.extractFieldData("velo",true);
      if (pressureFieldNeeded)
	pressureField=PL.extractFieldData("pressure",true);
      if (gradPnFieldNeeded)
	gradPnField=PL.extractFieldData("gradPn",true);
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of Fields' recovery.");
    
    // Array's parameters:
    
    Array gravity;
    try {    
      if (forceArrayExpected) {
	force=PL.extractArrayData("force");
	if (force.getData()!=NULL) {
	  forceArrayFound=true;
	  if (force.getOrder()!=1)
	    errThrow("ReinerRivlin2D::ReinerRivlin2D","force order must be 1");
	  if (force.getDim(0)!=veloField.getNbrComponents())
	    errThrow("ReinerRivlin2D::ReinerRivlin2D",
		     "number of components of force must match velocity Field.");
	}
      }
      
      if (gravityArrayExpected) {
	gravity=PL.extractArrayData("gravity");
	if (gravity.getData()!=NULL) {
	  gravityArrayFound=true;
	  if (gravity.getOrder()!=1)
	    errThrow("ReinerRivlin2D::ReinerRivlin2D","gravity order must be 1");
	  if (gravity.getDim(0)!=veloField.getNbrComponents())
	    errThrow("ReinerRivlin2D::ReinerRivlin2D",
		     "number of components of gravity must match velocity Field.");
	}
      }
      
      if (forceArrayFound&&gravityArrayFound)
	errThrow("ReinerRivlin2D::ReinerRivlin2D",
		 "force and gravity together incompatible.");
      
      if (gradPnArrayExpected) {
	gradPn=PL.extractArrayData("gradPn");
	if (gradPn.getData()!=NULL) {
	  gradPnArrayFound=true;
	  if (gradPn.getOrder()!=0)
	    errThrow("ReinerRivlin2D::ReinerRivlin2D",
		     "order of gradPn Array must be 0.");
	}
      }
      
      if (flowRateArrayExpected) {
	flowRate=PL.extractArrayData("flowRate");
	if (flowRate.getData()!=NULL) {
	  flowRateArrayFound=true;
	  if (flowRate.getOrder()!=0)
	    errThrow("ReinerRivlin2D::ReinerRivlin2D","flowRate order must be 0.");
	}
      }
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of Arrays' recovery.");
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("ReinerRivlin2D::ReinerRivlin2D","");
    }
    
    // Intermediate data construction:
    
    if (gradPnFieldNeeded) {
      gradPn=Array((TYPE="arrayField",ORDER=0,DERIV=true,
		    NAME="normal gradient of pressure",
		    COMPONENT()=gradPnField.component(0)));
      gradPnArrayFound=true;
    }
    
    // Coordinates and associated entities:
    
    Int veloFieldNbrC=veloField.getNbrComponents();
    if (veloFieldNbrC<2||veloFieldNbrC>3)
      errThrow("ReinerRivlin2D::ReinerRivlin2D",
	       "veloField has an invalid number of components.");
    
    if (boolAxisym||veloFieldNbrC==3)
      eqDim=3;
    else
      eqDim=2;
    
    if (eqDim!=globDim)
      errThrow("ReinerRivlin2D::ReinerRivlin2D",
	       "invalid equation dimensions for "+globLocTypeName);
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("ReinerRivlin2D::ReinerRivlin2D",
		 "coordinate Field must have 2 components.");
      
      coordinates=Array((TYPE="arrayField",ORDER=1,DIM=eqDim,
			 NAME="Coordinates",DERIV=boolCoordMobil,
			 COMPONENT(0)=coordField.component(0),
			 COMPONENT(1)=coordField.component(1)));
      
      if (boolTransient&&boolCoordMobil) {
	coordinatesTimeDeriv=timeDeriv(coordinates,localTimeDep);
	coordinatesTimeDeriv.setName("time derivative of Coordinates");
      }
      
      if (eqDim==3)
	gl=GlobLoc((COORD=coordinates,GLTYPE="3D2D"));
      else
	gl=GlobLoc((COORD=coordinates,GLTYPE="2D2D"));
      
      if (boolAxisym)
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
      else if (eqDim==2)
	tcs=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=coordinates));
      else
	tcs=TensorCoordSys((TCSTYPE="Cartesian3D",COORD=coordinates));
      
      GlobLoc::setDefaultGL(gl);
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of coordinates and associated entities.");
    
    // Quantities related to velocities:
    
    Array testMvt;
    try {
      if (veloFieldNeeded) {
	if (eqDim==2)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=2,NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1)));
	else {
	  if (veloField.getNbrComponents()==3)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			      NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1),
			      COMPONENT(2)=veloField.component(2)));
	  else if (veloField.getNbrComponents()==2)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			      NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1)));
	}
	testMvt=Array((TYPE="arrayTest",TEST=velocities,
		       NAME="Momentum equations"));
      }
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of quantities related to velocities.");
    
    // Quantities related to pressure:
    
    Array testDiv;
    try {
      if (pressureFieldNeeded) {
	pressure=Array((TYPE="arrayField",ORDER=0,NAME="Pressure",
			DERIV=true,TENSORCOORDSYS=tcs,
			COMPONENT()=pressureField.component(0)));
	testDiv=Array((TYPE="arrayTest",TEST=pressure,
		       NAME="Incompressibility equation"));
      }
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of quantities related to pressure.");

    // Normal vector to the domain :
    
    Array normal;
    if (gl.getTypeName()!="2D2D") {
      normal=gl.getNormal(tcs);
      normal.setName("normal vector");
    }
    
    // Momentum equations:
    
    Array momentumExpr;
    
    try {
      rho=Array((TYPE="array",ORDER=0,COMPONENT()=dRho,
		 TENSORCOORDSYS=tcs,NAME="density"));
      
      gradVeloTensor=grad(velocities);
      gradVeloTensor.setName("gradient of velocities");
      strainRateTensor=0.5*(gradVeloTensor+transpose(gradVeloTensor));
      strainRateTensor.setName("strain rate tensor");
      strainRate=sqrt(2.0*accumulation(strainRateTensor,strainRateTensor));
      strainRate.setName("strain rate");
      
      if (boolFViscosity) {
	viscosity=Array(strainRate,*fViscosity);
	viscosity.setName("viscosity");
      }
      else 
        viscosity=Array((TYPE="array",ORDER=0,COMPONENT()=dViscosity,
			 TENSORCOORDSYS=tcs,NAME="viscosity"));
      
      if (boolFRR) {
	RRviscosity=Array(strainRate,*fRR);
	RRviscosity.setName("RRviscosity");
      }
      else 
	RRviscosity=Array((TYPE="array",ORDER=0,COMPONENT()=dRR,
			   TENSORCOORDSYS=tcs,NAME="RRviscosity"));
      
      Array unit;
      if (eqDim==2)
        unit=Array((TYPE="array",ORDER=2,DIM=2,
		    COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,
		    NAME="unit tensor",TENSORCOORDSYS=tcs));
      else if (eqDim==3)
        unit=Array((TYPE="array",ORDER=2,DIM=3,
		    COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,COMPONENT(2,2)=1.0,
		    NAME="unit tensor",TENSORCOORDSYS=tcs));

      CauchyStressTensor=2.0*viscosity*strainRateTensor-pressure*unit
	+4.0*RRviscosity
	  *contractedMultiplication(strainRateTensor,1,strainRateTensor,0);
      CauchyStressTensor.setName("Cauchy stress tensor");
      
      if (boolTransient) {
	Array velocitiesTimeDeriv=timeDeriv(velocities,localTimeDep);
	velocitiesTimeDeriv.setName("time derivative of Velocities");
	if (boolCoordMobil)
          inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication((velocities-coordinatesTimeDeriv),0,gradVeloTensor,1));
	else
          inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication(velocities,0,gradVeloTensor,1));
      }
      else
        inertia=rho*contractedMultiplication(velocities,0,gradVeloTensor,1);
      inertia.setName("inertia tensor");
      
      Array testMomentum((TYPE="arrayTest",TEST=velocities,
			  NAME="test for momentum equations"));  
      Array gradTestMomentum=grad(testMomentum);
      gradTestMomentum.setName("gradient of test for momentum equations.");
      
      if (gravityArrayFound) {
	force=rho*gravity;
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia-force+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia-force);
      }
      else if (forceArrayFound) {
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia-force+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia-force);
      }
      else {
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia);
      }
      momentumExpr.setName("momentum expression");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of momentum equations.");
    
    // Incompressibility equation:
    
    Array incompressibilityExpr;
    try {
      Array divV=contract(gradVeloTensor);
      divV.setName("divergence of velocities");
      Array testIncompressibility((TYPE="arrayTest",TEST=pressure,
				   NAME="test for incompressibility equation"));
      incompressibilityExpr=testIncompressibility*divV;
      incompressibilityExpr.setName("incompressibility expression");
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of incompressibility equation.");
    
    // Flow rate equation:
    
    Array flowRateExpr;
    
    try {
      if (flowRateArrayFound) {
	Array veloN=accumulation(velocities,normal);
	veloN.setName("normal component of velocities");
	testFlowRate=Array((TYPE="arrayTest",TEST=gradPn,
			    NAME="test for flow rate equations"));
	flowRateExpr=testFlowRate*veloN;
	flowRateExpr.setName("flowRate expression");
      }
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D",
	     "at the end of flow rate equation.");
        
    // Integration of all the equations:
    
    try {
      Array totalExpr=momentumExpr;
      if (incompressibilityExpr.getData()!=NULL)
	totalExpr=totalExpr+incompressibilityExpr;
      if (flowRateExpr.getData()!=NULL)
	totalExpr=totalExpr+flowRateExpr;
      totalExpr.setName("addition of integrable expressions");
      
      if (flowRateExpr.getData()!=NULL)
	Array::operator=(integrate(totalExpr,dom,itg)-testFlowRate*flowRate);
      else
	Array::operator=(integrate(totalExpr,dom,itg));
      
      setName(objectName);
    }
    catchAll("ReinerRivlin2D::ReinerRivlin2D","at the end of integration.");
  }
  catchAll("ReinerRivlin2D::ReinerRivlin2D","at the very end.");
}
