
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "Thompson2D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



Thompson2D::Thompson2D(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName=="2D2D") {
      globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
      locDim=GlobLoc::getLocDimFromName(globLocTypeName);
      if (locDim!=2)
	errThrow("Thompson2D::Thompson2D",
		 "invalide GlobLoc type name: "+globLocTypeName+".");	
    }
    else
      errThrow("Thompson2D::Thompson2D",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    
    // Field type names:
    
    if (fieldTypeNames!="coordinates")
      errThrow("Thompson2D::Thompson2D",
	       "unknown problem type: "+problemTypeName+".");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("Thompson2D::Thompson2D",
	     "at the end of strings recovery.");
        
    // Domain:
    
    Domain dom;
    try {  
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("Thompson2D::Thompson2D","invalid domain dimension.");
    }
    catchAll("Thompson2D::Thompson2D","at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField;
    try {      
      coordField=PL.extractFieldData("coord",true);
      if (coordField.getNbrComponents()!=2)
        errThrow("Thompson2D::Thompson2D",
		 "invalid number of components for coordinates field.");
    }
    catchAll("Thompson2D::Thompson2D",
	     "at the end of Fields' recovery.");
    
    // Coordinates and associated entities:
    
    try {
      coordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,DERIV=true,
			 NAME="Coordinates",
			 COMPONENT(0)=coordField.component(0),
			 COMPONENT(1)=coordField.component(1)));     
      testCoordinates=Array((TYPE="arrayTest",TEST=coordinates,
                             NAME="testCoordinates"));  
      
      gl=GlobLoc((COORD=coordinates,GLTYPE=globLocTypeName));
      tcs=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=coordinates));
      
      GlobLoc::setDefaultGL(gl);
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
      
      refCoordField.makeCopy(coordField);
      refCoordField.setName("Reference coordinate Field");
      refCoordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,DERIV=false,
			    NAME="Reference coordinates",
			    COMPONENT(0)=refCoordField.component(0),
			    COMPONENT(1)=refCoordField.component(1)));
    }
    catchAll("Thompson2D::Thompson2D",
	     "at the end of coordinates entities.");
    
    // Thompson equations:
    
    try {
      X=Array((TYPE="arrayField",ORDER=0,NAME="X",DERIV=true,
	       COMPONENT()=coordField.component(0),TENSORCOORDSYS=tcs));
      Y=Array((TYPE="arrayField",ORDER=0,NAME="Y",DERIV=true,
	       COMPONENT()=coordField.component(1),TENSORCOORDSYS=tcs));
      testX=Array((TYPE="arrayTest",TEST=X,NAME="testX"));  
      testY=Array((TYPE="arrayTest",TEST=Y,NAME="testY"));  
      refX=Array((TYPE="arrayField",ORDER=0,NAME="refX",DERIV=false,
		  COMPONENT()=refCoordField.component(0),TENSORCOORDSYS=tcs));
      refY=Array((TYPE="arrayField",ORDER=0,NAME="refY",DERIV=false,
		  COMPONENT()=refCoordField.component(1),TENSORCOORDSYS=tcs));
      
      Array thompson=accumulation(grad(testX,gl),grad(refX,gl))
	+accumulation(grad(testY,gl),grad(refY,gl));
      unfilteredEquations=integrate(thompson,dom,itg);  
      
      // Filters and filtered equations:
      
      paramList PFL=PL.getListOfType("filter");
      PL=PL.getListNotOfType("filter");
      nbrFilters=PFL.getNbrParams();
      if (nbrFilters>0) {
        theFilters=new Filter[nbrFilters];
        paramList FL;
        for (int i=0;i<PFL.getNbrParams();i++) {
      	  theFilters[i]=Filter((NAME="Filter for Thompson2D equations",
      	  			DOM=PFL[i].getDomainData(),
      	  			ARRAY=coordinates));
      	  FL.pushBack(FILTER=theFilters[i]);
        }
        Array::operator=(Array((TYPE="arrayFilter",
      	  			EXPR=unfilteredEquations,FL))); 
      } 
      else
        Array::operator=(unfilteredEquations);
			      
      if (objectName=="")
        setName("2D Thompson equations "+dom.getName());
      else
        setName(objectName);  
	
cerr << getName() << endl ;  
    }
    catchAll("Thompson2D::Thompson2D",
	     "at the end of Thompson equations.");
    
    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("Thompson2D::Thompson2D","");
    }
    
  }
  catchAll("Thompson2D::Thompson2D","at the very end.");
}



Thompson2D::~Thompson2D(void) {
  try{
    nbrFilters=0;
    theFilters=NULL;
  }
  catchAll("Thompson2D::~Thompson2D","at the very end.");
}
    
