
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "FENE.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



FENE::FENE(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1,veloDim=-1;
    bool coordFieldNeeded=false,veloFieldNeeded=false,
	 pressureFieldNeeded=false,gradPnFieldNeeded=false,
	 configFieldNeeded=false;
    bool forceArrayExpected=false,forceArrayFound=false,
	 gradPnArrayExpected=false,gradPnArrayFound=false,
	 flowRateArrayExpected=false,flowRateArrayFound=false,
											gravityArrayExpected=false,gravityArrayFound=false;
    Array testFlowRate;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // Boolean parameters:
    
    bool boolCoordMobil=false;
    bool boolAxisym=false;
    bool boolFENECR=false;
    bool boolUpwindA=false;
    try {    
      if (PL.getNbrParamsNamed("coordMobil")>0)
	boolCoordMobil=PL.extractBoolData("coordMobil");
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
      if (PL.getNbrParamsNamed("FENE-CR")>0)
	boolFENECR=PL.extractBoolData("FENE-CR");
      if (PL.getNbrParamsNamed("upwindA")>0)
	boolUpwindA=PL.extractBoolData("upwindA");
    }
    catchAll("FENE::FENE",
	     "at the end bools recovery.");
    
    // GlobLoc type name:
    
    if (globLocTypeName=="3D1D"||globLocTypeName=="3D2D") {
      globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
      locDim=GlobLoc::getLocDimFromName(globLocTypeName);
    }
    else
      errThrow("FENE::FENE",
	       "invalid GlobLoc type name: "+globLocTypeName+".");
    
    if ((!boolFENECR)&&(globDim!=3))
      errThrow("FENE::FENE",
	       "FENE-P model requires 3D definition of equations.");
    
    // Time Dependence:
    
    bool boolTransient=false;
    try {
      if (PL.getNbrParamsNamed("timeDep")>0) {
	localTimeDep=PL.extractTimeDepData("timeDep");
        boolTransient=true;
      }
    }
    catchAll("FENE::FENE",
	     "at the end of TimeDep recovery.");
    
    // Field type names:
    
    if (fieldTypeNames=="A V P") {
      configFieldNeeded=true;
      veloFieldNeeded=true;
      pressureFieldNeeded=true;
    }
    else if (fieldTypeNames=="A V P GPn") {
      configFieldNeeded=true;
      veloFieldNeeded=true;
      pressureFieldNeeded=true;
      gradPnFieldNeeded=true;
    }
    else
      errThrow("FENE::FENE",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames=="GPn")
      gradPnArrayExpected=true;
    else if (arrayTypeNames=="Q")
      flowRateArrayExpected=true;
    else if (arrayTypeNames=="F Q") {
      forceArrayExpected=true;
      gravityArrayExpected=true;
      flowRateArrayExpected=true;
    }
    else if (arrayTypeNames=="F GPn") {
      forceArrayExpected=true;
      gravityArrayExpected=true;
      gradPnArrayExpected=true;
    }
    else
      errThrow("FENE::FENE",
	       "unknown problem type: "+problemTypeName+".");
    
    if (flowRateArrayExpected&&(!gradPnFieldNeeded))
      errThrow("FENE::FENE",
	       "unknown problem type: "+problemTypeName+".");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("FENE::FENE",
	     "at the end of strings recovery.");
    
    // Material parameters:
    
    Double dRho=0.0;
    
    Double dNewtVisc=0.0;
    bool boolFNewtVisc=false;
    function1 *fNewtVisc=NULL;
    
    Double dElastVisc=0.0;
    Double dLSquare=0.0;
    Double dRelaxTime=0.0;
    
    try {    
      param P;
      
      P=PL.extractParamNamed("rho",true);
      if (P.getTypeName()=="Double")
	dRho=P.getDoubleData();
      else
	errThrow("FENE::FENE","invalid type name for rho: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("newtVisc",true);
      if (P.getTypeName()=="function1") {
	boolFNewtVisc=true;
	fNewtVisc=(P.getFunction1Data());
      }
      else if (P.getTypeName()=="Double") {
	dNewtVisc=P.getDoubleData();
      }
      else
	errThrow("FENE::FENE","invalid type name for newtonian viscosity: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("elastVisc",true);
      if (P.getTypeName()=="Double")
	dElastVisc=P.getDoubleData();
      else
	errThrow("FENE::FENE","invalid type name for elastVisc: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("lSquare",true);
      if (P.getTypeName()=="Double")
	dLSquare=P.getDoubleData();
      else
	errThrow("FENE::FENE","invalid type name for lSquare: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("relaxTime",true);
      if (P.getTypeName()=="Double")
	dRelaxTime=P.getDoubleData();
      else
	errThrow("FENE::FENE","invalid type name for relaxTime: "
		 +P.getTypeName()+".");
    }
    catchAll("FENE::FENE",
	     "at the end of material parameters recovery.");
    
    // Domain:
    
    Domain dom;
    try {    
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("FENE::FENE",
	         "invalid domain dimension.");
    }
    catchAll("FENE::FENE",
	     "at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,configField,veloField,pressureField,tempField,gradPnField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (configFieldNeeded)
	configField=PL.extractFieldData("config",true);
      if (veloFieldNeeded)
	veloField=PL.extractFieldData("velo",true);
      if (pressureFieldNeeded)
	pressureField=PL.extractFieldData("pressure",true);
      if (gradPnFieldNeeded)
	gradPnField=PL.extractFieldData("gradPn",true);
    }
    catchAll("FENE::FENE",
	     "at the end of Fields' recovery.");
    
    Int veloFieldNbrC=veloField.getNbrComponents();
    if (veloFieldNbrC<2||veloFieldNbrC>3)
      errThrow("FENE::FENE",
	       "veloField has an invalid number of components.");
    
    Int configFieldNbrC=configField.getNbrComponents();
    if ((veloFieldNbrC==2&&configFieldNbrC!=4)||(veloFieldNbrC==3&&configFieldNbrC!=6))
      errThrow("FENE::FENE",
	       "incompatibility of velocity and configuration number of components.");
    veloDim=veloFieldNbrC;
    
    // Array's parameters:
    
    Array gravity;
    try {    
      if (forceArrayExpected) {
	force=PL.extractArrayData("force");
	if (force.getData()!=NULL) {
	  forceArrayFound=true;
	  if (force.getOrder()!=1)
	    errThrow("FENE::FENE","force order must be 1.");
	  if (force.getDim(0)!=veloField.getNbrComponents())
	    errThrow("FENE::FENE",
		     "number of components of force must match velocity Field.");
	}
      }
      
      if (gravityArrayExpected) {
	gravity=PL.extractArrayData("gravity");
	if (gravity.getData()!=NULL) {
	  gravityArrayFound=true;
	  if (gravity.getOrder()!=1)
	    errThrow("FENE::FENE","gravity order must be 1.");
	  if (gravity.getDim(0)!=veloField.getNbrComponents())
	    errThrow("FENE::FENE",
		     "number of components of gravity must match velocity Field.");
	}
      }
      
      if (forceArrayFound&&gravityArrayFound)
	errThrow("FENE::FENE",
		 "force and gravity parameters incompatible.");
      
      if (gradPnArrayExpected) {
	gradPn=PL.extractArrayData("gradPn");
	if (gradPn.getData()!=NULL) {
	  gradPnArrayFound=true;
	  if (gradPn.getOrder()!=0)
	    errThrow("FENE::FENE",
		     "order of gradPn Array must be 0.");
	}
      }
      
      if (flowRateArrayExpected) {
	flowRate=PL.extractArrayData("flowRate");
	if (flowRate.getData()!=NULL) {
	  flowRateArrayFound=true;
	  if (flowRate.getOrder()!=0)
	    errThrow("FENE::FENE","flowRate order must be 0.");
	}
      }
    }
    catchAll("FENE::FENE","at the end of Arrays' recovery.");
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("FENE::FENE","");
    }
    
    // Intermediate data construction:
    
    if (gradPnFieldNeeded) {
      gradPn=Array((TYPE="arrayField",ORDER=0,DERIV=true,
		    NAME="normal gradient of pressure",
		    COMPONENT()=gradPnField.component(0)));
      gradPnArrayFound=true;
    }
    
    // Coordinates and associated entities:
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("FENE::FENE",
		 "coordinate Field must have 2 components.");
      
      if (coordField.getNbrComponents()==1)
        coordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,
			   NAME="Coordinates",DERIV=boolCoordMobil,
			   COMPONENT(0)=coordField.component(0)));
      else if (coordField.getNbrComponents()==2)
        coordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,
			   NAME="Coordinates",DERIV=boolCoordMobil,
			   COMPONENT(0)=coordField.component(0),
			   COMPONENT(1)=coordField.component(1)));
      else
	errThrow("FENE::FENE",
		 "coordinate Field must have 1 or 2 components.");
      
      if (boolTransient&&boolCoordMobil) {
	coordinatesTimeDeriv=timeDeriv(coordinates,localTimeDep);
	coordinatesTimeDeriv.setName("coordinatesTimeDeriv");
      }
      
      gl=GlobLoc((COORD=coordinates,GLTYPE=globLocTypeName));
      
      if (boolAxisym) {
        if (globDim!=3)
	  errThrow("FENE::FENE",
		   "globDim must be 3 if axisymmetric.");
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
      }
      else
	tcs=TensorCoordSys((TCSTYPE="Cartesian3D",COORD=coordinates));
      
      GlobLoc::setDefaultGL(gl);
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
    }
    catchAll("FENE::FENE",
	     "at the end of coordinates and associated entities.");
    
    Array unit;
    unit=Array((TYPE="array",ORDER=2,DIM=3,
		COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,COMPONENT(2,2)=1.0,
		NAME="unit tensor",TENSORCOORDSYS=tcs));
    
    // Quantities related to velocities:
    
    Array testMvt;
    try {
      if (veloFieldNeeded) {
	if (veloField.getNbrComponents()==3)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			    NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1),
			    COMPONENT(2)=veloField.component(2)));
	else if (veloField.getNbrComponents()==2)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			    NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1)));
	testMvt=Array((TYPE="arrayTest",TEST=velocities,
		       NAME="Momentum equations"));
      }
    }
    catchAll("FENE::FENE",
	     "at the end of quantities related to velocities.");
    
    // Quantities related to pressure:
    
    Array testDiv;
    try {
      if (pressureFieldNeeded) {
	pressure=Array((TYPE="arrayField",ORDER=0,NAME="Pressure",
			DERIV=true,TENSORCOORDSYS=tcs,
			COMPONENT()=pressureField.component(0)));
	testDiv=Array((TYPE="arrayTest",TEST=pressure,
		       NAME="Incompressibility equation"));
      }
    }
    catchAll("FENE::FENE",
	     "at the end of quantities related to pressure.");
	     
    // Quantities related to configuration tensor:
    
    Array testConfig,testConfigUpwind;
    try {
      if (configFieldNeeded) {
        if (veloDim==3) {
	  if (configField.getNbrComponents()!=6)
	    errThrow("FENE::FENE",
		     "if veloDim==3 nbr of components of Field A must be 6.");
	}
	else if ((!boolFENECR)&&veloDim==2) {
	  if (configField.getNbrComponents()!=4)
	    errThrow("FENE::FENE",
		     "if veloDim==2 nbr of components of A must be 4 for FENE-P.");
	}
	
	if (configField.getNbrComponents()==6)
	  configuration=Array((TYPE="arrayField",ORDER=2,DIM=globDim,
			       NAME="Configuration",
			       DERIV=true,TENSORCOORDSYS=tcs,
			       COMPONENT(0,0)=configField.component(0),
			       COMPONENT(0,1)=configField.component(1),
			       COMPONENT(0,2)=configField.component(2),
			       COMPONENT(1,0)=configField.component(1),
			       COMPONENT(1,1)=configField.component(3),
			       COMPONENT(1,2)=configField.component(4),
			       COMPONENT(2,0)=configField.component(2),
			       COMPONENT(2,1)=configField.component(4),
			       COMPONENT(2,2)=configField.component(5)));
	else if (configField.getNbrComponents()==4)
	  configuration=Array((TYPE="arrayField",ORDER=2,DIM=3,
			       NAME="Configuration",
			       DERIV=true,TENSORCOORDSYS=tcs,
			       COMPONENT(0,0)=configField.component(0),
			       COMPONENT(0,1)=configField.component(1),
			       COMPONENT(1,0)=configField.component(1),
			       COMPONENT(1,1)=configField.component(2),
			       COMPONENT(2,2)=configField.component(3)));
	else if (configField.getNbrComponents()==3)
	  configuration=Array((TYPE="arrayField",ORDER=2,DIM=3,
			       NAME="Configuration",
			       DERIV=true,TENSORCOORDSYS=tcs,
			       COMPONENT(0,0)=configField.component(0),
			       COMPONENT(0,1)=configField.component(1),
			       COMPONENT(1,0)=configField.component(1),
			       COMPONENT(1,1)=configField.component(2),
			       COMPONENT(2,2)=1.0));
			       
	testConfig=Array((TYPE="arrayTest",TEST=configuration,
			  NAME="Configuration equations"));
	
	if (boolUpwindA) {
	  double dInterp;
	  if (configField.getInterpolationTypeName()=="none"
	      ||configField.getInterpolationTypeName()=="constantOnDomain"
	      ||configField.getInterpolationTypeName()=="constantPerElement"
	      ||configField.getInterpolationTypeName()=="linear"
	      ||configField.getInterpolationTypeName()=="linearPerElement")
	    dInterp=1.0;
	  else
	    dInterp=2.0;
	      
          if (boolTransient&&boolCoordMobil)
	    testConfigUpwind=upwindF(velocities,1.0,0.0,testConfig,
	  			     coordinatesTimeDeriv,gl,dInterp);
	  else
	    testConfigUpwind=upwindF(velocities,1.0,0.0,testConfig,
	  			     0.0,gl,dInterp);
	}
      }
    }
    catchAll("FENE::FENE",
	     "at the end of quantities related to velocities.");
    
    // Vector normal to domain :
    
    Array normal;
    if (gl.getTypeName()!="2D2D") {
      normal=gl.getNormal(tcs);
      normal.setName("normal vector");
    }
    
    // Configuration equation:
    
    Array configurationExpr;
    Array invLSquare;
    try {
      relaxTime=Array((TYPE="array",ORDER=0,COMPONENT()=dRelaxTime,
		       TENSORCOORDSYS=tcs,NAME="relaxation time"));
      lSquare=Array((TYPE="array",ORDER=0,COMPONENT()=dLSquare,
		     TENSORCOORDSYS=tcs,NAME="L"));
      elastVisc=Array((TYPE="array",ORDER=0,COMPONENT()=dElastVisc,
		       TENSORCOORDSYS=tcs,NAME="viscoelastic viscosity"));
      
      invLSquare=inverse(lSquare);
      invLSquare.setName("1/L");
      
      gradVeloTensor=grad(velocities);
      gradVeloTensor.setName("gradient of velocities");
      
      Array configUCD,configUCD1,configUCD2;
      if (boolTransient) {
	configUCD1=timeDeriv(configuration,localTimeDep)
	  -contractedMultiplication(gradVeloTensor,1,configuration,0)
            -contractedMultiplication(configuration,1,transpose(gradVeloTensor),0);
	if (boolCoordMobil) {	
	  configUCD=timeDeriv(configuration,localTimeDep)
	    +contractedMultiplication(velocities-coordinatesTimeDeriv,0,grad(configuration),2)
	      -contractedMultiplication(gradVeloTensor,1,configuration,0)
		-contractedMultiplication(configuration,1,transpose(gradVeloTensor),0);
	  configUCD2=contractedMultiplication(velocities-coordinatesTimeDeriv,0,grad(configuration),2);
	}
	else {
	  configUCD=timeDeriv(configuration,localTimeDep)
	    +contractedMultiplication(velocities,0,grad(configuration),2)
	      -contractedMultiplication(gradVeloTensor,1,configuration,0)
		-contractedMultiplication(configuration,1,transpose(gradVeloTensor),0);
	  configUCD2=contractedMultiplication(velocities,0,grad(configuration),2);
        }
      }
      else {
	configUCD=contractedMultiplication(velocities,0,grad(configuration),2)
	  -contractedMultiplication(gradVeloTensor,1,configuration,0)
	    -contractedMultiplication(configuration,1,transpose(gradVeloTensor),0);
	configUCD1=-contractedMultiplication(gradVeloTensor,1,configuration,0)
          -contractedMultiplication(configuration,1,transpose(gradVeloTensor),0);
	configUCD2=contractedMultiplication(velocities,0,grad(configuration),2);
      }
      configUCD.setName("configUCD");
      configUCD1.setName("configUCD1");
      configUCD2.setName("configUCD2");
      
      Array tmpFactor=1.0-contract(configuration,0,1)*invLSquare;
      tmpFactor.setName("1-(tr A)/L");
      Array tmpFactor2=inverse(1.0-3.0*invLSquare);
      tmpFactor2.setName("1/(1-3/L)");
      Array tmpFactor3=inverse(tmpFactor);
      tmpFactor3.setName("1/(1-(tr A)/L)");
      
      if (boolUpwindA) {
        Array Aconstit1;
        Array Aconstit2;
        if (boolFENECR)
          Aconstit1=configuration+relaxTime*tmpFactor*configUCD1-unit;
        else
          Aconstit1=configuration+relaxTime*tmpFactor*configUCD1
      	    -(tmpFactor*tmpFactor2)*unit;
	Aconstit2=relaxTime*tmpFactor*configUCD2;
      	    
        Array configTest((TYPE="arrayTest",TEST=configuration,
      	  		  NAME="Test for config equations"));
        configurationExpr=accumulation(configTest,Aconstit1)
	  +accumulation(testConfigUpwind,Aconstit2);
        configurationExpr.setName("Expression for configuration equations");
      }
      else {
        Array Aconstit;
        if (boolFENECR)
          Aconstit=configuration+relaxTime*tmpFactor*configUCD-unit;
        else
          Aconstit=configuration+relaxTime*tmpFactor*configUCD
      	    -(tmpFactor*tmpFactor2)*unit;
      	    
        Array configTest((TYPE="arrayTest",TEST=configuration,
      	  		  NAME="Test for config equations"));
        configurationExpr=accumulation(configTest,Aconstit);
        configurationExpr.setName("Expression for configuration equations");
      }
      
      extraStressTensor=(elastVisc*inverse(relaxTime))
	*(tmpFactor3*configuration-tmpFactor2*unit);
      extraStressTensor.setName("extra stress tensor");
    }
    catchAll("FENE::FENE",
	     "at the end of configuration equation.");
    
    // Momentum equations:
    
    Array momentumExpr;
    
    try {
      rho=Array((TYPE="array",ORDER=0,COMPONENT()=dRho,TENSORCOORDSYS=tcs,
		 NAME="density"));
      
      strainRateTensor=0.5*(gradVeloTensor+transpose(gradVeloTensor));
      strainRateTensor.setName("strain rate tensor");
      strainRate=sqrt(2.0*accumulation(strainRateTensor,strainRateTensor));
      strainRate.setName("strain rate");
      
      if (boolFNewtVisc) {
	newtVisc=Array(strainRate,*fNewtVisc);
	newtVisc.setName("Newtonian viscosity");
      }
      else 
        newtVisc=Array((TYPE="array",ORDER=0,COMPONENT()=dNewtVisc,
			TENSORCOORDSYS=tcs,NAME="Newtonian viscosity"));
      
      NewtonStressTensor=2.0*newtVisc*strainRateTensor-pressure*unit;	
      NewtonStressTensor.setName("Newtonian part of stress tensor");
      
      CauchyStressTensor=extraStressTensor+NewtonStressTensor;
      CauchyStressTensor.setName("Cauchy stress tensor");
      
      if (boolTransient) {
	Array velocitiesTimeDeriv=timeDeriv(velocities,localTimeDep);
	velocitiesTimeDeriv.setName("time derivative of Velocities");
	if (boolCoordMobil)
          inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication((velocities-coordinatesTimeDeriv),0,gradVeloTensor,1));
	else
          inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication(velocities,0,gradVeloTensor,1));
      }
      else
        inertia=rho*contractedMultiplication(velocities,0,gradVeloTensor,1);
      inertia.setName("inertia tensor");
      
      Array testMomentum((TYPE="arrayTest",TEST=velocities,
			  NAME="test for momentum equations"));  
      Array gradTestMomentum=grad(testMomentum);
      gradTestMomentum.setName("gradient of test for momentum equations.");
      
      if (gravityArrayFound) {
	force=rho*gravity;
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+force+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+force);
      }
      else if (forceArrayFound) {
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+force+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+force);
      }
      else {
	if (gradPnArrayFound)
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia+gradPn*normal);
	else
	  momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	    +accumulation(testMomentum,inertia);
      }
      momentumExpr.setName("momentum expression");
    }
    catchAll("FENE::FENE",
	     "at the end of momentum equations.");
    
    // Incompressibility equation:
    
    Array incompressibilityExpr;
    try {
      Array divV=contract(gradVeloTensor);
      divV.setName("divergence of velocities");
      Array testIncompressibility((TYPE="arrayTest",TEST=pressure,
				   NAME="test for incompressibility equation"));
      incompressibilityExpr=testIncompressibility*divV;
      incompressibilityExpr.setName("incompressibility expression");
    }
    catchAll("FENE::FENE",
	     "at the end of incompressibility equation.");
    
    // Flow rate equation:
    
    Array flowRateExpr;
    
    try {
      if (flowRateArrayFound) {
	Array veloN=accumulation(velocities,normal);
	veloN.setName("normal component of velocities");
	testFlowRate=Array((TYPE="arrayTest",TEST=gradPn,
			    NAME="test for flow rate equations"));
	flowRateExpr=testFlowRate*veloN;
	flowRateExpr.setName("flowRate expression");
      }
    }
    catchAll("FENE::FENE",
	     "at the end of flow rate equation.");
    
    // Integration of all the equations:
    
    try {
      Array totalExpr=configurationExpr;
      if (momentumExpr.getData()!=NULL)
	totalExpr=totalExpr+momentumExpr;
      if (incompressibilityExpr.getData()!=NULL)
	totalExpr=totalExpr+incompressibilityExpr;
      if (flowRateExpr.getData()!=NULL)
	totalExpr=totalExpr+flowRateExpr;
      totalExpr.setName("addition of integrable expressions");
      
      if (flowRateExpr.getData()!=NULL)
	Array::operator=(integrate(totalExpr,dom,itg)-testFlowRate*flowRate);
      else
	Array::operator=(integrate(totalExpr,dom,itg));
      
      setName(objectName);
    }
    catchAll("FENE::FENE","at the end of integration.");
  }
  catchAll("FENE::FENE","at the very end.");
}
