
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <typeinfo>
#include <iostream>
#include <iomanip>

#include "Array.h"
#include "array_base.h"
#include "arrayComp.h"
#include "arrayField.h"
#include "arrayFilter.h"
#include "arrayFunctionsBuild.h"
#include "arrayInteg.h"
#include "arrayOp1.h"
#include "arrayOp2.h"
#include "arrayOpN.h"
#include "arrayTime.h"
#include "coordSys/GlobLoc.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/TensorCoordSys.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "field/Field.h"
#include "field/field_base.h"
#include "filter/Filter.h"
#include "filter/filter_base.h"
#include "object/Object.h"
#include "param/param.h"
#include "param/paramList.h"
#include "scal/scal.h"
#include "step/step.h"
#include "step/stepList.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



Array::Array(void) {
  try {
    data=NULL;
  }
  catchAll("Array::Array(void)","at the very end")
}



Array::Array(const Array &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("Array::Array(const Array &in)","at the very end")  
}



Array::Array(const Array &in,function1 &F) {
  array *arg=in.getData();
  arrayOp1 *newArray=functionBuild(*arg,&F);
  data=NULL;
  setData(dynamic_cast<array*>(newArray));
}



Array::Array(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Array A(PL);
    data=NULL;
    setData(A.getData());
  }
  catchAll("Array::Array(param P)","at the very end")  
}



Array::Array(paramList PL) {
  
  try {
    
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plLayer=PL.getListOfType(LAYER);
    PL=PL.getListNotOfType(LAYER);
    paramList plOrder=PL.getListOfType(ORDER);
    PL=PL.getListNotOfType(ORDER);
    paramList plDim=PL.getListOfType(DIM);
    PL=PL.getListNotOfType(DIM);
    paramList plDeriv=PL.getListOfType(DERIV);
    PL=PL.getListNotOfType(DERIV);
    paramList plTCS=PL.getListOfType(TENSORCOORDSYS);
    PL=PL.getListNotOfType(TENSORCOORDSYS);
    paramList plComponent=PL.getListOfType(COMPONENT);
    PL=PL.getListNotOfType(COMPONENT);
    paramList plTest=PL.getListOfType(TEST);
    PL=PL.getListNotOfType(TEST);
    paramList plExpr=PL.getListOfType(EXPR);
    PL=PL.getListNotOfType(EXPR);
    paramList plFilter=PL.getListOfType(FILTER);
    PL=PL.getListNotOfType(FILTER);
    paramList plType=PL.getListOfType(TYPE);
    PL=PL.getListNotOfType(TYPE);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      errWrite("Array construction",tmpString);
    }
    
    if (plType.getNbrParams()!=1)
      errWrite("Array construction","exactly one TYPE parameter is expected.");
    
    if (!err::getOk())
      errThrow("Array construction","bailing out due to preceding errors.");
    
    // Creation of the output array :
    
    array *output=NULL;
    if (plType[0].getStringData()=="array") {
      output=new array;
    }
    else if (plType[0].getStringData()=="arrayComp") {
      arrayComp *outputA=new arrayComp;
      output=dynamic_cast<array*>(outputA);
    }
    else if (plType[0].getStringData()=="arrayField") {
      arrayField *outputA=new arrayField;
      output=dynamic_cast<array*>(outputA);
    }
    else if (plType[0].getStringData()=="arrayTest")
      output=NULL;
    else if (plType[0].getStringData()=="arrayFilter")
      output=NULL;
    else
      errWrite("Array construction",string("invalid TYPE= argument :  ")
	       +plType[0].getStringData());
    
    if (!err::getOk())
      errThrow("Array construction","bailing out due to preceding errors");
    
    if (plType[0].getStringData()!="arrayTest"
	&&plType[0].getStringData()!="arrayFilter") {
      
      // ******************************************************
      // *                                                    *
      // *  Follows the parameters common to all array types  *
      // *     except arrayTest and arrayFilter               *
      // *                                                    *
      // ******************************************************
      
      try {
	
	if (plTest.getNbrParams()>0)
	  errWrite("arrayTest or arrayFilter type Array construction",
		   "no TEST= parameter is allowed.");
	
	if (plFilter.getNbrParams()>0)
	  errWrite("arrayTest or arrayFilter type Array construction",
		   "no FILTER= parameter is allowed.");
	
	if (plExpr.getNbrParams()>0)
	  errWrite("arrayTest or arrayFilter type Array construction",
		   "no EXPR= parameter is allowed.");
	
	// Sets the order of the array :    
	Int arrayOrder=0;
	if (plOrder.getNbrParams()>1) {
	  errWrite("Array construction",
		   "maximum one ORDER= parameter is expected.");
        }
	else if (plOrder.getNbrParams()==1)
	  arrayOrder=plOrder[0].getIntData();
	output->setOrder(arrayOrder);    
	if (!err::getOk())
	  errThrow("Array construction","bailing out due to preceding errors");
	
	// Sets indices dimensions :
	Int dim;
	if (plDim.getNbrParams()==1) {
	  const Int *components=plDim[0].getComponentsData();
	  if (components[0]==-1) {
	    dim=plDim[0].getIntData();
	    for (Int i=0;i<arrayOrder;i++)
	      output->setDim(i,dim);
	  }
	  else {
	    if (components[1]!=-1) {
	      errWrite("Array construction",
		       "an index is not defined by more that one component for defining a dimension.");
	    }
	    else if (components[0]==0&&arrayOrder==1) {
	      dim=plDim[0].getIntData();
	      output->setDim(components[0],dim);
	    }
	    else
	      errWrite("Array construction","all indices must be given a dimension.");
	  }
	}
	else {
	  if (plDim.getNbrParams()!=arrayOrder) {
	    errWrite("Array construction","all indices must be given a dimension.");
	  }
	  else {
	    for (Int i=0;i<arrayOrder;i++) {
	      param pDim=plDim[i];
	      bool dimDefined[MAX_ORDER_ARRAY];
	      for (Int j=0;j<MAX_ORDER_ARRAY;j++)
		dimDefined[j]=false;
	      const Int *components=pDim.getComponentsData();
	      if (components[0]==-1||components[1]!=-1) {
		errWrite("Array construction",
			 "a single index dimension is defined by exactly one component");
	      }
	      else {
		dim=plDim[i].getIntData();
		output->setDim(components[0],dim);
		if (dimDefined[components[0]])
		  errWrite("Array construction",
			   string("index ")+strFromInt(components[0])
			   +" is given several dimensions.");
		dimDefined[components[0]]=true;
	      }
	    }
	  }
	}
	if (!err::getOk())
	  errThrow("Array construction","bailing out due to preceding errors");
	
	// Sets the tensorCoordSys of the array :    
	if (plTCS.getNbrParams()>1) {
	  errWrite("Array construction",
		   "only one TENSORCOORDSYS argument is allowed.");
	}
	else if (plTCS.getNbrParams()==1) {
	  tensorCoordSys *TCS
	    =dynamic_cast<tensorCoordSys*>(plTCS[0].getParentData());
	  if (TCS==NULL)
	    errWrite("Array construction",
		     "TENSORCOORDSYS= has to be followed by a TensorCoordSys argument.");
	  output->setTCS(TCS);
	}    
	if (!err::getOk())
	  errThrow("Array construction","bailing out due to preceding errors");
	
      }
      catchAll("Array::Array(paramList PL)","in generic part of constructor");
      
      // ********************
      // *                  *
      // *  Specific parts  *
      // *                  *
      // ********************
      
      if (plType[0].getStringData()=="array") {
	
	// ******************************
	// *                            *
	// *  Creation of normal array  *
	// *                            *
	// ******************************
	
	try {
	  
	  array *outputA=output;
	  
	  // Checks for illegal arguments :
	  
	  if (plTest.getNbrParams()>0)
	    errWrite("array type Array construction",
		     "no TEST= parameter is allowed.");
	  
	  if (plDeriv.getNbrParams()>0) 
	    errWrite("array type Array construction",
		     "no DERIV= parameter is allowed.");
	  
	  if (!err::getOk())
	    errThrow("array type Array construction",
		     "bailing out due to preceding errors");
	  
	  // sets the component values :
	  
	  Int nbrField=0;
	  Int nbrArray=0;
	  for (Int i=0;i<plComponent.getNbrParams();i++) {
	    parents *tmpP=plComponent[i].getParentData();
	    if (dynamic_cast<field*>(tmpP)!=NULL)
	      nbrField++;
	    else if (dynamic_cast<array*>(tmpP)!=NULL)
	      nbrArray++;
	  }
	  if (nbrField>0||nbrArray>0) {
	    errThrow("array type Array construction",
		     "components cannot point to Field or Array.");
	  }
	  else {
	    for (Int i=0;i<plComponent.getNbrParams();i++) {
	      param PC=plComponent[i];
	      outputA->component(PC.getComponentsData())=PC.getDoubleData();
	    }
	  }
	  
	}
	catchAll("Array::Array(paramList PL)","in part specific to array type array");
      }
      else if (plType[0].getStringData()=="arrayComp") {
	
	try {
	  
	  arrayComp *outputA=dynamic_cast<arrayComp*>(output);
	  
	  // ***************************
	  // *                         *
	  // *  Creation of arrayComp  *
	  // *                         *
	  // ***************************
	  
	  // Checks for illegal arguments :
	  
	  if (plTest.getNbrParams()>0)
	    errWrite("arrayComp type Array construction",
		     "no TEST= parameter is allowed.");
	  
	  if (plDeriv.getNbrParams()>0)
	    errWrite("arrayComp type Array construction",
		     "no DERIV= parameter is allowed.");
	  
	  if (!err::getOk())
	    errThrow("arrayComp type Array construction",
		     "bailing out due to preceding errors");
	  
	  // sets the component values or pointers :
	  
	  Int nbrField=0;
	  Int nbrArray=0;
	  field *lastField=NULL;
	  for (Int i=0;i<plComponent.getNbrParams();i++) {
	    parents *tmpP=plComponent[i].getParentData();
	    if (dynamic_cast<field*>(tmpP)!=NULL) {
	      lastField=dynamic_cast<field*>(tmpP);
	      nbrField++;
	    }
	    else if (dynamic_cast<array*>(tmpP)!=NULL)
	      nbrArray++;
	  }
	  
	  if (nbrField!=0)
	    errWrite("arrayComp type Array construction",
		     "no Field argument after COMPONENT= is expected.");
	  
	  if (nbrArray==0)
	    errWrite("arrayComp type Array construction",
		     "COMPONENT=Array arguments are expected.");
	  
	  if (!err::getOk())
	    errThrow("arrayComp type Array construction",
		     "bailing out due to preceding errors");
	  
	  for (Int i=0;i<plComponent.getNbrParams();i++) {
	    param PC=plComponent[i];
	    array *tmpA=dynamic_cast<array*>(PC.getParentData());
	    if (tmpA!=NULL) {
	      tmpA->setCurrentComponent(PC.getIntData());
	      outputA->component(PC.getComponentsData())=(*tmpA);
	      tmpA->setCurrentComponent(-1);
	    }
	    else
	      outputA->component(PC.getComponentsData())=PC.getDoubleData();
	  }
	  
	}
	catchAll("Array::Array(paramList PL)","in part specific to arrayComp type array");
	
      }
      else if (plType[0].getStringData()=="arrayField") {
	
	try {
	  
	  arrayField *outputA=dynamic_cast<arrayField*>(output);
	  
	  // ****************************
	  // *                          *
	  // *  Creation of arrayField  *
	  // *                          *
	  // ****************************
	  
	  // Checks for illegal arguments :
	  
	  if (plTest.getNbrParams()>0) 
	    errThrow("arrayField type Array construction",
		     "no TEST= parameter is allowed.");
	  
	  // sets the component values :
	  
	  Int nbrField=0;
	  Int nbrArray=0;
	  field *lastField=NULL;
	  for (Int i=0;i<plComponent.getNbrParams();i++) {
	    parents *tmpP=plComponent[i].getParentData();
	    if (dynamic_cast<field*>(tmpP)!=NULL) {
	      if (dynamic_cast<field*>(tmpP)!=lastField) {
		nbrField++;
		lastField=dynamic_cast<field*>(tmpP);
	      }
	    }
	    else if (dynamic_cast<array*>(tmpP)!=NULL)
	      nbrArray++;
	  }
	  
	  if (nbrArray>0)
	    errThrow("arrayField type Array construction",
		     "COMPONENT= cannot be followed by an Array argument.");
	  
	  if (nbrField!=1) {
	    errThrow("arrayField type Array construction",
		     "exactly one Field must be pointed to by components.");
	  }
	  else {
	    outputA->setField(lastField);
	    for (Int i=0;i<plComponent.getNbrParams();i++) {
	      param PC=plComponent[i];
	      field *tmpF=dynamic_cast<field*>(PC.getParentData());
	      if (tmpF!=NULL) {
		if (tmpF!=lastField) 
		  errThrow("arrayField type Array construction",
			   "exactly one Field as component argument is allowed");
		tmpF->setCurrentComponent(PC.getIntData());
		outputA->component(PC.getComponentsData())=(*tmpF);
		tmpF->setCurrentComponent(-1);
	      }
	      else
		outputA->component(PC.getComponentsData())=PC.getDoubleData();
	    }
	  }
	  
	  // Derivation of array :
	  if (plDeriv.getNbrParams()>1) {
	    errThrow("arrayField type Array construction",
		     "maximum one DERIV argument is allowed.");
	  }
	  else if (plDeriv.getNbrParams()==1) {
	    outputA->setDeriv(plDeriv[0].getBoolData());
	  }	  
	}
	catchAll("Array::Array(paramList PL)","in part specific to arrayField type array");
      }
    } // if (plType[0].getStringData()!="arrayTest"
      //     &&plType[0].getStringData()!="arrayFilter")
    else if (plType[0].getStringData()=="arrayTest") {
      
      try {
	
	// ***************************
	// *                         *
	// *  Creation of arrayTest  *
	// *                         *
	// ***************************
	
	// Checks for illegal arguments :
	
	if (plDeriv.getNbrParams()>0)
	  errWrite("arrayTest type Array construction",
		   "no DERIV= parameter is allowed.");
	
	if (plOrder.getNbrParams()!=0)
	  errWrite("arrayTest type Array construction",
		   "no ORDER= parameter is allowed.");
	
	if (plDim.getNbrParams()!=0)
	  errWrite("arrayTest type Array construction",
		   "no DIM= parameter is allowed.");
	
	if (plComponent.getNbrParams()!=0)
	  errWrite("arrayTest type Array construction",
		   "no COMPONENT= parameter is allowed.");
	
	if (plTCS.getNbrParams()!=0)
	  errWrite("arrayTest type Array construction",
		   "no TENSORCOORDSYS= parameter is allowed.");
	
	if (plFilter.getNbrParams()>0)
	  errWrite("arrayTest type Array construction",
		   "no FILTER= parameter is allowed.");
	
	if (plExpr.getNbrParams()>0)
	  errWrite("arrayTest type Array construction",
		   "no EXPR= parameter is allowed.");
	
	if (!err::getOk())
	  errThrow("arrayTest type Array construction",
		   "bailing out due to preceding errors");
	
	arrayField *outputA=NULL;
	
	parents *tmpParents=plTest[0].getParentData();
	arrayField *AF=dynamic_cast<arrayField*>(tmpParents);
	if (AF==NULL) {
	  errThrow("arrayTest type Array construction",
		   "TEST= must be followed by an arrayField type array.");
	}
	else {
	  outputA=new arrayField(*AF);
	  outputA->setLigDeriv(false);
	}
	
	output=dynamic_cast<array*>(outputA);
      }
      catchAll("Array::Array(paramList PL)","in part specific to arrayTest type array");
    } // else if (plType[0].getStringData()=="arrayTest")
    else if (plType[0].getStringData()=="arrayFilter") {
      
      try {
	
	// *****************************
	// *                           *
	// *  Creation of arrayFilter  *
	// *                           *
	// *****************************
	
	// Checks for illegal arguments :
	
	if (plDeriv.getNbrParams()>0)
	  errWrite("arrayFilter type Array construction",
		   "no DERIV= parameter is allowed.");
	
	if (plOrder.getNbrParams()!=0)
	  errWrite("arrayFilter type Array construction",
		   "no ORDER= parameter is allowed.");
	
	if (plDim.getNbrParams()!=0)
	  errWrite("arrayFilter type Array construction",
		   "no DIM= parameter is allowed.");
	
	if (plComponent.getNbrParams()!=0)
	  errWrite("arrayFilter type Array construction",
		   "no COMPONENT= parameter is allowed.");
	
	if (plTCS.getNbrParams()!=0)
	  errWrite("arrayFilter type Array construction",
		   "no TENSORCOORDSYS= parameter is allowed.");
	
	if (plTest.getNbrParams()!=0)
	  errWrite("arrayFilter type Array construction",
		   "no TEST= parameter is allowed.");
	
	if (plExpr.getNbrParams()!=1)
	  errWrite("arrayFilter type Array construction",
		   "exactly one EXPR= parameter is needed.");
	
	if (!err::getOk())
	  errThrow("arrayFilter type Array construction",
		   "bailing out due to preceding errors.");
	
	arrayFilter *outputA=NULL;
	
	parents *tmpParents=plExpr[0].getParentData();
	array *AExpr=dynamic_cast<array*>(tmpParents);
	if (AExpr==NULL) {
	  errThrow("arrayTest type Array construction",
		   "EXPR= must be followed by an Array.");
	}
	else
	  outputA=new arrayFilter(*AExpr);
	
	outputA->setExpr(AExpr);
	
	Int nbrFilters=plFilter.getNbrParams();
	outputA->setNbrFilters(nbrFilters);
	for (Int i=0;i<nbrFilters;i++) {
	  parents *tmpP=plFilter[i].getParentData();
	  filter *tmpF=dynamic_cast<filter*>(tmpP);
	  if (tmpF==NULL) {
	    errThrow("Array::Array(paramList PL)",
		     "FILTER= has to be followed by a Filter argument.");
	  }
	  else
	    outputA->setFilter(i,tmpF);
	}
	
	output=dynamic_cast<array*>(outputA);
      }
      catchAll("Array::Array(paramList PL)","in part specific to arrayFilter type array");
      
    } // else if (plType[0].getStringData()=="arrayFilter")
    
    // ******************************************************
    // *                                                    *
    // *  Follows the parameters common to all array types  *
    // *                                                    *
    // ******************************************************
    
    // sets the component names :  
    // not yet implemented !
    
    // Sets the name of the array :  
    
    if (plName.getNbrParams()>1) {
      errWrite("Array construction","maximum one NAME argument is allowed.");
    }
    else if (plName.getNbrParams()==1) {
      string arrayName=plName[0].getStringData();
      if (arrayName=="")
	errWrite("Array construction",
		 "NAME= has to be followed by a string argument.");
      if (output!=NULL)
        output->setName(arrayName);
      else
	errWrite("Array construction",
		 "unable to set name of NULL array.");
    }
    
    // Sets the layer of the array :  
    
    if (plLayer.getNbrParams()>1) {
      errWrite("Array construction","maximum one LAYER argument is allowed.");
    }
    else if (plLayer.getNbrParams()==1) {
      if (plLayer[0].getTypeName()!="Int") {
	errWrite("Array construction",
		 "LAYER= has to be followed by an Int argument.");
      }
      Int arrayLayer=plLayer[0].getIntData();
      if (output!=NULL)
        output->setLayer(arrayLayer);
      else {
	errWrite("Array construction",
		 "unable to set layer of NULL array.");
      }
    }
    
    // Final check :
    
    if (!err::getOk())
      errThrow("Array construction","bailing out due to preceding errors");
    
    data=NULL;
    setData(output);
    
  }
  catchAll("Array::Array(paramList PL)","at the very end of the function");
}



Array::~Array(void) {
  try {
    setData(NULL);
  }
  catchAll("Array::~Array(void)","at the very end")
}



void Array::init(void) {
  try{
    array *arrayIn=new array;
    setData(arrayIn);
  }
  catchAll("Array::init(void)","at the very end")  
}



void Array::reset(void) {
  try{
    init();
  }
  catchAll("Array::reset(void)","at the very end")  
}



void Array::setData(array *newArrayData) {
  try{
    array *A=getData();
    delinkAndDelete(A,&data);
    if (newArrayData!=NULL)
      newArrayData->addParent(&data);
    data=static_cast<void*>(newArrayData);
  }
  catchAll("Array::setData","at the very end")  
}



array* Array::getData(void) const {
  try{
    return(static_cast<array*>(data));
  }
  catchAll("Array::getData","at the very end")  
}



void Array::setName(string s) {
  try{
    array *arrayData=getData();
    arrayData->setName(s);
  }
  catchAll("Array::setName","at the very end")  
}



string Array::getName(void) const {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL)
      return(arrayData->getName());
    else
      return("");
  }
  catchAll("Array::getName","at the very end")  
}



void Array::setLayer(Int nbr) {
  try{
    array *arrayData=getData();
    arrayData->setLayer(nbr);
  }
  catchAll("Array::setLayer","at the very end")  
}



Int Array::getLayer(void) const {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL)
      return(arrayData->getLayer());
    else
      return(0);
  }
  catchAll("Array::getLayer","at the very end")  
}



Int Array::getOrder(void) const {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL)
      return(arrayData->getOrder());
    else
      return(-1);
  }
  catchAll("Array::getOrder","at the very end")  
}



Int Array::getDim(Int index) const {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL)
      return(arrayData->getDim(index));
    else
      return(0);
  }
  catchAll("Array::getDim","at the very end")  
}



void Array::setTCS(TensorCoordSys &TCS) {
  try {
    array *arrayData=getData();
    tensorCoordSys *tcsData=TCS.getData();
    if (arrayData!=NULL)
      arrayData->setTCS(tcsData);
    else
      errThrow("Array::setTCS","NULL Array!");
  }
  catchAll("Array::setTCS","at the very end")  
}



TensorCoordSys Array::getTCS(void) const {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL) {
      tensorCoordSys *tcs=arrayData->getTCS();
      TensorCoordSys TCS;
      TCS.setData(tcs);
      return(TCS);
    }
    else {
      TensorCoordSys TCS;
      return(TCS);
    }
  }
  catchAll("Array::getTCS","at the very end")  
}



Array& Array::component(const Int &index0,const Int &index1,
                        const Int &index2,const Int &index3,
                        const Int &index4,const Int &index5,
                        const Int &index6,const Int &index7) {
  try {
    array *arrayData=getData();
    if (arrayData!=NULL)
      arrayData->component(index0,index1,index2,index3,
			   index4,index5,index6,index7);
  }
  catchAll("Array::component","at the very end");
  
  return(*this);   
}



Array Array::makeDeriv(Int nbrScal,scal **scalList,GlobLoc &GL,
		       string derivType) const {
  try {
    array *tmpArray=getData();
    
    if (tmpArray==NULL)
      errThrow("Array::makeDeriv","uninitialized Array");
	
    array *newArray=tmpArray->makeDeriv(nbrScal,scalList,GL.getData(),
					derivType);
    Array NewArray;
    NewArray.setData(newArray);
    
    return(NewArray);
  }
  catchAll("Array::makeDeriv","at the very end");
}



Array Array::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  try {
    array *tmpArray=getData();
    
    if (tmpArray==NULL)
      errThrow("Array::makeExplicitDeriv","uninitialized Array");
	
    array *newArray=tmpArray->makeExplicitDeriv(nbrScal,scalList);
    Array NewArray;
    NewArray.setData(newArray);
    
    return(NewArray);
  }
  catchAll("Array::makeExplicitDeriv","at the very end");
}



Array Array::makeCoordDeriv(GlobLoc &GL) const {
  try {
    array *tmpArray=getData();
    
    if (tmpArray==NULL)
      errThrow("Array::makeCoordDeriv","uninitialized Array");
	
    array *newArray=tmpArray->makeCoordDeriv(GL.getData());
    Array NewArray;
    NewArray.setData(newArray);
    
    return(NewArray);
  }
  catchAll("Array::makeCoordDeriv","at the very end");
}



Array Array::makeCorlocDeriv(Int nbrDP) const {
  try {
    array *tmpArray=getData();
    
    if (tmpArray==NULL)
      errThrow("Array::makeCorlocDeriv","uninitialized Array");
	
    array *newArray=tmpArray->makeCorlocDeriv(nbrDP);
    Array NewArray;
    NewArray.setData(newArray);
    
    return(NewArray);
  }
  catchAll("Array::makeCorlocDeriv","at the very end");
}



Array Array::makePhysicalDeriv(GlobLoc &GL) const {
  try {
    array *tmpArray=getData();
    
    if (tmpArray==NULL)
      errThrow("Array::makePhysicalDeriv","uninitialized Array");
	
    array *newArray=tmpArray->makePhysicalDeriv(GL.getData());
    Array NewArray;
    NewArray.setData(newArray);
    
    return(NewArray);
  }
  catchAll("Array::makePhysicalDeriv","at the very end");
}



void Array::write(void) const {
  try {
    array *arrayData=getData();
    
    if (arrayData==NULL)
      errThrow("Array::write","uninitialized Array");
	
    if (arrayData!=NULL)
      arrayData->write();
    else
      cerr << "Array contains no data!" << endl ;
  }
  catchAll("Array::write","at the very end");
}



void Array::writeValues(void) const {
  try {
    array *arrayData=getData();
    
    if (arrayData==NULL)
      errThrow("Array::writeValues","uninitialized Array");
	
    if (arrayData!=NULL)
      arrayData->writeValues();
    else
      cerr << "Array contains no data!" << endl ;
  }
  catchAll("Array::writeValues","at the very end");
}



void Array::write(ostream &out) const {
  try {
    array *arrayData=getData();
    
    if (arrayData==NULL)
      errThrow("Array::write","uninitialized Array");
	
    if (arrayData!=NULL)
      arrayData->write(out);
    else
      out << "Array contains no data!" << endl ;
  }
  catchAll("Array::write","at the very end");
}



void Array::writeValues(ostream &out) const {
  try {
    array *arrayData=getData();
    
    if (arrayData==NULL)
      errThrow("Array::writeValues","uninitialized Array");
	
    if (arrayData!=NULL)
      arrayData->writeValues(out);
    else
      out << "Array contains no data!" << endl ;
  }
  catchAll("Array::writeValues","at the very end");
}



void Array::makeAtPoint(Object &obj,Double *localCoord) {
  try {
    array *arrayData=getData();
    if (arrayData==NULL)
      errThrow("Array::makeAtPoint","attempt to make void Array");
    
    arrayData->setNewPoint(true);
    arrayData->makeAtPoint(obj.getData(),obj.getDirect(),localCoord,true,false);
  }
  catchAll("Array::makeAtPoint","at the very end");
}



void Array::makeOnObject(Object &obj) {
  try {
    array *arrayData=getData();
    if (arrayData==NULL)
      errThrow("Array::makeOnObject","attempt to make void Array");
    
    arrayData->setNewPoint(true);
    arrayData->makeOnObject(obj.getData(),obj.getDirect(),true,false);
  }
  catchAll("Array::makeOnObject","at the very end");
}



void Array::operator=(const Array &Ain) {
  if (this==&Ain)
    return;
  try {
    setData(Ain.getData());
  }
  catchAll("Array::operator= A","at the very end");
}



void Array::operator=(Double d) {
  try {
    if (getData()==NULL)
      errThrow("Array::operator=(Double d)","uninitialized Array");
	
    *(getData())=d;
  }
  catchAll("Array::operator= d","at the very end");
}



Double Array::value(void) {
  try {
    if (getData()==NULL)
      errThrow("Array::value(void)","uninitialized Array");
    
    array *tmpA=getData();
    Int index=tmpA->getCurrentComponent();
    scal *tmpScal=tmpA->getComponentValues();
    tmpA->setCurrentComponent(-1);
    return (tmpScal[index].getVal());
  }
  catchAll("Array::value(void)","at the very end");
}




