
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <set>
#include <string>
#include <stdio.h>

#include "arrayComp.h"
#include "arrayField.h"
#include "arrayOp1.h"
#include "arrayOp2.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "scal/scal.h"
#include "print/printTree.h"
#include "timeDep/timeDep_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



arrayComp::arrayComp(void) : array() {
  GL=NULL;
  arrayList=new array* [nbrVal];
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++) {
    arrayList[i]=NULL;
    compCorresp[i]=-1;
  }
  dependence=NULL;
  dependenceSize=0;
  dependenceCapacity=0;
}



arrayComp::arrayComp(array &in) : array(in) {
  GL=NULL;
  arrayList=new array* [nbrVal];
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++) {
    arrayList[i]=NULL;
    compCorresp[i]=-1;
  }
  dependence=NULL;
  dependenceSize=0;
  dependenceCapacity=0;
}



arrayComp::arrayComp(arrayComp &in) : array(in) {
  Int i;
  
  GL=NULL;
  setGlobLoc(in.getGlobLoc());
  arrayList=new array* [nbrVal];
  compCorresp=new Int[nbrVal];
  for (i=0;i<nbrVal;i++) {
    arrayList[i]=in.arrayList[i];
    arrayList[i]->addParent(arrayList+i);
    compCorresp[i]=in.compCorresp[i];
  }
  dependenceSize=in.dependenceSize;
  dependenceCapacity=dependenceSize;
  if (dependenceCapacity>0)
    dependence=new array* [dependenceCapacity];
  else
    dependence=NULL;
  for (i=0;i<dependenceSize;i++)
    dependence[i]=in.dependence[i];
}



arrayComp::~arrayComp(void) {
  setGlobLoc(NULL);
  for (Int i=0;i<nbrVal;i++)
    delinkAndDelete(arrayList[i],arrayList+i);
  delete[] arrayList;
  arrayList=NULL;
  delete[] compCorresp;
  compCorresp=NULL;
  
  delete[] dependence;
  dependence=NULL;
  dependenceSize=0;
  dependenceCapacity=0;
}



array* arrayComp::makeCloneForArrayTime(timeDep &TD) {
  Int i,j;
  
  arrayComp *newArray=new arrayComp;
  
  newArray->setGlobLoc(getGlobLoc());
  
  newArray->copyArrayData(*this);
  
  newArray->arrayList=new array* [nbrVal];
  newArray->compCorresp=new Int[nbrVal];
  newArray->dependenceSize=dependenceSize;
  newArray->dependenceCapacity=dependenceSize;
  newArray->dependence=new array* [dependenceSize];
  for (j=0;j<dependenceSize;j++)
    newArray->dependence[j]=dependence[j]->makeCloneForArrayTime(TD);
  
  for (i=0;i<nbrVal;i++) {
    newArray->arrayList[i]=NULL;
    newArray->compCorresp[i]=compCorresp[i];
    for (j=0;j<dependenceSize;j++) {
      if (arrayList[i]==dependence[j]) {
        newArray->arrayList[i]=newArray->dependence[j];
        newArray->arrayList[i]->addParent(newArray->arrayList+i);
      }
    }
  }
  
  return(newArray);
}



void arrayComp::setOrder(Int nbr) {
  array::setOrder(nbr);
  adaptToNbrVal();
}



void arrayComp::setDim(Int index,Int nbrComponents) {
  array::setDim(index,nbrComponents);
  adaptToNbrVal();
}



void arrayComp::adaptToNbrVal(void) {
  if (arrayList!=NULL)
    delete[] arrayList;
  arrayList=new array* [nbrVal];
  if (compCorresp!=NULL)
    delete[] compCorresp;
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++) {
    arrayList[i]=NULL;
    compCorresp[i]=-1;
  }
  if (dependence!=NULL)
    delete[] dependence;
  dependence=NULL;
  dependenceSize=0;
  dependenceCapacity=0;
}



void arrayComp::makeDependence(void) {
  try {
    Int i;
    set<array*> arraySet;
    
    for (i=0;i<nbrVal;i++)
      if (arrayList[i]!=NULL)
	arraySet.insert(arrayList[i]);
    dependenceSize=arraySet.size();
    if (dependenceSize>dependenceCapacity) {
      delete[] dependence;
      dependenceCapacity=dependenceSize;
      dependence=new array* [dependenceCapacity];
    }
    
    set<array*>::iterator j;
    i=0;
    for (j=arraySet.begin();j!=arraySet.end();j++) {
      dependence[i]=*j;
      i++;
    }
  }
  catchAll("arrayComp::makeDependence","")
}



void arrayComp::operator=(const Double &d) {
  try {
    if (currentComponent==-1)
      errThrow("arrayComp::operator=(const Double &d)",
	       "   only components may be assigned to.");
    
    componentValues[currentComponent]=d;
    if (arrayList[currentComponent]!=NULL)
      delinkAndDelete(arrayList[currentComponent],arrayList+currentComponent);
    arrayList[currentComponent]=NULL;
    compCorresp[currentComponent]=-1;
    makeDependence();
    
    currentComponent=-1;
  }
  catchAll("arrayComp::operator= d","")
}



void arrayComp::operator=(array &A) {
  if (this==&A)
    return;
  try {
    if (currentComponent==-1)
      errThrow("arrayComp::operator=(array &A)",
	       "   only components may be assigned to.");
    
    if (A.getCurrentComponent()==-1&&(&A)!=NULL)
      errThrow("arrayComp::operator=(array &A)",
	       "   only components may be assigned from.");
    
    if (arrayList[currentComponent]!=NULL)
      delinkAndDelete(arrayList[currentComponent],arrayList+currentComponent);
    arrayList[currentComponent]=&A;
    if ((&A)!=NULL)
      A.addParent(arrayList+currentComponent);
    compCorresp[currentComponent]=A.getCurrentComponent();
    makeDependence();
    
    currentComponent=-1;
    A.setCurrentComponent(-1);
  }
  catchAll("arrayComp::operator= A","")
}



void arrayComp::makeAtPoint(object *element,Int direct,Double *localCoord,
			    bool makeValue,bool makeDeriv) {
  Int i;
  
  if (!getNewPoint())
    return;
  newPoint::setNewPoint(false);
  
  try {
    if (GL!=NULL)
      GL->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
    
    newPoint::setNewPoint(false);
    for (i=0;i<dependenceSize;i++)
      dependence[i]->makeAtPoint(element,direct,localCoord,makeValue,
				 makeDeriv);
    
    for (i=0;i<nbrVal;i++)
      if (arrayList[i]!=NULL) {
      if (compCorresp[i]==-1)
	errThrow("arrayComp::makeAtPoint",
		 "   compCorresp=-1 for a non NUll arrayList component.");
      
      componentValues[i]=arrayList[i]->componentValues[compCorresp[i]];
    }
    updateBoolZeroValue();
  }
  catchAll("arrayComp::makeAtPoint","")
}



array* arrayComp::makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
			    string derivType) const {
  try {
    Int i,j,k,index=-1;
    
    if (!(derivType=="explicit"||derivType=="coord"||derivType=="corloc"))
      errThrow("arrayComp::makeDeriv",
	       string("   unknown derivation type : ")+derivType+".");
    
    if (derivType=="coord") {
      if (GL!=NULL) {
	nbrScal=GL->getCoordComponentsNbr();
	scalList=GL->getCoordComponents();
      }
      else
	errThrow("arrayComp::makeDeriv",
		 "   no coord derivative can be calculated if GL==NULL");
    }
    
    arrayComp *newArray=new arrayComp;
    newArray->setGlobLoc(getGlobLoc());
    newArray->setOrder(order+1);
    for (i=0;i<order;i++)
      newArray->setDim(i,dim[i]);
    newArray->setDim(order,nbrScal);
    newArray->adaptToNbrVal();
    
    if (derivType=="explicit"||derivType=="coord") {
      array **tmpDependence=new array* [dependenceSize];
      for (i=0;i<dependenceSize;i++) {
	tmpDependence[i]=dependence[i]->makeDeriv(nbrScal,scalList,GL,
						  derivType);
	tmpDependence[i]->addParent(tmpDependence+i);
      }
      for (i=0;i<nbrVal;i++) {
	for (j=0;j<nbrScal;j++) {
	  for (k=0;k<dependenceSize;k++) {
	    if (arrayList[i]==dependence[k]) {
	      newArray->arrayList[i*nbrScal+j]=tmpDependence[k];
	      tmpDependence[k]->addParent(newArray->arrayList+i*nbrScal+j);
	      newArray->compCorresp[i*nbrScal+j]=compCorresp[i]*nbrScal+j;
	    }
	  }
	}
      }
      
      for (i=0;i<dependenceSize;i++) {
	delinkAndDelete(tmpDependence[i],tmpDependence+i);
	tmpDependence[i]=NULL;
      }
      delete[] tmpDependence;
      tmpDependence=NULL;
      
      Int newIndex;
      for (i=0;i<nbrVal;i++) {
	for (j=0;j<nbrScal;j++) {
	  newIndex=i*nbrScal+j;
	  if (componentValues+i==scalList[j]) {
	    newArray->componentValues[newIndex]=1.0;
	    delinkAndDelete(newArray->arrayList[newIndex],
			    newArray->arrayList+newIndex);
	    newArray->arrayList[newIndex]=NULL;
	    newArray->compCorresp[newIndex]=-1;
	  }
	}
      }
      newArray->makeDependence();
    } // if explicit or coord
    else if (derivType=="corloc") {
      newArray->dependenceSize=dependenceSize;
      newArray->dependenceCapacity=dependenceSize;
      newArray->dependence=new array* [newArray->dependenceSize];
      for (i=0;i<dependenceSize;i++) {
	newArray->dependence[i]=dependence[i]->makeDeriv(nbrScal,scalList,GL,
							 derivType);
	newArray->dependence[i]->addParent(newArray->dependence+i);
      }
      for (i=0;i<nbrVal;i++) {
	for (j=0;j<nbrScal;j++) {
	  if (arrayList[i]!=NULL) {
	    for (k=0;k<dependenceSize;k++)
	      if (arrayList[i]==dependence[k])
		index=k;
	    newArray->arrayList[nbrScal*i+j]=newArray->dependence[index];
	    newArray->dependence[index]
	      ->addParent(newArray->arrayList+nbrScal*i+j);
	    newArray->compCorresp[nbrScal*i+j]=compCorresp[i]*nbrScal+j;
	  }
	  else
	    newArray->componentValues[nbrScal*i+j]=0.0;
	}
      }
    } // if corloc
    
    return newArray;
  }
  catchAll("arrayComp::makeDeriv","")
}



array* arrayComp::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  try {
    return(makeDeriv(nbrScal,scalList,NULL,"explicit"));
  }
  catchAll("arrayComp::makeExplicitDeriv","")
}



array* arrayComp::makeCoordDeriv(globLoc *GL) const {
  try {
    return(makeDeriv(0,NULL,GL,"coord"));
  }
  catchAll("arrayComp::makeCoordDeriv","")
}



array* arrayComp::makeCorlocDeriv(Int nbrDP) const {
  try {
    return(makeDeriv(nbrDP,NULL,NULL,"corloc"));
  }
  catchAll("arrayComp::makeCorlocDeriv","")
}



void arrayComp::setNewPoint(bool b) {
  try {
    if (GL!=NULL)
      GL->setNewPoint(b);
    for (Int i=0;i<dependenceSize;i++)
      dependence[i]->setNewPoint(b);
    newPoint::setNewPoint(b);
  }
  catchAll("arrayComp::setNewPoint","")
}



void arrayComp::accessNbrInit(void) {
  try {
    if (GL!=NULL)
      GL->accessNbrInit();
    for (Int i=0;i<dependenceSize;i++)
      dependence[i]->accessNbrInit();
    newPoint::accessNbrInit();
  }
  catchAll("arrayComp::accessNbrInit","")
}



printTree arrayComp::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int index,indices[MAX_ORDER_ARRAY],argIndex,argIndices[MAX_ORDER_ARRAY];
    Int i,pos;
    array *arrayArg;
    Double d;
    
    sprintf(cbuf,"arrayComp %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    for (index=0;index<nbrVal;index++) {
      getIndicesFromIndex(index,indices);
      sbuf="   component[";
      for (i=0;i<order;i++) {
	if (i==0) 
	  sprintf(cbuf,"%d",indices[i]);
	else
	  sprintf(cbuf,",%d",indices[i]);
	sbuf+=cbuf;
      }
      sprintf(cbuf,"] = ");
      sbuf+=cbuf;
      
      arrayArg=arrayList[index];
      argIndex=compCorresp[index];
      pos=0;
      if (arrayArg!=NULL) {
	while (pos<dependenceSize) {
	  if (arrayArg==dependence[pos])
	    break;
	  pos++;
	}
	if (pos==dependenceSize)
	  errThrow("arrayComp::getPrintTree","   unknown referenced array.");
	
	sprintf(cbuf,"arrayArg%d[",pos+1);
	sbuf+=cbuf;
	arrayArg->getIndicesFromIndex(argIndex,argIndices);
	for (i=0;i<arrayArg->order;i++) {
	  if (i==0)
	    sprintf(cbuf,"%d",argIndices[i]);
	  else
	    sprintf(cbuf,",%d",argIndices[i]);
	  sbuf+=cbuf;
	}
	sprintf(cbuf,"]");
	sbuf+=cbuf;
      }
      else {
	d=componentValues[index].getVal();
	sprintf(cbuf,"%g",d);
	sbuf+=cbuf;
      }
      pt+=sbuf;
    }
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    for (i=0;i<dependenceSize;i++) {
      sprintf(cbuf,"arrayArg%d",i+1);
      ptArg=dependence[i]->getFirstPrintTree(cbuf,values);
      if (i<dependenceSize-1)
	ptArg.prepareForAdd();
      else
	ptArg.prepareForAddLast();
      pt+=ptArg;
    }
    
    return(pt);  
  }
  catchAll("arrayComp::getPrintTree","")
}



set<field*>* arrayComp::buildActiveFieldList(void) {
  try {
    static set<field*> *tmpFL;
    set<field*>::iterator SFit;
    Int i;
    
    FL.clear();
    
    if (GL!=NULL) {
      tmpFL=GL->buildActiveFieldList();
      for (SFit=tmpFL->begin();SFit!=tmpFL->end();SFit++)
	FL.insert(*SFit);
    }
    
    for (i=0;i<dependenceSize;i++) {
      tmpFL=dependence[i]->buildActiveFieldList();
      for (SFit=tmpFL->begin();SFit!=tmpFL->end();SFit++)
	FL.insert(*SFit);
    }
    
    return(&FL);
  }
  catchAll("arrayComp::buildActiveFieldList","")
}



set<parents*> arrayComp::getListOfType(string typeName) {
  try {
    Int i;
    set<parents*> S,tmpS;
    set<parents*>::iterator Sit;
    
    if (isOfType(typeName))
      S.insert(dynamic_cast<parents*>(this));
    
    for (i=0;i<dependenceSize;i++) {
      tmpS=dependence[i]->getListOfType(typeName);
      for (Sit=tmpS.begin();Sit!=tmpS.end();Sit++)
	S.insert(*Sit);
    }
    return(S);
  }
  catchAll("arrayComp::getListOfType","")
}



void arrayComp::setGlobLoc(const globLoc *newGL) {
  try {
    globLoc *oldGL=GL;
    if (oldGL!=NULL)
      delinkAndDelete(oldGL,&GL);
    GL=const_cast<globLoc*>(newGL);
    if (GL!=NULL)
      GL->addParent(&GL);
  }
  catchAll("arrayComp::setGlobLoc","")
}



globLoc* arrayComp::getGlobLoc(void) const {
  return(GL);
}



