
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "arrayField.h"
#include "arrayFunctionsBuild.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "object/object_base.h"
#include "print/printTree.h"
#include "scal/scal.h"
#include "shape/shape.h"
#include "timeDep/timeDep_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



arrayField::arrayField(void) : array() {
  F=NULL;
  
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++)
    compCorresp[i]=-1;
  
  deriv=true;
  ligDeriv=true;
  
  DPDim=0;
  
  globL=NULL;
}



arrayField::arrayField(field &Fin) : array() {
  F=NULL;
  setField(&Fin);;
  
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++)
    compCorresp[i]=-1;
  
  deriv=true;
  ligDeriv=true;
  
  DPDim=0;
  
  globL=NULL;
}



arrayField::arrayField(const array &Ain) : array(Ain) {
  F=NULL;
  
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++)
    compCorresp[i]=-1;
  
  deriv=true;
  ligDeriv=true;
  
  DPDim=0;
  
  globL=NULL;
}



arrayField::arrayField(const arrayField &AFin) : array(AFin) {
  Int i;
  
  F=NULL;
  setField(AFin.getField());
  
  compCorresp=new Int[nbrVal];
  for (i=0;i<nbrVal;i++)
    compCorresp[i]=AFin.compCorresp[i];
  
  deriv=AFin.deriv;
  ligDeriv=AFin.ligDeriv;
  
  DPDim=AFin.DPDim;
  
  globL=NULL;
  setGlobLoc(AFin.getGlobLoc());
}



arrayField::~arrayField(void) {
  setField(NULL);
  
  nbrVal=0;
  delete[] compCorresp;
  compCorresp=NULL;
  
  deriv=true;
  ligDeriv=true;
  
  DPDim=0;
  
  setGlobLoc(NULL);
}



array* arrayField::makeCloneForArrayTime(timeDep &TD) {
  try {
    field *F=getField();
    arrayField *newArray=new arrayField(*this);
    newArray->setField(TD.addActiveField(F));
    newArray->deriv=false;
    return(newArray);
  }
  catchAll("arrayField::makeCloneForArrayTime","")
}



void arrayField::setOrder(Int nbr) {
  array::setOrder(nbr);
  delete[] compCorresp;
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++)
    compCorresp[i]=-1;
}



void arrayField::setDim(Int index,Int nbrComponents) {
  array::setDim(index,nbrComponents);
  delete[] compCorresp;
  compCorresp=new Int[nbrVal];
  for (Int i=0;i<nbrVal;i++)
    compCorresp[i]=-1;
}



void arrayField::setField(field *fref) {
  delinkAndDelete(F,&F);
  F=fref;
  if (F!=NULL)
    F->addParent(&F);
}



field* arrayField::getField(void) const {
  return(F);
}



void arrayField::setGlobLoc(globLoc *GL) {
  delinkAndDelete(globL,&globL);
  globL=GL;
  if (globL!=NULL)
    globL->addParent(&globL);
}



globLoc* arrayField::getGlobLoc(void) const {
  return(globL);
}



void arrayField::setDeriv(bool logic) {
  deriv=logic;
}



void arrayField::setLigDeriv(bool logic) {
  ligDeriv=logic;
}



void arrayField::operator=(field &Farg) {
  // This function is used to set the correspondence between the components
  // of the arrayField and the components of the field.
  if (&Farg!=F)
    errThrow("arrayField = field",
	     "   the assigned field must be the reference field.");
  
  if (currentComponent==-1||F->currentComponent==-1)
    errThrow("arrayField = field","   only components may be assigned.");
  
  compCorresp[currentComponent]=F->currentComponent;
  currentComponent=-1;
  F->currentComponent=-1;
}



void arrayField::makeAtPoint(object *element,Int direct,Double *localCoord,
			     bool makeValue,bool makeDeriv) {
    
  if (!getNewPoint())
    return;
  newPoint::setNewPoint(false);
  
  try {    
    Int i,j,k;
    
    Int interpType=F->interpolationType;
    F->fillLocalNodalValues(element,direct,false,false);
    Int localNodalValuesSize=F->localNodalValuesSize;
    Double *localNodalValues=F->localNodalValues;
    Int nbrComponents=F->nbrComponents;
    Int localNbrNVPerComponent=localNodalValuesSize/nbrComponents;
    domain *defDomain=F->getDefDomain();
    Int objectType;
    if (element!=NULL)
      objectType=element->getType();
    else
      objectType=-1;
    
    Int nbr_SF_PBC=1;
    if (DPDim)
      nbr_SF_PBC=DPDim;
    
    Double funct[MAX_OBJ_PER_ELT];
    Int corlocDeriv[3];
    Int nbrFunct;
    
    static scal *currentScal,currentScalCorlocDeriv[3],tmpCoordDeriv[3];
    Double currentScalCorlocDerivValues[3];
    Double value;
    
    Double *dtmp;
    bool *btmp;

    register Double *diter1,*diter2,*dFinal,dValue;
    
    for (i=0;i<nbrVal;i++)
      if (compCorresp[i]!=-1)
        componentValues[i]=0.0;
  
    // Partial derivatives wrt local coordinates (or no partial derivatives) :
    
    for (k=0;k<nbr_SF_PBC;k++) {
      
      // One calculates the component derivatives :
      
      corlocDeriv[0]=0;
      corlocDeriv[1]=0;
      corlocDeriv[2]=0;
      if (DPDim)
    	corlocDeriv[k]=1;
      
      // The appropriate shape functions are called :
      
      nbrFunct=shape::function(interpType,objectType,defDomain,localCoord,
        		       corlocDeriv,funct);
      
      // The values of the components are calculated :
      
      if (makeValue) {
        for (i=k;i<nbrVal;i+=nbr_SF_PBC) {
          if (compCorresp[i]!=-1) {
            currentScal=componentValues+i;
            value=0.0;
            for (j=0;j<nbrFunct;j++) {
              value+=localNodalValues[compCorresp[i]*localNbrNVPerComponent+j]
        	*funct[j];
            }
            currentScal->val=value;
          } //  if (compCorresp[i]!=-1)
        } // for i
      } // if (makeValue)
      
      // The derivatives with respect to nodal values are calculated :
      
      if (deriv&&makeDeriv) {
        for (i=k;i<nbrVal;i+=nbr_SF_PBC) {
          if (compCorresp[i]!=-1) {
            currentScal=componentValues+i;
            if (ligDeriv) {
              currentScal->setLigSize(nbrFunct);
              dtmp=currentScal->lig;
              btmp=currentScal->ligBool;
              currentScal->ligIndex
        	=F->mappingLocalScal[compCorresp[i]*localNbrNVPerComponent];
            }
            else {
              currentScal->setColSize(nbrFunct);
              dtmp=currentScal->col;
              btmp=currentScal->colBool;
              currentScal->colIndex
        	=F->mappingLocalScal[compCorresp[i]*localNbrNVPerComponent];
            }
            for (j=0;j<nbrFunct;j++) {
              dtmp[j]=funct[j];
              btmp[j]=true;
            }
          } //  if (compCorresp[i]!=-1)
        } // for i
      } // if (deriv&&makeDeriv)
    } // for k  

    for (i=0;i<nbrVal;i++)
      componentValues[i].updateBoolZeroValue();

    if (globL!=NULL) {
    
      // Global coordinates partial derivatives :
    
      if (!DPDim)
        errThrow("arrayField::makeAtPoint","   globL!=NULL&&DPDim==0");
    
      globL->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
      array *invJac=globL->getInvJac();
      
      static bool optiScal;
      
      if (!scal::getWhite()) {
      
        // One first checks whether the optimized version is to be used :
      
        static Int maxCoordLigSize,maxCoordColSize;
        static Int maxThisLigSize,maxThisColSize;
	
	maxCoordLigSize=0;
	maxCoordColSize=0;
        for (i=0;i<invJac->nbrVal;i++) {
	  maxCoordLigSize=max(maxCoordLigSize,
	                      invJac->componentValues[i].ligSize);
	  maxCoordColSize=max(maxCoordColSize,
	                      invJac->componentValues[i].colSize);
	}
	
	maxThisLigSize=0;
	maxThisColSize=0;
        for (i=0;i<nbrVal;i++) {
	  maxThisLigSize=max(maxThisLigSize,componentValues[i].ligSize);
	  maxThisColSize=max(maxThisColSize,componentValues[i].colSize);
	}
	
	if (maxCoordLigSize&&!maxCoordColSize&&!maxThisLigSize&&maxThisColSize)
	  optiScal=true;
	else
	  optiScal=false;	  

      } // if (!scal::getWhite())
      else
        optiScal=false;
	
      if (!optiScal) {
      
        // Classical calculation of the components :
	
        for (k=0;k<nbrVal;k+=nbr_SF_PBC) {
          if (compCorresp[k]!=-1) {
	    for (i=0;i<nbr_SF_PBC;i++) {
	      currentScalCorlocDeriv[i]=componentValues[k+i];
	      componentValues[k+i]=0.0; 
	    }
	    for (i=0;i<nbr_SF_PBC;i++)
	      for (j=0;j<nbr_SF_PBC;j++)
	        componentValues[k+i]+=
		  invJac->componentValues[j*nbr_SF_PBC+i]
		    *currentScalCorlocDeriv[j];
	  } // if (compCorresp[k]!=-1)
	} // for (k=0;k<nbrVal;k+=nbr_SF_PBC)
      } // if (!optiScal)
      else {
      
        // Optimized calculation of the components :
     
        static Int firstNonNULLComp=-1;
        static Int firstNonNULLCompCorresp=-1; 
        static Int tmpColIndex; 
	     
        for (i=0;i<nbrVal;i+=nbr_SF_PBC)
          if (compCorresp[i]!=-1) {
            firstNonNULLComp=i;
      	    firstNonNULLCompCorresp=compCorresp[i];
      	    break;
      	  }
        if (firstNonNULLComp==-1)
          errThrow("arrayField::makeAtPoint",
      	  	   "   no Field component in arrayField.");

        // (First not NULL component calculation) :
	
	for (j=0;j<nbr_SF_PBC;j++)
	  currentScalCorlocDeriv[j]=componentValues[firstNonNULLComp+j];
	for (j=0;j<nbr_SF_PBC;j++) {
          tmpCoordDeriv[j]=0.0;
	  for (k=0;k<nbr_SF_PBC;k++)
	    tmpCoordDeriv[j]+=invJac->componentValues[k*nbr_SF_PBC+j]
	                        *currentScalCorlocDeriv[k];
	}
	for (j=0;j<nbr_SF_PBC;j++)
          componentValues[firstNonNULLComp+j]=tmpCoordDeriv[j];
	  
        // (Other not NULL components calculation) :
	
        for (i=firstNonNULLComp+nbr_SF_PBC;i<nbrVal;i+=nbr_SF_PBC) {
	  if (compCorresp[i]!=-1) {
	    for (j=0;j<nbr_SF_PBC;j++) {
	      currentScalCorlocDerivValues[j]=componentValues[i+j].val;
	      tmpColIndex=componentValues[i+j].colIndex;
	      componentValues[i+j]=tmpCoordDeriv[j];
	      componentValues[i+j].colIndex=tmpColIndex;
	    } // for (j=0;j<nbr_SF_PBC;j++)
	    for (j=0;j<nbr_SF_PBC;j++) {
	      componentValues[i+j].val=0.0;
	      diter1=componentValues[i+j].lig;
	      dFinal=diter1+componentValues[i+j].ligSize;
	      while (diter1<dFinal)
	    	*(diter1++)=0.0;
	    	
	      for (k=0;k<nbr_SF_PBC;k++) {
	        dValue=currentScalCorlocDerivValues[k];
	    	componentValues[i+j].val+=
	    	  invJac->componentValues[j*nbr_SF_PBC+k].val*dValue;
    		diter1=componentValues[i+j].lig+
		  invJac->componentValues[j*nbr_SF_PBC+k].ligIndex-
		    componentValues[i+j].ligIndex;
    		diter2=invJac->componentValues[j*nbr_SF_PBC+k].lig;  
    		dFinal=diter1+invJac->componentValues[j*nbr_SF_PBC+k].ligSize;
    		while (diter1<dFinal)
    		  *(diter1++)+=dValue*(*(diter2++));
	      } // for (k=0;k<nbr_SF_PBC;k++)
	    } // for (j=0;j<nbr_SF_PBC;j++)
	  } // if (compCorresp[i]!=-1)
	} // for (i=firstNonNULLComp+nbr_SF_PBC;i<nbrVal;i+=nbr_SF_PBC)
      } // if (!optiScal) else
    } // if (globL!=NULL)

    for (i=0;i<nbrVal;i++)
      componentValues[i].updateBoolZeroValue();
    updateBoolZeroValue();  
    
  }
  catchAll ("arrayField::makeAtPoint","")
}



array* arrayField::makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
			     string derivType) const {
  try {
    if (derivType=="explicit")
      return(makeExplicitDeriv(nbrScal,scalList));
    else if (derivType=="coord")
      return(makeCoordDeriv(GL));
    else if (derivType=="corloc")
      return(makeCorlocDeriv(nbrScal));
    else
      errThrow("arrayField::makeDeriv",
	       string("   unknown derivation type : ")+derivType+".");
  }
  catchAll("arrayField::makeDeriv","");
}



array* arrayField::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  try {
    return(array::makeDeriv(nbrScal,scalList,NULL,"explicit"));
  }
  catchAll("arrayField::makeExplicitDeriv","");
}



array* arrayField::makeCoordDeriv(globLoc *GL) const {
  try {
    Int i;
    array *newArray;
    
    array *Coord=GL->getCoord();
    if (this==dynamic_cast<arrayField*>(Coord)) {
      newArray=new array;
      newArray->setOrder(order+1);
      for (i=0;i<order;i++)
	newArray->setDim(i,dim[i]);
      newArray->setDim(order,Coord->getDim(0));
      for (i=0;i<newArray->nbrVal;i++)
	newArray->componentValues[i]=0.0;
      for (i=0;i<dim[0];i++)
	newArray->component(i,i)=1.0;
    }
    else {
      array *tmpArray1=makeCorlocDeriv(Coord->getDim(0));
      {
	arrayField *tmpAF=dynamic_cast<arrayField*>(tmpArray1);
	tmpAF->setGlobLoc(GL);
      }
      newArray=tmpArray1;
    }
    return(newArray);
  }
  catchAll("arrayField::makeCoordDeriv","");
}



array* arrayField::makeCorlocDeriv(Int nbrDP) const {
  try {
    Int i;
    
    if (DPDim!=0) {
      errThrow("arrayField::makeCorlocDeriv",
                string("   cannot derive twice wrt local coordinates."));
    }	 arrayField* newArray=new arrayField(*this);
     
    newArray->setTCS(NULL);    
    newArray->DPDim=nbrDP;
    
    newArray->setOrder(order+1);
    for (i=0;i<order;i++)
      newArray->setDim(i,dim[i]);
    newArray->setDim(order,nbrDP);
    
    delete[] newArray->compCorresp;
    newArray->compCorresp=new Int[newArray->nbrVal];
    for (i=0;i<nbrVal*nbrDP;i++)
      newArray->compCorresp[i]=compCorresp[i/nbrDP];
    
    return(newArray);
  }
  catchAll("arrayField::makeCorlocDeriv","");
}



void arrayField::setNewPoint(bool b) {
  try {
    if (globL!=NULL)
      globL->setNewPoint(b);
    newPoint::setNewPoint(b);
  }
  catchAll("arrayField::setNewPoint","")
}



void arrayField::accessNbrInit(void) {
  try {
    newPoint::accessNbrInit();
    if (F!=NULL)
      F->accessNbrInit();
    if (globL!=NULL)
      globL->accessNbrInit();
  }
  catchAll("arrayField::accessNbrInit","")
}



printTree arrayField::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int index,indices[MAX_ORDER_ARRAY];
    Int i;
    
    sprintf(cbuf,"arrayField %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (globL)
      sprintf(cbuf,"   nbr partial deriv wrt global coordinates : %d",
              DPDim);
    else
      sprintf(cbuf,"   nbr partial deriv wrt local coordinates : %d",
              DPDim);    
    sbuf=cbuf;
    pt+=sbuf;
    
    for (index=0;index<nbrVal;index++) {
      getIndicesFromIndex(index,indices);
      sbuf="   component[";
      for (i=0;i<order;i++) {
	if (i==0) 
	  sprintf(cbuf,"%d",indices[i]);
	else
	  sprintf(cbuf,",%d",indices[i]);
	sbuf+=cbuf;
      }
      if (compCorresp[index]!=-1)
        sprintf(cbuf,"] --> field[%d]",compCorresp[index]);
      else
        sprintf(cbuf,"] --> %g",componentValues[index].getVal());
      sbuf+=cbuf;
      pt+=sbuf;
    }
    
    if (deriv)
      sbuf="   deriv=true";
    else
      sbuf="   deriv=false";
    pt+=sbuf;
    
    if (ligDeriv)
      sbuf="   ligDeriv=true";
    else
      sbuf="   ligDeriv=false";
    pt+=sbuf;
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    if (F!=NULL) {
      ptArg=F->getFirstPrintTree("field",false);
      ptArg.prepareForAddLast();
    }
    else {
      sbuf="undefined field";
      ptArg=sbuf;
      sbuf="   ";
      ptArg+="sbuf";
      ptArg.prepareForAddLast();
    }
    pt+=ptArg;
    
    return(pt);
  }
  catchAll("arrayField::getPrintTree","")
}



set<field*>* arrayField::buildActiveFieldList(void) {
  try {
    FL.clear();
    
    if (globL!=NULL)
      FL=(*(globL->buildActiveFieldList()));
    
    if (deriv)
      FL.insert(F);
    
    return(&FL);
  }
  catchAll("arrayField::buildActiveFieldList","")
}



set<parents*> arrayField::getListOfType(string typeName) {
  try {
    set<parents*> S;
    
    if (isOfType(typeName))
      S.insert(dynamic_cast<parents*>(this));
    if (F->isOfType(typeName))
      S.insert(dynamic_cast<parents*>(F));
    
    return(S);
  }
  catchAll("arrayField::getListOfType","")
}
