
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <list>
#include <string>
#include <stdio.h>

#include "arrayFilter.h"
#include "err/err.h"
#include "filter/filter_base.h"
#include "scal/scal.h"
#include "print/printTree.h"
#include "step/step.h"
#include "step/stepList.h"
#include "timeDep/timeDep_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



arrayFilter::arrayFilter(void) : array() {
  expr=NULL;
  nbrFilters=0;
  filterList=NULL;
}



arrayFilter::arrayFilter(array &in) : array(in) {
  expr=NULL;
  nbrFilters=0;
  filterList=NULL;
  lines.clear();
  cols.clear();
}



arrayFilter::arrayFilter(arrayFilter &in) : array(in) {
  expr=NULL;
  setExpr(in.getExpr());
  nbrFilters=0;
  filterList=NULL;
  setNbrFilters(in.getNbrFilters());
  for (Int i=0;i<nbrFilters;i++)
    setFilter(i,in.getFilter(i));
  lines.clear();
  cols.clear();
}



arrayFilter::~arrayFilter(void) {
  setExpr(NULL);
  setNbrFilters(0);
  cleanFilters();
}



void arrayFilter::setExpr(array *A) {
  delinkAndDelete(expr,&expr);
  expr=A;
  if (expr!=NULL)
    expr->addParent(&expr);
}



array* arrayFilter::getExpr(void) const {
  return(expr);
}



void arrayFilter::setNbrFilters(Int nbr) {
  Int i;
  for (i=0;i<nbrFilters;i++)
    setFilter(i,NULL);
  delete[] filterList;
  nbrFilters=nbr;
  filterList=new filter* [nbr];
  for (i=0;i<nbrFilters;i++)
    filterList[i]=NULL;
}



Int arrayFilter::getNbrFilters(void) const {
  return(nbrFilters);
}



void arrayFilter::setFilter(Int pos,filter *F) {
  delinkAndDelete(filterList[pos],filterList+pos);
  filterList[pos]=F;
  if (F!=NULL)
    F->addParent(filterList+pos);
}



filter* arrayFilter::getFilter(Int pos) const {
  return(filterList[pos]);
}



array* arrayFilter::makeCloneForArrayTime(timeDep &TD) {
  try {
    arrayFilter *newArray=new arrayFilter(*this);
    newArray->setExpr(expr->makeCloneForArrayTime(TD));
    return(newArray);
  }
  catchAll("arrayFilter::makeCloneForArrayTime","")
}



array* arrayFilter::makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
		              string derivType) const {
  errThrow("arrayFilter::makeDeriv",
	   "no derivation of arrayFilter is allowed.");
  return(NULL);
}



array* arrayFilter::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  errThrow("arrayFilter::makeExplicitDeriv",
	   "no derivation of arrayFilter is allowed.");
  return(NULL);
}



array* arrayFilter::makeCoordDeriv(globLoc *GL) const {
  errThrow("arrayFilter::makeCoordDeriv",
	   "no derivation of arrayFilter is allowed.");
  return(NULL);
}



array* arrayFilter::makeCorlocDeriv(Int nbrDP) const {
  errThrow("arrayFilter::makeCorlocDeriv",
	   "no derivation of arrayFilter is allowed.");
  return(NULL);
}



void arrayFilter::setNewPoint(bool b) {
  try {
    if (expr!=NULL)
      expr->setNewPoint(b);
    newPoint::setNewPoint(b);
  }
  catchAll("arrayFilter::setNewPoint","");
}



void arrayFilter::accessNbrInit(void) {
  try {
    if (expr!=NULL)
      expr->accessNbrInit();
    for (Int i=0;i<nbrFilters;i++)
      if (filterList[i]!=NULL)
        filterList[i]->accessNbrInit();
    newPoint::accessNbrInit();
  }
  catchAll("arrayFilter::accessNbrInit","")
}



void arrayFilter::makeAtPoint(object *element,Int direct,Double *localCoord,
		              bool makeValue,bool makeDeriv) {
  try {
    Int i;
    
    if (!getNewPoint())
      return;
    newPoint::setNewPoint(false);
    
    expr->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
    for (i=0;i<nbrVal;i++)
      componentValues[i]=expr->componentValues[i];
    applyFiltersToComponents();
  }
  catchAll("arrayFilter::makeAtPoint","")
}



void arrayFilter::makeOnObject(object *element,Int direct,bool makeValue,
                               bool makeDeriv) {
  try {
    Int i;
    
    if (!getNewPoint())
      return;
    newPoint::setNewPoint(false);
    
    expr->makeOnObject(element,direct,makeValue,makeDeriv);
    for (i=0;i<nbrVal;i++)
      componentValues[i]=expr->componentValues[i];
    applyFiltersToComponents();
  }
  catchAll("arrayFilter::makeOnObject","")
}



printTree arrayFilter::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int i;
    
    sprintf(cbuf,"arrayFilter");
    sbuf=cbuf;
    sprintf(cbuf," %p",static_cast<const void*>(this));
    sbuf+=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sprintf(cbuf,"   number of filters = %d",nbrFilters);
    sbuf=cbuf;
    pt+=sbuf;

    ptArg="   ";
    pt+=ptArg;    
    
    pt.prepareForArguments();
    
    if (expr!=NULL)
      ptArg=expr->getFirstPrintTree("expression",values);
    else
      ptArg="undefined expression";
    if (nbrFilters>0)
      ptArg.prepareForAdd();
    else
      ptArg.prepareForAddLast();
    pt+=ptArg;
    
    for (i=0;i<nbrFilters;i++) {
      ptArg=filterList[i]->getFirstPrintTree("filter",false);
      if (i==(nbrFilters-1))
        ptArg.prepareForAddLast();
      else
        ptArg.prepareForAdd();
      pt+=ptArg;
    }
    
    ptArg="   ";
    pt+=ptArg;    
    
    return(pt);
  }
  catchAll("arrayFilter::getPrintTree","")
}



stepList arrayFilter::getStepList(void) {
  try {
    Int i,size;
    
    if (order!=0)
      errThrow("arrayFilter::getStepList",
	       "   the array has more than one component.");
    
    stepList SL=expr->getStepList();
    size=SL.getSize();
    for (i=0;i<size;i++)
      SL[i]->addArrayFilter(this);
    
    return(SL);
  }
  catchAll("arrayFilter::getStepList","")
}



set<field*>* arrayFilter::buildActiveFieldList(void) {
  try {
    FL=*(expr->buildActiveFieldList());
    return(&FL);
  }
  catchAll("arrayFilter::buildActiveFieldList","")
}



set<parents*> arrayFilter::getListOfType(string typeName) {
  try {
    set<parents*> S,tmpS;
    set<parents*>::iterator it;
    
    if (isOfType(typeName))
      S.insert(dynamic_cast<parents*>(this));
    tmpS=expr->getListOfType(typeName);
    for (it=tmpS.begin();it!=tmpS.end();it++)
      S.insert(*it);
    
    return(S);
  }
  catchAll("arrayFilter::getListOfType","");
}



void arrayFilter::generateFilters(void) {
  try {
    Int i;
    filter *tmpFilter;
    set<Int>::iterator sit;
    bool lineType;
    
    cleanFilters();
    for (i=0;i<nbrFilters;i++) {
      tmpFilter=filterList[i];
      tmpFilter->generateBlacks();
      lineType=tmpFilter->lineType;
      if (lineType)
	for (sit=tmpFilter->blacks.begin();sit!=tmpFilter->blacks.end();sit++)
	  lines.insert(*sit);
      else
	for (sit=tmpFilter->blacks.begin();sit!=tmpFilter->blacks.end();sit++)
	  cols.insert(*sit);
    } 
  }
  catchAll("arrayFilter::generateFilters","");
}



void arrayFilter::applyFiltersToComponents(void) {
  try {
    for (Int i=0;i<nbrVal;i++)
      applyFiltersToScal(componentValues+i);
    
    updateBoolZeroValue();
  }
  catchAll("arrayFilter::applyFiltersToComponents","");
}



void arrayFilter::applyFiltersToScal(scal *tmpScal) const {
  try {
    Int j;
    Double *doublePtr,*doublePtrEnd;
    Int ligCapacity;
    
    // elimination of lines :
    
    for (j=0;j<tmpScal->colSize;j++) {
      if (tmpScal->colBool[j]
	  &&lines.find(tmpScal->map[tmpScal->colIndex+j])!=lines.end()) {
	tmpScal->colBool[j]=false;
	tmpScal->col[j]=0.0;
	doublePtr=tmpScal->mat[j];
	doublePtrEnd=tmpScal->mat[j]+tmpScal->ligSize;
	while(doublePtr<doublePtrEnd)
	  *(doublePtr++)=0.0;
      }
    }
    
    // elimination of columns :
    
    for (j=0;j<tmpScal->ligSize;j++) {
      if (tmpScal->ligBool[j]
	  &&cols.find(tmpScal->map[tmpScal->ligIndex+j])!=cols.end()) {
	tmpScal->ligBool[j]=false;
	tmpScal->lig[j]=0.0;
	doublePtr=tmpScal->mat[0]+j;
	ligCapacity=tmpScal->ligCapacity;
	doublePtrEnd=tmpScal->mat[0]+j+tmpScal->colSize*ligCapacity;
	while(doublePtr<doublePtrEnd) {
	  *(doublePtr)=0.0;
	  doublePtr+=ligCapacity;
	}
      }
    }
    
    tmpScal->updateBoolZeroValue();
  }
  catchAll("arrayFilter::applyFiltersToScal","")
}



void arrayFilter::cleanFilters(void) {
  try {
    lines.clear();
    cols.clear();
  }
  catchAll("arrayFilter::cleanFilters","")
}



