
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <typeinfo>

#include "array_base.h"
#include "arrayInteg.h"
#include "arrayOp1.h"
#include "arrayOp2.h"
#include "arrayOpN.h"
#include "arrayFunctionsBuild.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "function/functions.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;


namespace keyFE2 {  
  
  
  
  arrayOp1* transpositionBuild(array &arg,Int index1,Int index2) {
    try {
      if (index1>arg.getOrder())
	errThrow("transpostionBuild",
		 string("   first index too large (")+strFromInt(index1)+").");
      
      if (index2>arg.getOrder())
	errThrow("transpostionBuild",
		 string("   second index too large (")+strFromInt(index2)+").");
      
      
      arrayOp1* newArray=new arrayOp1(arg);
      if (arg.getDim(index2)!=arg.getDim(index1)) {
	newArray->setDim(index1,arg.getDim(index2));
	newArray->setDim(index2,arg.getDim(index1));
      }
      newArray->setIParam1(index1);
      newArray->setIParam2(index2);
      newArray->setType("transposition");
      newArray->setArrayArg(&arg);
      
      return(newArray);
    }
    catchAll ("transpositionBuild","")
  }
  
  
  
  arrayOp1* contractionBuild(array &arg,Int index1,Int index2) {
    try {
      Int i,index;
      
      if (arg.getOrder()<2)
	errThrow("contractionBuild",
		 string("   order too small (")+strFromInt(arg.getOrder())+").");
      
      if (index1>arg.getOrder())
	errThrow("contractionBuild",
		 string("   first index too large (")+strFromInt(index1)+").");
      
      if (index2>arg.getOrder())
	errThrow("contractionBuild",
		 string("   second index too large (")+strFromInt(index2)+").");
      
      if (arg.getDim(index1)!=arg.getDim(index2))
	errThrow("contractionBuild",
		 string("   dimensions must coincide (")
		 +strFromInt(arg.getDim(index1))+"!="
		 +strFromInt(arg.getDim(index2))+").");
      
      arrayOp1* newArray=new arrayOp1(arg);
      newArray->setTCS(NULL);
      newArray->setOrder(arg.getOrder()-2);
      index=0;
      for (i=0;i<arg.getOrder();i++)
	if (i!=index1&&i!=index2)
	  newArray->setDim(index++,arg.getDim(i));
      newArray->setIParam1(index1);
      newArray->setIParam2(index2);
      newArray->setType("contraction");
      newArray->setArrayArg(&arg);
      newArray->setTCS(arg.getTCS());
      
      return(newArray);
    }
    catchAll("contractionBuild","")
  }
  
  
  
  arrayOp1* inversionBuild(array &arg)  {
    try {
      if (arg.getOrder()==2) {
	if (arg.getDim(0)!=arg.getDim(1))
	  errThrow("inversionBuild",
		   "   the array must be square for inversion.");
	arrayOp1* newArray=new arrayOp1;
	newArray->setOrder(2);
	newArray->setDim(0,arg.getDim(0));
	newArray->setDim(1,arg.getDim(1));
	newArray->setType("inversion");
	newArray->setArrayArg(&arg);
        return(newArray);
      }
      else if (arg.getOrder()==0) {
	arrayOp1* newArray=new arrayOp1;
	newArray->setOrder(0);
	newArray->setType("inversion");
	newArray->setArrayArg(&arg);
        return(newArray);
      }
      else
	errThrow("inversionBuild",
		 "   order of the array must be 0 or 2 for inversion.");
    }
    catchAll("inversionBuild","")
  }
  
  
  
  arrayOp1* functionBuild(array &arg,function1 *F) {
    try {
      arrayOp1* newArray=new arrayOp1(arg);
      newArray->setType("function");
      newArray->setFunction(F);
      newArray->setArrayArg(&arg);
      
      return(newArray);
    }
    catchAll("functionBuild","")
  }
  
  
  
  arrayOp1* indicesPermutationBuild(array &arg) {
    try {
      // Construction without permutation.
      arrayOp1 *newArray=new arrayOp1(arg);
      newArray->setTCS(NULL);
      newArray->setIListNbr(arg.getOrder());
      for (Int i=0;i<arg.getOrder();i++) {
	newArray->setDim(i,arg.getDim(i));
	newArray->setIList(i,i);
      }
      newArray->setType("indicesPermutation");
      newArray->setArrayArg(&arg);
      newArray->setTCS(arg.getTCS());
      
      return(newArray);
    }
    catchAll("indicesPermutationBuild","")
  }
  
  
  
  arrayOp1* discretizedTranspositionBuild(array &arg) {
    try {
      arrayOp1* newArray=new arrayOp1(arg);
      newArray->setType("discretizedTransposition");
      newArray->setArrayArg(&arg);
      
      return(newArray);
    }
    catchAll("discretizedTranspositionBuild","")
  }
  
  
  
  arrayOp2* vectProductBuild(array *first,array *second) {
    try {
      if ((first->getOrder()!=1)&&(second->getOrder()!=1))
	errThrow("vectProductBuild","   order of the arrays must be 1.");
      if (!((first->getDim(0)==3)&&(second->getDim(0)==3))
	      ||((first->getDim(0)==2)&&(second->getDim(0)==2)))
	errThrow("vectProductBuild","   dim of the arrays must be 2 or 3.");
      
      arrayOp2 *arrayVP=new arrayOp2;
      if (first->getDim(0)==3) {
	arrayVP->setOrder(1);
	arrayVP->setDim(0,3);
      }
      else
	arrayVP->setOrder(0);
      arrayVP->setType("vectProduct");
      arrayVP->setArrayArg1(first);
      arrayVP->setArrayArg2(second);
      
      return(arrayVP);
    }
    catchAll("vectProductBuild","")
  }
  
  
  
  arrayOp2* specialContractBuild(array *first,array *second,Int indexMax) {
    try {
      arrayOp2 *newArray=new arrayOp2;
      newArray->setOrder(first->getOrder()+second->getOrder()-indexMax);
      for (Int i=0;i<newArray->getOrder();i++)
	newArray->setDim(i,second->getDim(i));
      newArray->setArrayArg1(first);
      newArray->setArrayArg2(second);
      newArray->setType("specialContract");
      newArray->setIParam1(indexMax);
      
      return(newArray);
    }
    catchAll("specialContractBuild","")
  }
  
  
  
  arrayOp2* additionBuild(array *first,array *second) {
    try {
      Int i;
      
      if (first->getOrder()!=second->getOrder())
	errThrow("additionBuild","   order of the arrays must be identical.");
      
      for (i=0;i<first->getOrder();i++)
	if (first->getDim(i)!=second->getDim(i))
	  errThrow("additionBuild",
		   string("   ")+strFromInt(i)
		   +"th dimensions of the arrays must be identical ("
		   +strFromInt(first->getDim(i))+"!="
		   +strFromInt(second->getDim(i))+").");
      
      arrayOp2 *arraySum=new arrayOp2;
      arraySum->setOrder(first->getOrder());
      for (i=0;i<first->getOrder();i++)
	arraySum->setDim(i,first->getDim(i));
      arraySum->setType("addition");
      arraySum->setArrayArg1(first);
      arraySum->setArrayArg2(second);
      
      return(arraySum);
    }
    catchAll("additionBuild","")
  }
  
  
  
  arrayOp2* soustractionBuild(array *first,array *second) {
    try {
      Int i;
      
      if (first->getOrder()!=second->getOrder())
	errThrow("soustractionBuild",
		 "   order of the arrays must be identical.");
      
      for (i=0;i<first->getOrder();i++)
	if (first->getDim(i)!=second->getDim(i))
	  errThrow("soustractionBuild",
		   string("   ")+strFromInt(i)
		   +"th dimensions of the arrays must be identical ("
		   +strFromInt(first->getDim(i))+"!="
		   +strFromInt(second->getDim(i))+").");
      
      arrayOp2 *arrayDiff=new arrayOp2;
      arrayDiff->setOrder(first->getOrder());
      for (i=0;i<first->getOrder();i++)
	arrayDiff->setDim(i,first->getDim(i));
      arrayDiff->setType("soustraction");
      arrayDiff->setArrayArg1(first);
      arrayDiff->setArrayArg2(second);
      
      return(arrayDiff);
    }
    catchAll("soustractionBuild","")
  }
  
  
  
  arrayOp2* exteriorMultiplicationBuild(array *first,array *second) {
    try {
      Int i;
      
      arrayOp2 *arrayMult=new arrayOp2;
      arrayMult->setOrder(first->getOrder()+second->getOrder());
      for (i=0;i<first->getOrder();i++)
	arrayMult->setDim(i,first->getDim(i));
      for (i=0;i<second->getOrder();i++)
	arrayMult->setDim(first->getOrder()+i,second->getDim(i));
      arrayMult->setType("exteriorMultiplication");
      arrayMult->setArrayArg1(first);
      arrayMult->setArrayArg2(second);
      
      return(arrayMult);
    }
    catchAll("exteriorMultiplicationBuild","")
  }
  
  
  
  arrayOp2* contractedMultiplicationBuild(array *first,Int index1,array *second,
					  Int index2) {
    try {
      Int i,index;
      
      arrayOp2 *arrayMult=new arrayOp2;
      arrayMult->setOrder(first->getOrder()+second->getOrder()-2);
      index=0;
      for (i=0;i<first->getOrder();i++)
	if (i!=index1)
	  arrayMult->setDim(index++,first->getDim(i));
      for (i=0;i<second->getOrder();i++)
	if (i!=index2)
	  arrayMult->setDim(index++,second->getDim(i));
      arrayMult->setType("contractedMultiplication");
      arrayMult->setIParam1(index1);
      arrayMult->setIParam2(index2);
      arrayMult->setArrayArg1(first);
      arrayMult->setArrayArg2(second);
      
      return(arrayMult);
    }
    catchAll("contractedMultiplicationBuild","")
  }
  
  
  
  arrayOp2* accumulationBuild(array *first,array *second,Int nbrC) {
    try {
      Int i,index;
      
      if (first->getOrder()<nbrC)
	errThrow("accumulationBuild",
		 string("   first operand of accumulation operator should have at least ")
		 +strFromInt(nbrC)+" for order.");
      
      if (second->getOrder()<nbrC)
	errThrow("accumulationBuild",
		 string("   second operand of accumulation operator should have at least ")
		 +strFromInt(nbrC)+" for order.");
      
      arrayOp2 *arrayMult=new arrayOp2;
      arrayMult->setOrder(first->getOrder()+second->getOrder()-2*nbrC);
      index=0;
      for (i=nbrC;i<first->getOrder();i++)
	arrayMult->setDim(index++,first->getDim(i));
      for (i=nbrC;i<second->getOrder();i++)
	arrayMult->setDim(index++,second->getDim(i));
      arrayMult->setType("accumulation");
      arrayMult->setIParam1(nbrC);
      arrayMult->setArrayArg1(first);
      arrayMult->setArrayArg2(second);
      
      return(arrayMult);
    }
    catchAll("accumulationBuild","")
  }
  
  
  
  arrayOp2* substitutionBuild(array *first,array *second,Int index1,Int index2) {
    try {
      Int i;
      
      if (second->getOrder()!=2)
	errThrow("substitutionBuild","   order of the second array must be 2.");
      
      if (second->getDim(0)!=second->getDim(1))
	errThrow("substitutionBuild","   the second array must be square.");
      
      if (second->getDim(0)!=first->getDim(index1))
	errThrow("substitutionBuild",
		 "   invalid dimension in index substitution.");
      
      if (index2<0||index2>1)
	errThrow("substitutionBuild","   second parameter must be 0 or 1.");    
      
      arrayOp2 *newArray=new arrayOp2;
      newArray->setOrder(first->getOrder());
      for (i=0;i<first->getOrder();i++)
	newArray->setDim(i,first->getDim(i));
      newArray->setType("substitution");
      newArray->setIParam1(index1);
      newArray->setIParam2(index2);
      newArray->setArrayArg1(first);
      newArray->setArrayArg2(second);
      
      return(newArray);
    }
    catchAll("substitutionBuild","")
  }
  
  
  
  bool specialCombiliNeeded(array *arg1,array *arg2) {
    try {
      // One checks the first argument :
      if (typeid(*arg1)==typeid(arrayInteg))
	return(true);
      if (typeid(*arg1)==typeid(arrayOp2)) {
	arrayOp2 *tmp=dynamic_cast<arrayOp2*>(arg1);
	if (tmp->getTypeName()=="addition")
	  return(true);
	else if (tmp->getTypeName()=="soustraction")
	  return(true);
      }
      if (typeid(*arg1)==typeid(arrayOpN)) {
	arrayOpN *tmp=dynamic_cast<arrayOpN*>(arg1);
	if (tmp->getTypeName()=="addition")
	  return(true);
	else if (tmp->getTypeName()=="combili")
	  return(true);
      }
      
      // One checks the second argument :
      if (typeid(*arg2)==typeid(arrayInteg))
	return(true);
      if (typeid(*arg2)==typeid(arrayOp2)) {
	arrayOp2 *tmp=dynamic_cast<arrayOp2*>(arg2);
	if (tmp->getTypeName()=="addition")
	  return(true);
	else if (tmp->getTypeName()=="soustraction")
	  return(true);
      }
      if (typeid(*arg2)==typeid(arrayOpN)) {
	arrayOpN *tmp=dynamic_cast<arrayOpN*>(arg2);
	if (tmp->getTypeName()=="addition")
	  return(true);
	else if (tmp->getTypeName()=="combili")
	  return(true);
      }
      
      // Default :
      return(false); 
    }
    catchAll("specialCombiliNeeded","")
  }
  
  
  
  arrayOpN* specialCombiliBuild(Double fact1,array *arg1,Double fact2,
				array *arg2) {
    try {
      Int nbrArgs1=0,nbrArgs2=0,nbrArgs=0;
      Int index;
      arrayOpN *newArray;
      Int i,k;
      
      if (arg1==NULL)
        errThrow("specialCombiliBuild","   arg1 is NULL.");
      if (arg2==NULL)
        errThrow("specialCombiliBuild","   arg2 is NULL.");
      
      // One calculates the number of arguments of the arrayOpN :
      
      for (k=0;k<2;k++) {
	array *argx;
	Int nbrArgsx;
	
	if (k==0)
	  argx=arg1;
	else
	  argx=arg2;
	
	if (typeid(*argx)==typeid(arrayOp2)) {
	  arrayOp2 *tmp=dynamic_cast<arrayOp2*>(argx);
	  if (tmp->getTypeName()=="addition")
	    nbrArgsx=2;
	  else if (tmp->getTypeName()=="soustraction")
	    nbrArgsx=2;
	  else
	    nbrArgsx=1;
	}
	else if (typeid(*argx)==typeid(arrayOpN)) {
	  arrayOpN *tmp=dynamic_cast<arrayOpN*>(argx);
	  if (tmp->getTypeName()=="addition")
	    nbrArgsx=tmp->getArrayArgsNbr();
	  else if (tmp->getTypeName()=="combili")
	    nbrArgsx=tmp->getArrayArgsNbr();
	  else
	    nbrArgsx=1;
	}
	else
	  nbrArgsx=1;
	
	if (k==0)
	  nbrArgs1=nbrArgsx;
	else
	  nbrArgs2=nbrArgsx;
      }
      
      nbrArgs=nbrArgs1+nbrArgs2;
      
      // One builds the new arrayOpN :
      
      newArray=new arrayOpN(*arg1);
      newArray->setArrayArgsNbr(nbrArgs);
      newArray->setDParamsNbr(nbrArgs);
      newArray->setType("combili");
      
      index=0;
      
      for (k=0;k<2;k++) {
	array *argx;
	Int nbrArgsx;
	Double factx;
	
	if (k==0) {
	  argx=arg1;
	  nbrArgsx=nbrArgs1;
	  factx=fact1;
	}
	else {
	  argx=arg2;
	  nbrArgsx=nbrArgs2;
	  factx=fact2;
	}
	
	if (typeid(*argx)==typeid(arrayOp2)) {
	  arrayOp2 *tmp=dynamic_cast<arrayOp2*>(argx);
	  if (tmp->getTypeName()=="addition") {
	    newArray->setArrayArg(index,tmp->getArrayArg1());
	    newArray->setDParam(index,factx);
	    index++;
	    newArray->setArrayArg(index,tmp->getArrayArg2());
	    newArray->setDParam(index,factx);
	    index++;
	  }
	  else if (tmp->getTypeName()=="soustraction") {
	    newArray->setArrayArg(index,tmp->getArrayArg1());
	    newArray->setDParam(index,factx);
	    index++;
	    newArray->setArrayArg(index,tmp->getArrayArg2());
	    newArray->setDParam(index,-factx);
	    index++;
	  }
	  else {
	    newArray->setArrayArg(index,argx);
	    newArray->setDParam(index,factx);
	    index++;
	  }
	}
	else if (typeid(*argx)==typeid(arrayOpN)) {
	  arrayOpN *tmp=dynamic_cast<arrayOpN*>(argx);
	  if (tmp->getTypeName()=="addition") {
	    for (i=0;i<tmp->getArrayArgsNbr();i++) {
	      newArray->setArrayArg(index,tmp->getArrayArg(i));
	      newArray->setDParam(index,factx);
	      index++;
	    }
	  }
	  else if (tmp->getTypeName()=="combili") {
	    for (i=0;i<tmp->getArrayArgsNbr();i++) {
	      newArray->setArrayArg(index,tmp->getArrayArg(i));
	      newArray->setDParam(index,factx*tmp->getDParam(i));
	      index++;
	    }
	  }
	  else {
	    newArray->setArrayArg(index,argx);
	    newArray->setDParam(index,factx);
	    index++;
	  }
	}
	else {
	  newArray->setArrayArg(index,argx);
	  newArray->setDParam(index,factx);
	  index++;
	}
      }
      
      return(newArray);
    }
    catchAll("specialCombiliBuild","")
  }
  
  
  
  arrayOpN* upwindF(array *V,array *rho,array *k,array *testU,array *dXdt,
                    globLoc *gl,Double dInterp) {
    try {
      if (V==NULL)
        errThrow("upwindF","V is NULL.");
      if (rho==NULL)
        errThrow("upwindF","rho is NULL.");
      if (k==NULL)
        errThrow("upwindF","k is NULL.");
      if (testU==NULL)
        errThrow("upwindF","testU is NULL.");
      if (gl==NULL)
        errThrow("upwindF","gl is NULL.");
	
      if (V->getOrder()!=1)
        errThrow("upwindF","order of V must be 1.");
      if (rho->getOrder()!=0)
        errThrow("upwindF","order of rho must be 0.");
      if (k->getOrder()!=0)
        errThrow("upwindF","order of k must be 0.");
      if (dXdt!=NULL)
        if (dXdt->getOrder()!=1)
          errThrow("upwindF","order of dXdt must be 1.");
	
      Int globDim=gl->getGlobDim();
      
      if (V->getDim(0)!=globDim)
        errThrow("upwindF","dim of V must be "+strFromInt(globDim)+".");
      if (dXdt!=NULL)
        if (dXdt->getDim(0)!=globDim)
          errThrow("upwindF","dim of dXdt must be "+strFromInt(globDim)+".");
	  
      if (dInterp<0.0||dInterp>10.0)
        errThrow("upwindF","dInterp must be between 0.0 or 10.0.");
	
      array *interp=new array;
      interp->component()=dInterp;
      interp->setName("interp");
	
      array *corlocDerivTestU=testU->makeCorlocDeriv(gl->getCoord()->getDim(0));
      corlocDerivTestU->setName("corlocDerivTestU");
      
      array *A;
      if (dXdt!=NULL) {
        A=dynamic_cast<array*>(soustractionBuild(V,dXdt));
	A->setName("A");
      }
      else
        A=V;
      
      array *aScal=NULL;
      {
         array *aSquare=dynamic_cast<array*>(contractedMultiplicationBuild(A,0,A,0));
	 aSquare->setName("aSquare");  
         function1 *f=new DPowMult(1.0,0.5);
	 aScal=functionBuild(*aSquare,f);
	 aScal->setName("aScal");
      }
      
      array *direct=new array(*A);
      direct->setName("direct");

      array *invJacDirect=NULL;
      array *invJacDirectNorm=NULL; 
      {
         array *invJac=gl->getInvJac();
         invJacDirect=dynamic_cast<array*>(
	    contractedMultiplicationBuild(invJac,1,direct,0));
	 invJacDirect->setName("invJacDirect");
         array *twoByH_Square=dynamic_cast<array*>(
	    contractedMultiplicationBuild(invJacDirect,0,invJacDirect,0));
	 twoByH_Square->setName("twoByH_Square");
         function1 *f=new DPowMult(1.0,0.5);
	 invJacDirectNorm=functionBuild(*twoByH_Square,f);
	 invJacDirectNorm->setName("invJacDirectNorm");
      }
      
      arrayOpN *newArray=new arrayOpN(*testU);
      newArray->setArrayArgsNbr(12);
      newArray->setArrayArg( 0,V);
      newArray->setArrayArg( 1,rho);
      newArray->setArrayArg( 2,k);
      newArray->setArrayArg( 3,testU);
      newArray->setArrayArg( 4,dXdt);
      newArray->setArrayArg( 5,interp);
      newArray->setArrayArg( 6,A);
      newArray->setArrayArg( 7,aScal);
      newArray->setArrayArg( 8,direct);
      newArray->setArrayArg( 9,invJacDirectNorm);
      newArray->setArrayArg(10,corlocDerivTestU);
      newArray->setArrayArg(11,invJacDirect);
      newArray->setType("upwind");
      
      return newArray;
    }
    catchAll("upwindF","")
  }

}
