
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "arrayInteg.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "integ/integ_base.h"
#include "object/object_base.h"
#include "print/printTree.h"
#include "scal/scal.h"
#include "step/step.h"
#include "step/stepList.h"
#include "timeDep/timeDep_base.h"

using namespace std ;
using namespace keyFE2 ;



arrayInteg::arrayInteg(void) : array() {
  expr=NULL;
  dom=NULL;
  integRule=NULL;
}



arrayInteg::arrayInteg(array *a,domain *d,integ *rule) : array(*a) {
  expr=NULL;
  dom=NULL;
  integRule=0;
  setExpression(a);
  setDomain(d);
  setIntegRule(rule);
}



arrayInteg::arrayInteg(const arrayInteg &arrayIn) : array(arrayIn) {
  expr=NULL;
  dom=NULL;
  integRule=0;
  setExpression(arrayIn.getExpression());
  setDomain(arrayIn.getDomain());
  setIntegRule(arrayIn.getIntegRule());
}



arrayInteg::~arrayInteg(void) {
  setExpression(NULL);
  setDomain(NULL);
  setIntegRule(NULL);
}



array* arrayInteg::makeCloneForArrayTime(timeDep &TD) {
  try {
    arrayInteg *newArray=new arrayInteg(*this);
    newArray->setExpression(expr->makeCloneForArrayTime(TD));
    return(newArray);
  }
  catchAll("arrayInteg::makeCloneForArrayTime","")
}



void arrayInteg::setExpression(array *e) {
  try {
    delinkAndDelete(expr,&expr);
    expr=NULL;
    
    if (e!=NULL)
      if (e->getOrder()!=0)
	errThrow("arrayInteg::setExpression",
		 "   order of integrated expression is to be 0.");
    
    expr=e;
    if (expr!=NULL)
      expr->addParent(&expr);
  }
  catchAll("arrayInteg::setExpression","")
}



array* arrayInteg::getExpression(void) const {
  return(expr);
}



void arrayInteg::setDomain(domain *d) {
  try {
    delinkAndDelete(dom,&dom);
    dom=d;
    if (dom!=NULL)
      dom->addParent(&dom);
  }
  catchAll("arrayInteg::setDomain","")
}



domain* arrayInteg::getDomain(void) const {
  return(dom);
}



void arrayInteg::setIntegRule(integ *rule) {
  try {
    delinkAndDelete(integRule,&integRule);
    integRule=rule;
    if (integRule!=NULL)
      integRule->addParent(&integRule);
  }
  catchAll("arrayInteg::setIntegRule","")
}



integ* arrayInteg::getIntegRule(void) const {
  return(integRule);
}



void arrayInteg::makeAtPoint(object *element,Int direct,Double *localCoord,
			     bool makeValue,bool makeDeriv) {
  errThrow("arrayInteg::makeAtPoint",
	   "   the function cannot be called for arrayInteg object.");
}



void arrayInteg::makeOnObject(object *element,Int direct,bool makeValue,
                              bool makeDeriv) {
  try {
    Double **integPoints,*integWeights;
    Int nbrPoints=0,nbrCoord=0;
    Int i;
    static scal factor;
    
    integRule->getIntegPoints(element->getType(),nbrPoints,nbrCoord,integPoints,
			      integWeights);
    
    if (scal::getWhite())
      nbrPoints=1;
    
    componentValues[0]=0.0;
    for (i=0;i<nbrPoints;i++) {
      expr->setNewPoint(true);
      integRule->setNewPoint(true);
      expr->makeAtPoint(element,direct,integPoints[i],makeValue,makeDeriv);
      integRule->makeAtPoint(element,direct,integPoints[i],makeValue,makeDeriv);    
      factor=integRule->getIntegFactor(element,direct,i,nbrPoints,nbrCoord,
				       integPoints,integWeights,makeValue,
				       makeDeriv);
      componentValues[0]+=expr->componentValues[0]*factor;
    }
  }
  catchAll("arrayInteg::makeOnObject","")
}



array* arrayInteg::makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
			     string derivType) const {
  errThrow("arrayInteg::makeDeriv",
	   "   the function cannot be called for arrayInteg object.");
}



array* arrayInteg::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  errThrow("arrayInteg::makeExplicitDeriv",
	   "   the function cannot be called for arrayInteg object.");
}



array* arrayInteg::makeCoordDeriv(globLoc *GL) const {
  errThrow("arrayInteg::makeCoordDeriv",
	   "   the function cannot be called for arrayInteg object.");
}



array* arrayInteg::makeCorlocDeriv(Int nbrDP) const {
  errThrow("arrayInteg::makeCorlocDeriv",
	   "   the function cannot be called for arrayInteg object.");
}



void arrayInteg::accessNbrInit(void) {
  try {
    newPoint::accessNbrInit();
    if (expr!=NULL)
      expr->accessNbrInit();
  }
  catchAll("arrayInteg::accessNbrInit","")
}



printTree arrayInteg::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    
    sprintf(cbuf,"arrayInteg %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    if (expr!=NULL) {
      ptArg=expr->getFirstPrintTree("expression",values);
      ptArg.prepareForAdd();
    }
    else {
      sbuf="undefined expression";
      ptArg=sbuf;
      sbuf="   ";
      ptArg+="sbuf";
    }
    pt+=ptArg;
    
    if (dom!=NULL) {
      ptArg=dom->getFirstPrintTree("domain",values);
      ptArg.prepareForAdd();
    }
    else {
      sbuf="undefined domain";
      ptArg=sbuf;
      sbuf="   ";
      ptArg+="sbuf";
    }
    pt+=ptArg;
    
    if (integRule!=NULL) {
      ptArg=integRule->getFirstPrintTree("integRule",values);
      ptArg.prepareForAddLast();
    }
    else {
      sbuf="undefined integRule";
      ptArg=sbuf;
      sbuf="   ";
      ptArg+="sbuf";
    }
    pt+=ptArg;
    
    return(pt);
  }
  catchAll("arrayInteg::getPrintTree","")
}



stepList arrayInteg::getStepList(void) {
  try {
    if (order!=0)
      errThrow("arrayInteg::getStepList",
	       "   the array has more than one component.");
    
    static stepList SL;
    SL.init();
    for (Int i=0;i<dom->getObjectsNbr();i++) {
      step tmp(this,dom->getObject(i),dom->getObjectDirect(i));
      SL.pushBack(tmp);
    }
    return(SL);
  }
  catchAll("arrayInteg::getStepList","")
}



set<field*>* arrayInteg::buildActiveFieldList(void) {
  try {
    FL=*(expr->buildActiveFieldList());
    return(&FL);
  }
  catchAll("arrayInteg::buildActiveFieldList","")
}



set<parents*> arrayInteg::getListOfType(string typeName) {
  try {
    set<parents*> S,tmpS;
    set<parents*>::iterator it;
    
    if (isOfType(typeName))
      S.insert(dynamic_cast<parents*>(this));
    tmpS=expr->getListOfType(typeName);
    for (it=tmpS.begin();it!=tmpS.end();it++)
      S.insert(*it);
    
    return(S);
  }
  catchAll("arrayInteg::getListOfType","")
}



