
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iomanip>
#include <iostream>
#include <list>
#include <set>
#include <stdio.h>
#include <string>

#include "constr1.h"
#include "array/array_base.h"
#include "array/arrayField.h"
#include "array/arrayFunctionsBuild.h"
#include "array/arrayInteg.h"
#include "err/err.h"
#include "expr/expression.h"
#include "field/field_base.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "domain/domain_base.h"
#include "integ/integ_base.h"
#include "object/object_base.h"
#include "param/param.h"
#include "param/paramList.h"
#include "print/printTree.h"
#include "step/step.h"
#include "util/parents.h"

using namespace std ;
using namespace keyFE2 ;



constr1::constr1(void) : constr() {
  test=NULL;
  equation=NULL;
  expr=NULL;
  dom=NULL;
  GL=NULL;
  TCS=NULL;
  integRule=-1;
}



constr1::constr1(paramList PL) : constr() {
  array *tmpTest=NULL;
  array *tmpEquation=NULL;
  domain *tmpDomain=NULL;
  globLoc *tmpGL=NULL;
  string tmpName;
  tensorCoordSys *tmpTCS=NULL;
  Int tmpLayer;
  
  try {
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plLayer=PL.getListOfType(LAYER);
    PL=PL.getListNotOfType(LAYER);
    paramList plDomain=PL.getListOfType(DOM);
    PL=PL.getListNotOfType(DOM);
    paramList plTest=PL.getListOfType(TEST);
    PL=PL.getListNotOfType(TEST);
    paramList plEquation=PL.getListOfType(EQUATION);
    PL=PL.getListNotOfType(EQUATION);
    paramList plGL=PL.getListOfType(GLOBLOC);
    PL=PL.getListNotOfType(GLOBLOC);
    paramList plTCS=PL.getListOfType(TENSORCOORDSYS);
    PL=PL.getListNotOfType(TENSORCOORDSYS);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString;
      tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++)
	tmpString=tmpString+" "+PL[i].getName();
      tmpString=tmpString+".";
      errWrite("constr1 construction",tmpString);
    }
    
    if (plName.getNbrParams()>1) {
      errWrite("constr1 construction","maximum one NAME argument is allowed.");
    }
    else if (plName.getNbrParams()==1) {
      tmpName=plName[0].getStringData();
      if (tmpName=="")
	errWrite("constr1 construction",
		 "NAME= has to be followed by a string argument.");
    }
    
    tmpLayer=0;
    if (plLayer.getNbrParams()>1) {
      errWrite("constr1 construction","maximum one LAYER argument is allowed.");
    }
    else if (plLayer.getNbrParams()==1) {
      if(plLayer[0].getTypeName()!="Int") {
	errWrite("constr1 construction",
		 "NAME= has to be followed by an Int argument.");
      }
      else
        tmpLayer=plLayer[0].getIntData();
    }
    
    if (plTest.getNbrParams()!=1) {
      errWrite("constr1 construction",
	       "exactly one TEST parameter is needed.");
    }
    else {
      parents *tmpParents=plTest[0].getParentData();
      tmpTest=dynamic_cast<array*>(tmpParents);
      if (tmpTest==NULL)
	errWrite("constr1 construction",
		 "TEST= must be followed by an Array.");
    }
    
    if (plEquation.getNbrParams()!=1) {
      errWrite("constr1 construction",
	       "exactly one EQUATION parameter is needed.");
    }
    else {
      parents *tmpParents=plEquation[0].getParentData();
      tmpEquation=dynamic_cast<array*>(tmpParents);
      if (tmpEquation==NULL)
	errWrite("constr1 construction",
		 "EQUATION= must be followed by an Array.");
    }
    
    if (plDomain.getNbrParams()!=1) {
      errWrite("constr1 construction",
	       "exactly one DOM parameter is needed.");
    }
    else {
      parents *tmpParents=plDomain[0].getParentData();
      tmpDomain=dynamic_cast<domain*>(tmpParents);
      if (tmpDomain==NULL)
	errWrite("constr1 construction",
		 "DOM= must be followed by a Domain.");
    }
    
    if (plGL.getNbrParams()>1) {
      errWrite("constr1 construction",
	       "maximum one GLOBLOC parameter is needed.");
    }
    else if (plGL.getNbrParams()==1) {
      parents *tmpParents=plGL[0].getParentData();
      tmpGL=dynamic_cast<globLoc*>(tmpParents);
      if (tmpGL==NULL)
	errWrite("constr1 construction",
		 "GLOBLOC= must be followed by a GlobLoc.");
    }
    
    if (plTCS.getNbrParams()>1) {
      errWrite("constr1 construction",
	       "maximum one TENSORCOORDSYS parameter is needed.");
    }
    else if (plTCS.getNbrParams()==1) {
      parents *tmpParents=plTCS[0].getParentData();
      tmpTCS=dynamic_cast<tensorCoordSys*>(tmpParents);
      if (tmpTCS==NULL)
	errWrite("constr1 construction",
		 "TENSORCOORDSYS= must be followed by a TensorCoordSys.");
    }
    
    if (!err::getOk())
      errThrow("constr1 construction","bailing out due to preceding errors.");
    
    test=NULL;
    equation=NULL;
    expr=NULL;
    dom=NULL;
    GL=NULL;
    TCS=NULL;

    setName(tmpName);
    setLayer(tmpLayer);
    setDomain(tmpDomain);
    setTest(tmpTest);
    setEquation(tmpEquation);
    setGlobLoc(tmpGL);
    setTCS(tmpTCS);
  }
  catchAll("constr1::constr1(paramList PL)","at the very end of constructor")
}



constr1::constr1(const constr1 &in) : constr(in) {
  try {
    test=NULL;
    setTest(in.test);
    equation=NULL;
    setEquation(in.equation);
    expr=NULL;
    setExpr(in.expr);
    dom=NULL;
    setDomain(in.dom);
    GL=NULL;
    setGlobLoc(in.GL);
    TCS=NULL;
    setTCS(in.TCS);
    integRule=in.integRule;
  }
  catchAll("constr1::constr1(const constr1 &in)","")
}



constr1::~constr1(void) {
  try {
    setTest(NULL);
    setEquation(NULL);
    setExpr(NULL);
    setDomain(NULL);
    setGlobLoc(NULL);
    setTCS(NULL);
    integRule=-1;
  }
  catchAll("constr1::~constr1","")
}



void constr1::operator=(const constr1 &in) {
  if (this==&in)
    return;
  
  try {
    constr::operator=(in);
    setTest(in.test);
    setEquation(in.equation);
    setExpr(in.expr);
    setDomain(in.dom);
    setGlobLoc(in.GL);
    setTCS(in.TCS);
    integRule=in.integRule;
  }
  catchAll("constr1::operator=","")
}



void constr1::setTest(array *arg) {
  try {
    delinkAndDelete(test,&test);
    test=arg;
    if (test!=NULL)
      test->addParent(&test);  
    checkTestEquationMatch();
  }
  catchAll("constr1::setTest","")
}



void constr1::setEquation(array *arg) {
  try {
    delinkAndDelete(equation,&equation);
    equation=arg;
    if (equation!=NULL)
      equation->addParent(&equation);
    checkTestEquationMatch();
  }
  catchAll("constr1::setEquation","")
}



void constr1::setExpr(array *arg) {
  try {
    delinkAndDelete(expr,&expr);
    expr=arg;
    if (expr!=NULL)
      expr->addParent(&expr);
  }
  catchAll("constr1::setExpr","")
}



void constr1::setDomain(domain *arg) {
  try {
    delinkAndDelete(dom,&dom);
    dom=arg;
    if (dom!=NULL)
      dom->addParent(&dom);
    checkTestEquationMatch();
  }
  catchAll("constr1::setDomain","")
}



void constr1::setGlobLoc(globLoc *arg) {
  try {
    delinkAndDelete(GL,&GL);
    GL=arg;
    if (GL!=NULL)
      GL->addParent(&GL);
    checkTestEquationMatch();
  }
  catchAll("constr1::setGlobLoc","")
}



void constr1::setTCS(tensorCoordSys *arg) {
  try {
    delinkAndDelete(TCS,&TCS);
    TCS=arg;
    if (TCS!=NULL)
      TCS->addParent(&TCS);
    checkTestEquationMatch();
  }
  catchAll("constr1::setTCS","")
}



void constr1::checkTestEquationMatch(void) {
  if (test==NULL||equation==NULL)
    return;
  
  try {
    if (test->getOrder()!=equation->getOrder())
      errThrow("constr1::checkTestEquationMatch",
	       string("order mismatch :  ")+strFromInt(test->getOrder())
	       +"!="+strFromInt(equation->getOrder())+".");
    
    for (Int i=0;i<test->getOrder();i++)
      if (test->getDim(i)!=equation->getDim(i)) 
	errWrite("constr1::checkTestEquationMatch",
		 string("dimension mismatch for index ")+strFromInt(i)
		 +" :  "+strFromInt(test->getDim(i))+"!="
		 +strFromInt(equation->getDim(i))+".");
    
    if (!err::getOk())
      errThrow("constr1::checkTestEquationMatch",
	       "bailing out due to preceding errors.");
    
    if (GL!=NULL&&dom!=NULL)
      if (GL->getLocDim()!=dom->getDim()) {
        cerr << "   GlobLoc local dim = " << GL->getLocDim() << endl ;
        cerr << "   Domain dim = " << dom->getDim() << endl ;
	errThrow("constr1::checkTestEquationMatch",
		 "domain dimension incompatible with globLoc argument.");
      }

    test->buildActiveFieldList();
    set<field*> *tmpSF=test->getFieldList();
    if (tmpSF->size()!=1)
      errThrow("constr1::checkTestEquationMatch",
	       "TEST expression should depend on exactly one field");
    field *F=*(tmpSF->begin());
    string ITN=F->getInterpolationTypeName();
    if (ITN=="constantOnDomain"||ITN=="constantPerElement")
      integRule=100;
    else if (ITN=="linear")
      integRule=101;
    else if (ITN=="quadratic")
      integRule=102;
    else
      integRule=-1;
    
    array *tmpArray=accumulationBuild(test,equation,test->getOrder());
    setExpr(tmpArray);
  }
  catchAll("constr1::checkTestEquationMatch","")
}



void constr1::generateExprCollect(bool makeValue,bool makeDeriv) {
  Int i,j,k,l;
  Int ligSize,colSize,ligIndex,colIndex,*map,counter;
  step currentStep;
  Double *col,*lig,**mat;
  Double tmpDouble1,tmpDouble2,factor;
  static scal scalFactor,tmpScal;
  
  object *element;
  Int elementDirect;
  
  integ ITG(GL,TCS,integRule);
  Double **integPoints,*integWeights;
  Int nbrPoints=0,nbrCoord=0;
  
  list<expression> EL1;
  list<expression>::iterator ELit;
  static expression tmpExpr,tmpExpr1,tmpExpr2;
  
  scal *CV;
  
  try {
    expr->buildActiveFieldList();
    CV=expr->getComponentValues();
    currentStep.setEquation(expr);
    for (i=0;i<dom->getObjectsNbr();i++) {
      element=dom->getObject(i);
      elementDirect=dom->getObjectDirect(i);
      ITG.getIntegPoints(element->getType(),nbrPoints,nbrCoord,integPoints,
			 integWeights);
      currentStep.setElement(element);
      currentStep.setDirection(elementDirect);
      for (l=0;l<nbrPoints;l++) {
	// One calculates entities at a point :
	currentStep.exec(integPoints[l]);
	ITG.setNewPoint(true);
	ITG.makeAtPoint(element,elementDirect,integPoints[l],makeValue,
			makeDeriv);    
	scalFactor=ITG.getIntegFactor(element,elementDirect,l,nbrPoints,
				      nbrCoord,integPoints,integWeights,
				      makeValue,makeDeriv);
	factor=scalFactor.getVal();
	tmpScal=(*CV);
	// One analyzes the tmpScal :
	ligSize=tmpScal.getLigSize();
	ligIndex=tmpScal.getLigIndex();
	colSize=tmpScal.getColSize();
	colIndex=tmpScal.getColIndex();
	col=tmpScal.getCol();
	mat=tmpScal.getMat();
	map=tmpScal.getMap();
	for (j=0;j<colSize;j++) {
	  lig=mat[j];
	  counter=0;
	  tmpDouble1=0.0;
	  tmpDouble2=0.0;
	  for (k=0;k<ligSize;k++) {
	    if (lig[k]!=0.0)
	      counter++;
	    if (map[ligIndex+k]==map[colIndex+j]) {
	      tmpDouble1=lig[k];
	      if (abs(tmpDouble1)>tmpExpr.zero)
		tmpDouble2=1.0/tmpDouble1;
	    }
	  } // for k ...
	  if (abs(tmpDouble1)>tmpExpr.zero) {
	    tmpExpr.clear();
	    tmpExpr.setCapacity(max(counter-1,tmpExpr.getCapacity()));
	    tmpExpr.index=map[colIndex+j];
	    tmpExpr.weight=factor;
	    tmpExpr.independentValue=-col[j]*tmpDouble2;
	    for (k=0;k<ligSize;k++)
	      if (lig[k]!=0.0&&map[colIndex+j]!=map[ligIndex+k])
		tmpExpr.addTerm(map[ligIndex+k],-lig[k]*tmpDouble2);
	    EL1.push_back(tmpExpr);
	  } // if (...)
	} // for j ...
      } // for l ...
    } // for i ...
    // One sorts the expressions :
    
    EL1.sort();
    ELit=EL1.begin();
    tmpExpr=*ELit;
    ELit++;
    EC.setCapacity(EL1.size());
    for (;ELit!=EL1.end();ELit++) {
      tmpExpr1=*ELit;
      if (tmpExpr1.index!=tmpExpr.index) {
	EC.pushBack(tmpExpr);
	tmpExpr=tmpExpr1;
      }
      else {
	tmpExpr2=tmpExpr;
	tmpExpr.merge(tmpExpr1,tmpExpr2);
      }
    }
    tmpExpr.simplify();
    EC.pushBack(tmpExpr);
    
    EC.setLayer(layer);
    
    if (EC.getSize()==0)
      errThrow("constr1::generateExprCollect",
	       string("no constraint generated for constr1 named \"")+getName()+"\".");
  }
  catchAll("constr1::generateExprCollect","at the end of function")
}



printTree constr1::getPrintTree(bool values) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt,ptArg;
    
  try {
    sprintf(cbuf,"constr1");
    sbuf=cbuf;
    sprintf(cbuf," %p",static_cast<const void*>(this));
    sbuf+=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    ptArg=test->getFirstPrintTree("test",values);
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    ptArg=equation->getFirstPrintTree("equation",values);
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    ptArg=dom->getFirstPrintTree("dom",values);
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    ptArg=GL->getFirstPrintTree("GL",values);
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    sprintf(cbuf,"integRule = %d",integRule);
    ptArg=cbuf;
    ptArg.prepareForAdd();
    pt+=ptArg;
    pt+="   ";
    
    return(pt);
  }
  catchAll("constr1::getPrintTree","")
}

