
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <typeinfo>
#include <stdio.h>
#include <string>

#include "globLoc_base.h"
#include "array/array_base.h"
#include "array/arrayComp.h"
#include "array/arrayFunctionsBuild.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "function/functions.h"
#include "print/printTree.h"

using namespace std ;
using namespace keyFE2 ;



void globLoc::init(void) {
  try {
    type=-1;
    coord=NULL;
    coordDeriv=NULL;
    Jac=NULL;
    invJac=NULL;
    detJac=0.0;
    vectors=NULL;
    setVectors();
  }
  catchAll("globLoc::init","at the very end")
}



globLoc::globLoc(void) {
  try {
    init();
  }
  catchAll("globLoc::globLoc(void)","at the very end")
}



globLoc::globLoc(string str) {
  try {
    init();
    setName(str);
  }
  catchAll("globLoc::globLoc(string str)","at the very end")
}



globLoc::~globLoc(void) {
  try {
    type=-1;
    setCoord(NULL);
    setCoordDeriv(NULL);
    setJac(NULL);
    setInvJac(NULL);
    detJac=0.0;
    setVectors(NULL);
  }
  catchAll("globLoc::~globLoc(void)","at the very end")
}



void globLoc::setType(Int nbr) {
  type=nbr;
}



void globLoc::setType(string typeName) {
  if (typeName=="none")
    type=-1;
  else if (typeName=="unknown")
    type=0;
  else if (typeName=="1D1D")
    type=1;
  else if (typeName=="2D2D")
    type=2;
  else if (typeName=="2D1D")
    type=3;
  else if (typeName=="3D3D")
    type=4;
  else if (typeName=="3D2D")
    type=5;
  else if (typeName=="3D1D")
    type=6;
  else if (typeName=="0D0D")
    type=7;
  else if (typeName=="1D0D")
    type=8;
  else if (typeName=="2D0D")
    type=9;
  else if (typeName=="3D0D")
    type=10;
  else
    errThrow("globLoc::setType",string("unknown globLoc type :  ")
	     +typeName+".");    
}



int globLoc::getType(void) const {
  return(type);
}



string globLoc::getTypeName(void) const {
  string typeName;
  
  switch(type) {
  case -1 :
    typeName="none";
    break;
  case 0 :
    typeName="unknown";
    break;
  case 1 :
    typeName="1D1D";
    break;
  case 2 :
    typeName="2D2D";
    break;
  case 3 :
    typeName="2D1D";
    break;
  case 4 :
    typeName="3D3D";
    break;
  case 5 :
    typeName="3D2D";
    break;
  case 6 :
    typeName="3D1D";
    break;
  case 7 :
    typeName="0D0D";
    break;
  case 8 :
    typeName="1D0D";
    break;
  case 9 :
    typeName="2D0D";
    break;
  case 10 :
    typeName="3D0D";
    break;
  }
  return(typeName);
}



void globLoc::setCoord(array *A) {
  try {
    if (A!=NULL)
      if (A->getOrder()!=1)
	errThrow("globLoc::setCoord","order of the array must be 1.");
    
    delinkAndDelete(coord,&coord);
    coord=A;
    if (coord!=NULL)
      coord->addParent(&coord);
  }
  catchAll("globLoc::setCoord","at the very end")
}



array* globLoc::getCoord(void) const {
  return(coord);
}



Int globLoc::getCoordComponentsNbr(void) const {
  try {
    if (coord!=NULL) 
      return(coord->getDim(0));
    else
      return(0);
  }
  catchAll("globLoc::getCoordComponentsNbr","at the very end")
}



scal** globLoc::getCoordComponents(void) const {
  try {
    if (coord!=NULL) 
      return(&(coord->componentValues));
    else
      return(NULL);
  }
  catchAll("globLoc::getCoordComponents","at the very end")
}



void globLoc::setCoordDeriv(array *A) {
  try {
    if (A!=NULL)
      if (A->getOrder()!=2)
	errThrow("globLoc::setCoordDeriv","order of the array must be 2.");
    delinkAndDelete(coordDeriv,&coordDeriv);
    coordDeriv=A;
    if (coordDeriv!=NULL)
      coordDeriv->addParent(&coordDeriv);
  }
  catchAll("globLoc::setCoordDeriv","at the very end")
}



array* globLoc::getCoordDeriv(void) const {
  return(coordDeriv);
}



void globLoc::setJac(array *A) {
  try {
    if (A!=NULL) {
      if (A->getOrder()!=2)
	errThrow("globLoc::setJac","order of the array must be 2.");
      if (A->getDim(0)!=A->getDim(1))
	errThrow("globLoc::setJac","the array must be square.");
    }
    
    delinkAndDelete(Jac,&Jac);
    Jac=A;
    if (Jac!=NULL)
      Jac->addParent(&Jac);
  }
  catchAll("globLoc::setJac","at the very end")
}



array* globLoc::getJac(void) const {
  return(Jac);
}



void globLoc::setInvJac(array *A) {
  try {
    if (A!=NULL) {
      if (A->getOrder()!=2)
	errThrow("globLoc::setInvJac","order of the array must be 2.");
      if (A->getDim(0)!=A->getDim(1))
	errThrow("globLoc::setInvJac","the array must be square.");
    }
    
    delinkAndDelete(invJac,&invJac);
    invJac=A;
    if (invJac!=NULL)
      invJac->addParent(&invJac);
  }
  catchAll("globLoc::setInvJac","at the very end")
}



array* globLoc::getInvJac(void) const {
  return(invJac);
}



const scal& globLoc::getDetJac(void) const {
  return(detJac);
}



Int globLoc::getLocDim(void) const {
  switch(type) {
  case 1 :
    return 1;
  case 2 :
    return 2;
  case 3 :
    return 1;
  case 4 :
    return 3;
  case 5 :
    return 2;
  case 6 :
    return 1;
  case 7 :
  case 8 :
  case 9 :
  case 10 :
    return 0;
  default :
    return -1;
  }
}



Int globLoc::getGlobDim(void) const {
  switch(type) {
  case 1 :
  case 8 :
    return 1;
  case 2 :
  case 3 :
  case 9 :
    return 2;
  case 4 :
  case 5 :
  case 6 :
  case 10 :
    return 3;
  case 7 :
    return 0;
  default :
    return -1;
  }
}



Int globLoc::getGlobDimFromName(string &name) {
  string globName=name.substr(0,2);
  if (globName=="0D")
    return(0);
  else if (globName=="1D")
    return(1);
  else if (globName=="2D")
    return(2);
  else if (globName=="3D")
    return(3);
  else errThrow("globLoc::getGlobDimFromName","invalid name: "+name+".");
}



Int globLoc::getLocDimFromName(string &name) {
  string locName=name.substr(2,4);
  if (locName=="0D")
    return(0);
  else if (locName=="1D")
    return(1);
  else if (locName=="2D")
    return(2);
  else if (locName=="3D")
    return(3);
  else errThrow("globLoc::getLocDimFromName","invalid name: "+name+".");
}

      
      
void globLoc::setVectors(array *A) {
  try {
    if (A!=NULL) {
      if (A->getOrder()!=2)
	errThrow("globLoc::setVectors","the array must be 3x3 square.");
      if (A->getDim(0)!=3||A->getDim(1)!=3)
	errThrow("globLoc::setVectors","the array must be 3x3 square.");
    }
    
    delinkAndDelete(vectors,&vectors);
    vectors=A;
    if (vectors!=NULL)
      vectors->addParent(&vectors);
  }
  catchAll("globLoc::setVectors(array *A)","at the very end")
}



void globLoc::setVectors(void) {
  try {
    array *tmp=new array;
    tmp->setOrder(2);
    tmp->setDim(0,3);
    tmp->setDim(1,3);
    setVectors(tmp);  
  }
  catchAll("globLoc::setVectors(void)","at the very end")
}



array* globLoc::getVectors(void) const {
  return(vectors);
}



array* globLoc::getNormal(tensorCoordSys *TCS) const {
  try {
    arrayComp *tmpVec=NULL;
    
    string TCSTypeName="none";
    if (TCS!=NULL)
      TCSTypeName=TCS->getTypeName();
    
    switch (type) {
    case 3 :
      // 2D1D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getNormal",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,2);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(1,0);
      tmpVec->component(1)=vectors->component(1,1);
      break;
    case 4 :
      // 3D3D :
      errThrow("globLoc::getNormal",
	       "globLoc type 4 not yet implemented.");
      break;
    case 5 :
      // 3D2D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getNormal",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,3);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(2,0);
      tmpVec->component(1)=vectors->component(2,1);
      tmpVec->component(2)=vectors->component(2,2);
      break;    
    case 6 :
      // 3D1D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getNormal",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,3);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(1,0);
      tmpVec->component(1)=vectors->component(1,1);
      tmpVec->component(2)=vectors->component(1,2);
      break;
    default :
      // others :
      errThrow("globLoc::getNormal",string("invalid globLoc type :  ")
	       +getTypeName());
      break;    
    }
    
    tmpVec->setGlobLoc(this);
    
    return(dynamic_cast<array*>(tmpVec));
  }
  catchAll("globLoc::getNormal(tensorCoordSys *TCS)","at the very end")
}



array* globLoc::getTangeant(tensorCoordSys *TCS) const {
  try {
    arrayComp *tmpVec=NULL;
    
    string TCSTypeName="none";
    if (TCS!=NULL)
      TCSTypeName=TCS->getTypeName();
    
    switch (type) {
    case 1 :
      // 1D1D :
      if (TCSTypeName!="Cartesian1D"&&TCSTypeName!="none")
	errThrow("globLoc::getTangeant",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,1);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(0,0);
      break;
    case 3 :
      // 2D1D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getTangeant",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,2);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(0,0);
      tmpVec->component(1)=vectors->component(0,1);
      break;
    case 4 :
      // 3D3D :
      errThrow("globLoc::getTangeant",
	       "globLoc type 4 not yet implemented.");
      break;
    case 6 :
      // 3D1D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getTangeant",string("invalid TCS type :  ")
		 +TCSTypeName);
      tmpVec=new arrayComp;
      tmpVec->setOrder(1);
      tmpVec->setDim(0,3);
      tmpVec->setTCS(TCS);
      tmpVec->component(0)=vectors->component(0,0);
      tmpVec->component(1)=vectors->component(0,1);
      tmpVec->component(2)=vectors->component(0,2);
      break;
    default :
      // others :
      errThrow("globLoc::getTangeant",string("invalid globLoc type :  ")
	       +getTypeName());
      break;    
    }
    
    tmpVec->setGlobLoc(this);
    
    return(dynamic_cast<array*>(tmpVec));
  }
  catchAll("globLoc::getTangeant(tensorCoordSys *TCS)","at the very end")
}



array* globLoc::getVector1(tensorCoordSys *TCS) const {
  try {
    Int i,dim;
    
    arrayComp *tmpVec=NULL;
    
    string TCSTypeName="none";
    if (TCS!=NULL)
      TCSTypeName=TCS->getTypeName();
    
    switch (type) {
    case 1 :
      // 1D1D :
      if (TCSTypeName!="Cartesian1D"&&TCSTypeName!="none")
	errThrow("globLoc::getVector1",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=1;
      break;
    case 2 :
      // 2D2D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getVector1",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=2;
      break;
    case 3 :
      // 2D1D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getVector1",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=2;
      break;
    case 4 :
      // 3D3D :
      errThrow("globLoc::getVector1",
	       "globLoc type 4 not yet implemented.");
      break;
    case 5 :
      // 3D2D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector1",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    case 6 :
      // 3D1D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector1",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    default :
      // others :
      errThrow("globLoc::getVector1",string("invalid globLoc type :  ")
	       +getTypeName());
      break;    
    }
    
    tmpVec=new arrayComp;
    tmpVec->setOrder(1);
    tmpVec->setDim(0,dim);
    tmpVec->setTCS(TCS);
    for (i=0;i<dim;i++)
      tmpVec->component(i)=vectors->component(0,i);
    
    tmpVec->setGlobLoc(this);
    
    return(dynamic_cast<array*>(tmpVec));
  }
  catchAll("globLoc::getVector1(tensorCoordSys *TCS)","at the very end")
}



array* globLoc::getVector2(tensorCoordSys *TCS) const {
  try {
    Int i,dim;
    
    arrayComp *tmpVec=NULL;
    
    string TCSTypeName="none";
    if (TCS!=NULL)
      TCSTypeName=TCS->getTypeName();
    
    switch (type) {
    case 2 :
      // 2D2D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getVector2",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=2;
      break;
    case 3 :
      // 2D1D :
      if (TCSTypeName!="Cartesian2D"&&TCSTypeName!="none")
	errThrow("globLoc::getVector2",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=2;
      break;
    case 4 :
      // 3D3D :
      errThrow("globLoc::getVector2",
	       "globLoc type 4 not yet implemented.");
      break;
    case 5 :
      // 3D2D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector2",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    case 6 :
      // 3D1D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector2",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    default :
      // others :
      errThrow("globLoc::getVector2",string("invalid globLoc type :  ")
	       +getTypeName());
      break;    
    }
    
    tmpVec=new arrayComp;
    tmpVec->setOrder(1);
    tmpVec->setDim(0,dim);
    tmpVec->setTCS(TCS);
    for (i=0;i<dim;i++)
      tmpVec->component(i)=vectors->component(0,i);
    
    tmpVec->setGlobLoc(this);
    
    return(dynamic_cast<array*>(tmpVec));
  }
  catchAll("globLoc::getVector2(tensorCoordSys *TCS)","at the very end")
}



array* globLoc::getVector3(tensorCoordSys *TCS) const {
  try {
    Int i,dim;
    
    arrayComp *tmpVec=NULL;
    
    string TCSTypeName="none";
    if (TCS!=NULL)
      TCSTypeName=TCS->getTypeName();
    
    switch (type) {
    case 4 :
      // 3D3D :
      errThrow("globLoc::getVector3",
	       "globLoc type 4 not yet implemented.");
      break;
    case 5 :
      // 3D2D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector3",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    case 6 :
      // 3D1D :
      if (TCSTypeName!="Cartesian3D"&&TCSTypeName!="CylindricalRZT"
	  &&TCSTypeName!="none")
	errThrow("globLoc::getVector3",string("invalid TCS type :  ")
		 +TCSTypeName);
      dim=3;
      break;
    default :
      // others :
      errThrow("globLoc::getVector3",string("invalid globLoc type :  ")
	       +getTypeName());
      break;    
    }
    
    tmpVec=new arrayComp;
    tmpVec->setOrder(1);
    tmpVec->setDim(0,dim);
    tmpVec->setTCS(TCS);
    for (i=0;i<dim;i++)
      tmpVec->component(i)=vectors->component(0,i);
    
    tmpVec->setGlobLoc(this);
    
    return(dynamic_cast<array*>(tmpVec));
  }
  catchAll("globLoc::getVector3(tensorCoordSys *TCS)","at the very end")
}



void globLoc::makeAtPoint(object *element,int direct,Double *localCoord,
                          bool makeValue,bool makeDeriv) {

//#warning Vrifier toutes les normales !!!

  if (!newPoint::getNewPoint())
    return;
  newPoint::setNewPoint(false);

  static scal Xxi,Yxi,Zxi,Xeta,Yeta,Zeta,tmpScal,tmpScalX,tmpScalY,tmpScalZ;
  static scal tmp1,tmp2;
  static scal invDetJac,tmpFact;
  const static scal unit(1.0);
  static DPowMult SQRT(1.0,0.5);
  scal *e1,*e2,*e3;
  static scal genericVectors[9];
  
  try {
    
    if (vectors!=NULL) {
      e1=vectors->componentValues;
      e2=vectors->componentValues+3;
      e3=vectors->componentValues+6;
    }
    else {
      e1=genericVectors;
      e2=genericVectors+3;
      e3=genericVectors+9;
    }

    if (coord!=NULL)
      coord->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
    if (coordDeriv!=NULL)
      coordDeriv->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);  
    
    switch(type) {
    case -1 :
    case 0 :
      // 0D0D :
      // nothing to do !
      Jac->componentValues[0]=0.0;
      invJac->componentValues[0]=0.0;
      detJac=0.0;
      e1[0]=0.0;
      e1[1]=0.0; 
      e1[2]=0.0;
      e2[0]=0.0; 
      e2[1]=0.0; 
      e2[2]=0.0;
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=0.0;
      break;
    case 1 :
      // 1D1D :
      Jac->componentValues[0]=coordDeriv->componentValues[0];
      invJac->componentValues[0]=unit/coordDeriv->componentValues[0];
      detJac=coordDeriv->componentValues[0];
      e1[0]=1.0; 
      e1[1]=0.0; 
      e1[2]=0.0;
      e2[0]=0.0; 
      e2[1]=0.0; 
      e2[2]=0.0;
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=0.0;
      break;
    case 2 :
      // 2D2D :
      Jac->componentValues[0]=coordDeriv->componentValues[0];
      Jac->componentValues[1]=coordDeriv->componentValues[1];
      Jac->componentValues[2]=coordDeriv->componentValues[2];
      Jac->componentValues[3]=coordDeriv->componentValues[3];
      detJac=coordDeriv->componentValues[0]*coordDeriv->componentValues[3];
      detJac-=coordDeriv->componentValues[1]*coordDeriv->componentValues[2];
      invDetJac=unit/detJac;
      invJac->componentValues[0]=Jac->componentValues[3]*invDetJac;
      invJac->componentValues[1]=-Jac->componentValues[1]*invDetJac;
      invJac->componentValues[2]=-Jac->componentValues[2]*invDetJac;
      invJac->componentValues[3]=Jac->componentValues[0]*invDetJac;
      // Vector e1 = e_xi :
      tmp1=coordDeriv->componentValues[0]*coordDeriv->componentValues[0];
      tmp1+=coordDeriv->componentValues[2]*coordDeriv->componentValues[2];    
      tmp2.functionExec(&tmp1,&SQRT,makeValue,makeDeriv);
      tmp1=unit/tmp2;
      e1[0]=coordDeriv->componentValues[0]*tmp1; 
      e1[1]=coordDeriv->componentValues[2]*tmp1; 
      e1[2]=0.0;
      // Vector e2 = e_eta :
      tmp1=coordDeriv->componentValues[1]*coordDeriv->componentValues[1];
      tmp1+=coordDeriv->componentValues[3]*coordDeriv->componentValues[3];    
      tmp2.functionExec(&tmp1,&SQRT,makeValue,makeDeriv);
      tmp1=unit/tmp2;
      e2[0]=coordDeriv->componentValues[1]*tmp1; 
      e2[1]=coordDeriv->componentValues[3]*tmp1; 
      e2[2]=0.0;
      // Vector e3 = 0 :
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=0.0;
      break;
    case 3 :
      // 2D1D :
      tmpScal=coordDeriv->componentValues[0]*coordDeriv->componentValues[0];
      tmpScal+=coordDeriv->componentValues[1]*coordDeriv->componentValues[1];    
      detJac.functionExec(&tmpScal,&SQRT,makeValue,makeDeriv);
      invDetJac=unit/detJac;
      Jac->componentValues[0]=coordDeriv->componentValues[0];
      Jac->componentValues[1]=coordDeriv->componentValues[1]*invDetJac;
      Jac->componentValues[2]=coordDeriv->componentValues[1];
      Jac->componentValues[3]=-coordDeriv->componentValues[0]*invDetJac;
      tmpFact=unit/tmpScal;
      invJac->componentValues[0]=-Jac->componentValues[3]*invDetJac;
      invJac->componentValues[1]=Jac->componentValues[1]*invDetJac;
      invJac->componentValues[2]=Jac->componentValues[2]*invDetJac;
      invJac->componentValues[3]=-Jac->componentValues[0]*invDetJac;
      // Vector e1 = e_xi :
      e1[0]=coordDeriv->componentValues[0]*invDetJac; 
      e1[1]=coordDeriv->componentValues[1]*invDetJac; 
      e1[2]=0.0;
      // Vector e2 orthogonal to e1 :
      e2[0]=e1[1]; 
      e2[1]=-e1[0]; 
      e2[2]=0.0;
      // Vector e3 = 0 :
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=0.0;			       
      break;
    case 4 :
      // 3D3D :
      errThrow("globLoc::makeAtPoint",
	       "globLoc type 4 not yet implemented.");
      break;
    case 5 :
      // 3D2D :
      // One makes the assumption that only the two first coordinates are not 
      //    zero (i.e. domain in plane XY).
      Jac->componentValues[0]=coordDeriv->componentValues[0];
      Jac->componentValues[1]=coordDeriv->componentValues[1];
      Jac->componentValues[3]=coordDeriv->componentValues[2];
      Jac->componentValues[4]=coordDeriv->componentValues[3];
      Jac->componentValues[8]=1.0;
      detJac=coordDeriv->componentValues[0]*coordDeriv->componentValues[3];
      detJac-=coordDeriv->componentValues[1]*coordDeriv->componentValues[2];
      invDetJac=unit/detJac;
      invJac->componentValues[0]=Jac->componentValues[4]*invDetJac;
      invJac->componentValues[1]=-Jac->componentValues[1]*invDetJac;
      invJac->componentValues[3]=-Jac->componentValues[3]*invDetJac;
      invJac->componentValues[4]=Jac->componentValues[0]*invDetJac;
      invJac->componentValues[8]=1.0;    
      // Vector e1 = e_xi :
      tmp1=coordDeriv->componentValues[0]*coordDeriv->componentValues[0];
      tmp1+=coordDeriv->componentValues[2]*coordDeriv->componentValues[2];    
      tmp2.functionExec(&tmp1,&SQRT,makeValue,makeDeriv);
      tmp1=unit/tmp2;
      e1[0]=coordDeriv->componentValues[0]*tmp1; 
      e1[1]=coordDeriv->componentValues[2]*tmp1; 
      e1[2]=0.0;
      // Vector e2 = e_eta :
      tmp1=coordDeriv->componentValues[1]*coordDeriv->componentValues[1];
      tmp1+=coordDeriv->componentValues[3]*coordDeriv->componentValues[3];    
      tmp2.functionExec(&tmp1,&SQRT,makeValue,makeDeriv);
      tmp1=unit/tmp2;
      e2[0]=coordDeriv->componentValues[1]*tmp1; 
      e2[1]=coordDeriv->componentValues[3]*tmp1; 
      e2[2]=0.0;
      // Vector e3 = [0,0,1] :
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=1.0;
      break;
    case 6:
      // 3D1D :
      // One makes the assumption that only the two first coordinates are not 
      //    zero (i.e. domain in plane XY).
      tmpScal=coordDeriv->componentValues[0]*coordDeriv->componentValues[0];
      tmpScal+=coordDeriv->componentValues[1]*coordDeriv->componentValues[1];    
      detJac.functionExec(&tmpScal,&SQRT,makeValue,makeDeriv);
      invDetJac=unit/detJac;
      Jac->componentValues[0]=coordDeriv->componentValues[0];
      Jac->componentValues[1]=coordDeriv->componentValues[1]*invDetJac;
      Jac->componentValues[3]=coordDeriv->componentValues[1];
      Jac->componentValues[4]=-coordDeriv->componentValues[0]*invDetJac;
      Jac->componentValues[8]=1.0;
      invJac->componentValues[0]=-Jac->componentValues[4]*invDetJac;
      invJac->componentValues[1]=Jac->componentValues[1]*invDetJac;
      invJac->componentValues[3]=Jac->componentValues[3]*invDetJac;
      invJac->componentValues[4]=-Jac->componentValues[0]*invDetJac;
      invJac->componentValues[8]=1.0;
      // Vector e1 = e_xi :
      e1[0]=coordDeriv->componentValues[0]*invDetJac; 
      e1[1]=coordDeriv->componentValues[1]*invDetJac; 
      e1[2]=0.0;
      // Vector e2 orthogonal to e1 :
      e2[0]=e1[1]; 
      e2[1]=-e1[0]; 
      e2[2]=0.0;
      // Vector e3 = [0,0,1] :
      e3[0]=0.0; 
      e3[1]=0.0; 
      e3[2]=1.0;
      break ;
    case 7:
    case 8:
    case 9:
    case 10:
      detJac=1.0;
      break ;
    }
  }
  catchAll("globLoc::makeAtPoint","at the very end")
}



void globLoc::setNewPoint(bool b) {
  try {
    newPoint::setNewPoint(b);
    
    if (coord!=NULL)
      coord->setNewPoint(b);
    if (coordDeriv!=NULL)
      coordDeriv->setNewPoint(b);
    if (Jac!=NULL)
      Jac->setNewPoint(b);
    if (invJac!=NULL)
      invJac->setNewPoint(b);
  }
  catchAll("globLoc::setNewPoint","at the very end")
}



void globLoc::accessNbrInit(void) {
  try {
    newPoint::accessNbrInit();
    
    if (coord!=NULL)
      coord->accessNbrInit();
    if (coordDeriv!=NULL)
      coordDeriv->accessNbrInit();
    if (Jac!=NULL)
      Jac->accessNbrInit();
    if (invJac!=NULL)
      invJac->accessNbrInit();
  }
  catchAll("globLoc::accessNbrInit","at the very end")
}



set<field*>* globLoc::buildActiveFieldList(void) {
  try {
    static set<field*> FL ;
    
    if (coord!=NULL)
      FL=(*(coord->buildActiveFieldList()));
    
    return(&FL);
  }
  catchAll("globLoc::buildActiveFieldList","at the very end")
}



printTree globLoc::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (getName()=="")
      pt+="   anonymous";
    else
      pt+=("   name = "+getName());
    sprintf(cbuf,"   type = ");
    sbuf=cbuf;
    sbuf+=getTypeName();
    pt+=sbuf;
    
    return pt;
  }
  catchAll("globLoc::getPrintTreeData","at the very end")
}



printTree globLoc::getPrintTreeValues(void) const {
  try {
    printTree pt;
    return pt;
  }
  catchAll("globLoc::getPrintTreeValues","at the very end")
}


  
printTree globLoc::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt,ptArg;
    
    sprintf(cbuf,"globLoc %p",static_cast<const void*>(this));
    pt+=cbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    if (coord!=NULL)
      ptArg=coord->getFirstPrintTree("coord",values);
    else
      ptArg="coord = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;
    
     if (coordDeriv!=NULL)
      ptArg=coordDeriv->getFirstPrintTree("coordDeriv",values);
    else
      ptArg="coordDeriv = NULL";
    ptArg.prepareForAddLast();
    pt+=ptArg;
    
   return(pt);
  }
  catchAll("globLoc::getPrintTree","at the very end")
}
    



