
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <typeinfo>
#include <string>
#include <stdio.h>
 
#include "tensorCoordSys_base.h"
#include "array/array_base.h"
#include "err/err.h"
#include "print/printTree.h"
#include "scal/scal.h"

using namespace std ;
using namespace keyFE2 ;



void tensorCoordSys::init(void) {
  Coord=NULL;
  CoordComponentsNbr=0;
  CoordComponents=NULL;
  CovaMetric=NULL;
  ContraMetric=NULL;
  Christoffel=NULL;
  type=0;
}



tensorCoordSys::tensorCoordSys(void) {
  try {
    init();
    name::name();
    newPoint::newPoint();
    parents::parents();
  }
  catchAll("tensorCoordSys::tensorCoordSys(void)","")
}



tensorCoordSys::tensorCoordSys(const tensorCoordSys &in) {
  try {
    init();
    
    setCoord(in.Coord);
    setCovaMetric(in.CovaMetric);
    setContraMetric(in.ContraMetric);
    setChristoffel(in.Christoffel);
    setType(in.getType());
    
    setName(in.getName());
    newPoint::newPoint();
    parents::parents();
  }
  catchAll("tensorCoordSys::tensorCoordSys(const tensorCoordSys &in)","")
}



tensorCoordSys::~tensorCoordSys(void) {
  try {
    setCoord(NULL);
    setCovaMetric(NULL);
    setContraMetric(NULL);
    setChristoffel(NULL);
  }
  catchAll("tensorCoordSys::~tensorCoordSys(void)","")
}



void tensorCoordSys::setCoord(array *coord) {
  try {
    if (coord!=NULL)
      if (coord->getOrder()!=1)
	errThrow("tensorCoordSys::setCoord","the order of the array must be 1.");
    delinkAndDelete(Coord,&Coord);
    Coord=coord;
    if (Coord!=NULL)
      Coord->addParent(&Coord);
    
    if (CoordComponents!=NULL) {
      delete[] CoordComponents;
      CoordComponents=NULL;
      CoordComponentsNbr=0;
    }
    if (Coord!=NULL) {
      CoordComponentsNbr=Coord->getDim(0);
      CoordComponents=new scal* [CoordComponentsNbr];
      for (Int i=0;i<CoordComponentsNbr;i++)
	CoordComponents[i]=Coord->componentValues+i;
    }
  }
  catchAll("tensorCoordSys::setCoord(array *coord)","")
}



array* tensorCoordSys::getCoord(void) const {
  return(Coord);
}



int tensorCoordSys::getCoordComponentsNbr(void) const {
  return(CoordComponentsNbr);
}



scal** tensorCoordSys::getCoordComponents(void) const {
  return(CoordComponents);
}



scal* tensorCoordSys::getCoordComponents(Int index) const {
  try {
    return(CoordComponents[index]);
  }
  catchAll("tensorCoordSys::getCoordComponents","")
}



void tensorCoordSys::setCovaMetric(array *metric) {
  try {
    if (metric!=NULL) {
      if (metric->getOrder()!=2)
	errThrow("tensorCoordSys::setCovaMetric",
		 "order of the array must be 2.");
      if (metric->getDim(0)!=metric->getDim(1))
	errThrow("tensorCoordSys::setCovaMetric",
		 "the array should be square.");
    }
    
    if (CovaMetric!=NULL)
      delinkAndDelete(CovaMetric,&CovaMetric);
    
    CovaMetric=metric;
    if (CovaMetric!=NULL)
      CovaMetric->addParent(&CovaMetric);
  }
  catchAll("tensorCoordSys::setCovaMetric","")
}



array* tensorCoordSys::getCovaMetric(void) const {
  return(CovaMetric);
}



void tensorCoordSys::setContraMetric(array *metric) {
  try {
    if (metric!=NULL) {
      if (metric->getOrder()!=2)
	errThrow("tensorCoordSys::setMetric","order of the array must be 2.");
      if (metric->getDim(0)!=metric->getDim(1))
	errThrow("tensorCoordSys::setMetric","the array should be square.");
    }
    
    if (ContraMetric!=NULL)
      delinkAndDelete(ContraMetric,&ContraMetric);
    
    ContraMetric=metric;
    if (ContraMetric!=NULL)
      ContraMetric->addParent(&ContraMetric);
  }
  catchAll("tensorCoordSys::setContraMetric","")
}



array* tensorCoordSys::getContraMetric(void) const {
  return(ContraMetric);
}



void tensorCoordSys::setChristoffel(array *christoffel) {
  try {
    if (christoffel!=NULL) {
      if (christoffel->getOrder()!=3)
	errThrow("tensorCoordSys::setChristoffel",
		 "order of the array must be 3.");
      if (christoffel->getDim(0)!=christoffel->getDim(1)
	  ||christoffel->getDim(0)!=christoffel->getDim(2))
	errThrow("tensorCoordSys::setChristoffel",
		 "the array should be cubic.");
    }
    delinkAndDelete(Christoffel,&Christoffel);
    Christoffel=christoffel;
    if (Christoffel!=NULL)
      Christoffel->addParent(&Christoffel);
  }
  catchAll("tensorCoordSys::setChristoffel","")
}



array* tensorCoordSys::getChristoffel(void) const {
  return(Christoffel);
}



void tensorCoordSys::setType(Int index) {
  type=index;
}



void tensorCoordSys::setType(string typeName) {
  if (typeName=="none")
    type=0;
  else if (typeName=="Cartesian1D")
    type=1;
  else if (typeName=="Cartesian2D")
    type=2;
  else if (typeName=="Cartesian3D")
    type=3;
  else if (typeName=="CylindricalRZT")
    type=4;
  else
    errThrow("tensorCoordSys::setType",string("unknown globLoc type :  ")
	     +typeName+".");    
}



int tensorCoordSys::getType(void) const {
  return(type);
}



string tensorCoordSys::getTypeName(void) const {
  switch(type) {
  case 0 :
    return("none");
  case 1 :
    return("Cartesian1D");
  case 2 :
    return("Cartesian2D");
  case 3 :
    return("Cartesian3D");
  case 4 :
    return("CylindricalRZT");
  }
  return("unknown");
}



tensorCoordSys* tensorCoordSys::makeCopyForCoordSys(void) const {
  try {
    return(new tensorCoordSys(*this));
  }
  catchAll("tensorCoordSys::makeCopyForCoordSys","")
}



void tensorCoordSys::makeAtPoint(object *element,Int direct,Double *localCoord,
				 bool makeValue,bool makeDeriv) {
  
  if (!newPoint::getNewPoint())
    return;
  newPoint::setNewPoint(false);
  
  try {
    switch(type) {
    case 1 :
    case 2 :
    case 3 :
      // nothing to do !
      break;
    case 4 :
      // Cylindrical RZT :
      {
	static scal unit(1.0);
	static scal R,invR;
	
	Coord->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
	
	R=Coord->componentValues[0];
	if (R.getVal()!=0.0)
	  invR=unit/R;
	else
	  invR=1000000.0;
	
	CovaMetric->componentValues[0]=1.0;
	CovaMetric->componentValues[4]=1.0;
	CovaMetric->componentValues[8]=R*R;
	
	ContraMetric->componentValues[0]=1.0;
	ContraMetric->componentValues[4]=1.0;
	ContraMetric->componentValues[8]=invR*invR;
	
	Christoffel->componentValues[8]=invR;
	Christoffel->componentValues[20]=-invR;
      }
      break;
    default :
      if (Coord!=NULL)
	Coord->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
      if (CovaMetric!=NULL)
	CovaMetric->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
      if (ContraMetric!=NULL)
	ContraMetric->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
      if (Christoffel!=NULL)
	Christoffel->makeAtPoint(element,direct,localCoord,makeValue,makeDeriv);
      break;
    }
    return;
  }
  catchAll("tensorCoordSys::makeAtPoint","")
}



void tensorCoordSys::setNewPoint(bool b) {
  try {
    newPoint::setNewPoint(b);
    
    if (Coord!=NULL)
      Coord->setNewPoint(b);
    if (CovaMetric!=NULL)
      CovaMetric->setNewPoint(b);
    if (ContraMetric!=NULL)
      ContraMetric->setNewPoint(b);
    if (Christoffel!=NULL)
      Christoffel->setNewPoint(b);
  }
  catchAll("tensorCoordSys::setNewPoint","")
}



void tensorCoordSys::accessNbrInit(void) {
  try {
    newPoint::accessNbrInit();
    
    if (Coord!=NULL)
      Coord->accessNbrInit();
    if (CovaMetric!=NULL)
      CovaMetric->accessNbrInit();
    if (ContraMetric!=NULL)
      ContraMetric->accessNbrInit();
    if (Christoffel!=NULL)
      Christoffel->accessNbrInit();
  }
  catchAll("tensorCoordSys::accessNbrInit","")
}



printTree tensorCoordSys::getPrintTreeData(void) const {
  try {
    printTree pt;
    if (getName()=="")
      pt+="   anonymous";
    else
      pt+=("   name = "+getName());
    pt+=("   type = "+getTypeName());
    return(pt);
  }
  catchAll("tensorCoordSys::getPrintTreeData","")
}



printTree tensorCoordSys::getPrintTreeValues(void) const {
  try {
    printTree pt;
    return pt;
  }
  catchAll("tensorCoordSys::getPrintTreeValues","")
}


  
printTree tensorCoordSys::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    
    sprintf(cbuf,"tensor coordinate system %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;    
    
    pt.prepareForArguments();
    
    if (Coord!=NULL)
      ptArg=Coord->getFirstPrintTree("Coord",values);
    else
      ptArg="Coord = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;
  
    if (CovaMetric!=NULL)
      ptArg=CovaMetric->getFirstPrintTree("CovaMetric",values);
    else
      ptArg="CovaMetric = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;

    if (ContraMetric!=NULL)
      ptArg=ContraMetric->getFirstPrintTree("ContraMetric",values);
    else
      ptArg="ContraMetric = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    if (Christoffel!=NULL)
      ptArg=Christoffel->getFirstPrintTree("Christoffel",values);
    else
      ptArg="Christoffel = NULL";
    ptArg.prepareForAddLast();
    pt+=ptArg;
    
    return(pt);
  }
  catchAll("tensorCoordSys::getPrintTree","")
}




