
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <cmath>
#include <list>
#include <map>
#include <string.h>

#include "exprCollect.h"
#include "expression.h"
#include "err/err.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



exprCollect::exprCollect(void) {
  try {
    capacity=10;
    size=0;
    exprData=new expression[capacity];
  }
  catchAll("exprCollect::exprCollect(void)","")
}



exprCollect::exprCollect(const exprCollect &in) {
  try {
    Int i;
    
    capacity=in.capacity;
    size=in.size;
    if (capacity>0)
      exprData=new expression[capacity];
    else
      exprData=NULL;
    for (i=0;i<size;i++)
      exprData[i]=in.exprData[i];
  }
  catchAll("exprCollect::exprCollect(const exprCollect &in)","")
}



exprCollect::~exprCollect(void) {
  try {
    capacity=0;
    size=0;
    delete[] exprData;
    exprData=NULL;
  }
  catchAll("exprCollect::~exprCollect(void)","")
}



void exprCollect::operator=(const exprCollect &in) {
  if (this==&in)
    return;
  
  try {
    Int i;
    
    capacity=max(10,in.size);
    size=in.size;
    delete[] exprData;
    exprData=new expression[capacity];
    for (i=0;i<size;i++)
      exprData[i]=in.exprData[i];
  }
  catchAll("void exprCollect::operator=(const exprCollect &in)","")
}



void exprCollect::clear(void) {
  size=0;
}



Int exprCollect::getSize(void) const {
  return(size);
}



void exprCollect::setCapacity(Int nbr) {
  try {
    capacity=nbr;
    size=0;
    delete[] exprData;
    if (capacity>0)
      exprData=new expression[capacity];
    else
      exprData=NULL;
  }
  catchAll("void exprCollect::setCapacity(Int nbr)","")
}



void exprCollect::pushBack(const expression &arg) {
  try {
    if (size==capacity) {
      Int i;
      expression *tmpExpr=new expression[2*capacity+1];
      for (i=0;i<size;i++)
	tmpExpr[i]=exprData[i];
      delete[] exprData;
      exprData=tmpExpr;
      capacity=2*capacity+1;
    }
    exprData[size++]=arg;
  }
  catchAll("void exprCollect::pushBack(const expression &arg)","")
}



void exprCollect::setLayer(Int nbr) {
    for (Int i=0;i<size;i++)
      exprData[i].setLayer(nbr);
}



void exprCollect::merge(const exprCollect &left,const exprCollect &right) {
  try {
    // One assumes that all expressions are sorted by increasing index order.
    setCapacity(left.size+right.size);
    
    Int leftIndex,rightIndex;
    leftIndex=0;
    rightIndex=0;
    
    while (leftIndex<left.size&&rightIndex<right.size)
      if (left.exprData[leftIndex].index<left.exprData[rightIndex].index)
	exprData[size++]=left.exprData[leftIndex++];
    else if (left.exprData[leftIndex].index>left.exprData[rightIndex].index)
      exprData[size++]=right.exprData[rightIndex++];
    else
      exprData[size++].merge(left.exprData[leftIndex++],
			     right.exprData[rightIndex++]);
    
    while (leftIndex<left.size)
      exprData[size++]=left.exprData[leftIndex++];
    
    while (rightIndex<right.size)
      exprData[size++]=right.exprData[rightIndex++];
  }
  catchAll("void exprCollect::merge(const exprCollect &left,const exprCollect &right)","")
}



Int exprCollect::getPosFromIndex(Int nbr) {
  try {
    // Dichotomic search of nbr in the indices.
    Int posUp,posDown,posCurrent;
    
    if (size==0)
      return(-1);
    
    posDown=0;
    posUp=size-1;
    if (exprData[posDown].index==nbr)
      return(posDown);
    if (exprData[posUp].index==nbr)
      return(posUp);
    
    while(posUp>posDown+1) {
      posCurrent=(posUp+posDown)/2;
      if (exprData[posCurrent].index==nbr)
	return(posCurrent);
      if (exprData[posCurrent].index>nbr)
	posUp=posCurrent;
      else
	posDown=posCurrent;
    }
    return(-1);
  }
  catchAll("Int exprCollect::getPosFromIndex(Int nbr)","")
}



void exprCollect::solve(void) {

  // One makes the assumption that the list of expressions is sorted by index!

  try {
    list<expression> eList,eList2;
    map<Int,expression> eMap;
    map<Int,expression>::iterator eMapIt;
    expression tmpExpr,tmpExpr2;
    Int i,iter,elimNbr,tmpIndex;
    bool found;
    Double tmpFactor;
    
    // One merges the expressions that share an identical index :
    
    elimNbr=0;
    eList.clear();
    if (size>0)
      tmpExpr=exprData[0];
    i=1;
    while (i<size) {
      if (tmpExpr.index==exprData[i].index) {
        tmpExpr2.merge(tmpExpr,exprData[i]);
	tmpExpr=tmpExpr2;
	i++;
      }
      else {
        eList.push_back(tmpExpr);
	tmpExpr=exprData[i];
	i++;
      }
    }
    if (size>0)
      eList.push_back(tmpExpr);
      
    // Forward (iter=1) and backward (iter=2) substitutions if needed :
      
    for (iter=1;iter<=2;iter++) {
      
      // Initialisation :
      
      eMap.clear();
      eList2.clear();
      
      // elimination :
      
      while(!eList.empty()) {
	tmpExpr=eList.front();
	eList.erase(eList.begin());
	if (tmpExpr.index!=-1) {
	  // index substitution :
	  eMapIt=eMap.find(tmpExpr.index);
	  if (eMapIt!=eMap.end())
	    tmpExpr.substitute(eMapIt->second,tmpExpr);
	  // indices substitution :
	  do {
	    found=false;
	    for (i=0;i<tmpExpr.size;i++) {
	      eMapIt=eMap.find(tmpExpr.indices[i]);
	      if (eMapIt!=eMap.end()) {
		found=true;
		tmpExpr.substitute(eMapIt->second,tmpExpr);
	      }
	    }
	  } while (found) ;
	  // disposing of expression :
	  if (tmpExpr.index!=-1) {
	    eMap[tmpExpr.index]=tmpExpr;
	    eList2.push_back(tmpExpr);
	  }
	  else
	    eList.push_back(tmpExpr);
	}
	else {
	  // indices substitution :
	  do {
	    found=false;
	    for (i=0;i<tmpExpr.size;i++) {
	      eMapIt=eMap.find(tmpExpr.indices[i]);
	      if (eMapIt!=eMap.end()) {
		found=true;
		tmpExpr.substitute(eMapIt->second,tmpExpr);
	      }
	    }
	  } while (found) ;
	  // Manipulation of expression :
	  if (tmpExpr.size>0) {
	    //Explicitation of expression :
	    tmpIndex=tmpExpr.indices[0];
	    tmpFactor=tmpExpr.factors[0];
	    for (i=1;i<tmpExpr.size;i++) {
	      if (abs(tmpExpr.factors[i])>abs(tmpFactor)) {
		tmpIndex=tmpExpr.indices[i];
		tmpFactor=tmpExpr.factors[i];
	      }
	    }
	    tmpExpr2.setCapacity(tmpExpr.size-1);
	    tmpExpr2.weight=tmpExpr.weight;
            tmpExpr2.setLayer(tmpExpr.layer);
	    tmpExpr2.index=tmpIndex;
	    tmpExpr2.independentValue=tmpExpr.independentValue/(-tmpFactor);
	    tmpExpr2.size=0;
	    for (i=0;i<tmpExpr.size;i++)
	      if (tmpExpr.indices[i]!=tmpExpr2.index)
		tmpExpr2.addTerm(tmpExpr.indices[i],
				 tmpExpr.factors[i]/(-tmpFactor));
	    tmpExpr=tmpExpr2;
	    eMap[tmpExpr.index]=tmpExpr;
	    eList2.push_back(tmpExpr);
	  }
	  else {
	    // Elimination of expression :
	    if (abs(tmpExpr.independentValue)<tmpExpr.zero)
	      elimNbr++;
	    else {
	      elimNbr++;
	      cerr << "warning :  "<< tmpExpr.independentValue << "!=" << 0.0 << endl ;
	      //	    errThrow("exprCollect::eliminate",
	      //		     "incompatible constraints have been detected");
	    }
	  } // if (tmpExpr.size>0)
	}
      } // while(!eList.empty())
      
      eList2.reverse();
      eList=eList2;
    } // for iter ...
    
    // Remake of the collection :
    
    size=0;
    for (eMapIt=eMap.begin();eMapIt!=eMap.end();eMapIt++) {
      tmpExpr=eMapIt->second;
      tmpExpr.weight=1.0;
      pushBack(tmpExpr);
    }
    
    return;
  }
  catchAll("void exprCollect::solve(void)","")
}



void exprCollect::write(ostream &os) const {
  try {
    Int i;
    for (i=0;i<size;i++)
      exprData[i].write(os);
  }
  catchAll("void exprCollect::write(ostream &os) const","")
}

