
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <cmath>
#include <iostream>
#include <iomanip>
#include <list>

#include "expression.h"

#include "err/err.h"
#include "types/params.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



Double expression::zero=1.0e-38;



expression::expression(void) {
  layer=0;
  weight=1.0;
  index=-1;
  independentValue=0.0;
  
  capacity=0;
  size=0;
  factors=NULL;
  indices=NULL;
}



expression::expression(Int initCapacity) {
  try {
    layer=0;
    weight=1.0;
    index=-1;
    independentValue=0.0;
    
    capacity=initCapacity;
    size=0;
    if (initCapacity>0) {
      factors=new Double[initCapacity];
      indices=new Int[initCapacity];
    }
    else {
      factors=NULL;
      indices=NULL;
    }    
  }
  catchAll("expression::expression(Int initCapacity)","")
}  



expression::expression(const expression &in) {
  try {
    Int i;
    
    layer=in.layer;
    weight=in.weight;
    index=in.index;
    independentValue=in.independentValue;
    
    capacity=in.capacity;
    size=in.size;
    
    if (capacity>0) {
      factors=new Double[capacity];
      indices=new Int[capacity];
    }
    else {
      factors=NULL;
      indices=NULL;
    }    
    
    for (i=0;i<size;i++) {
      factors[i]=in.factors[i];
      indices[i]=in.indices[i];
    }
  }
  catchAll("expression::expression(const expression &in)","")
}



expression::~expression(void) {
  try {
    layer=0;
    weight=1.0;
    index=-1;
    independentValue=0.0;
    
    capacity=0;
    size=0;
    delete[] factors;
    factors=NULL;
    delete[] indices;
    indices=NULL;
  }
  catchAll("expression::~expression(void)","")
}  



void expression::setZero(Double d) {
  if (d<=0.0)
    errThrow("expression::setZero","invalid value :  zero<=0.0");
  zero=d;
}



Double expression::getZero(void) const {
  return(zero);
}



void expression::operator=(const expression &in) {
  if (this==&in)
    return;
  
  try {
    Int i;
    
    layer=in.layer;
    weight=in.weight;
    index=in.index;
    independentValue=in.independentValue;
    
    capacity=in.capacity;
    size=in.size;
    
    delete[] factors;
    delete[] indices;
    if (capacity>0) {
      factors=new Double[capacity];
      indices=new Int[capacity];
    }
    else {
      factors=NULL;
      indices=NULL;
    }
    for (i=0;i<size;i++) {
      factors[i]=in.factors[i];
      indices[i]=in.indices[i];
    }
  }
  catchAll("void expression::operator=(const expression &in)","")
}



bool expression::operator<(const expression &right) {
  if (index<right.index)
    return true;
  if (index==right.index&&size<right.size)
    return true;
  return false;
}


void expression::setLayer(Int nbr) {
  layer=nbr;
}



void expression::setIndex(Int nbr) {
  index=nbr;
}



void expression::setIndependentValue(Double val) {
  independentValue=val;
}



void expression::setWeight(Double w) {
  weight=w;
}



Double expression::getWeight(void) const {
  return(weight);
}



void expression::setCapacity(Int nbr) {
  try {
    if (nbr!=capacity) {
      delete[] factors;
      delete[] indices;
      capacity=nbr;
      if (capacity>0) {
	factors=new Double[capacity];
	indices=new Int[capacity];
      }
      else {
	factors=NULL;
	indices=NULL;
      }
    }
    size=0;
  }
  catchAll("void expression::setCapacity(Int nbr)","")
}



Int expression::getCapacity(void) const {
  return(capacity);
}



void expression::clear(void) {
  weight=1.0;
  index=-1;
  independentValue=0.0;
  size=0;
}  



void expression::addTerm(Int nbr,Double val) {
  if (size==capacity)
    errThrow("expression::addTerm","expression overflow.");
  
  if (val==0.0)
    return;
  
  indices[size]=nbr;
  factors[size++]=val;
}



void expression::merge(const expression &left,const expression &right) {
  try {    
    static expression tmpExpr;
    
    // One assumes that the indices are sorted in ascending order !
    
    if (left.index!=right.index)
      errThrow("expression::merge",
	       string("indices of the arguments do not correspond :  ")
	       +strFromInt(left.index)+"!="+strFromInt(right.index)+".");
	       
    if (left.layer==right.layer) {
      // One first counts the number of common indices in order to estimate
      //    the number of dependent indices of the merged expression :
      
      Int leftPos=0;
      Int rightPos=0;
      Int commonCounter=0;
      while (leftPos<left.size&&rightPos<right.size) {
   	if (left.indices[leftPos]<right.indices[rightPos])
 	  leftPos++;
   	else if (left.indices[leftPos]>right.indices[rightPos])
 	  rightPos++;
   	else {
 	  leftPos++;
 	  rightPos++;
 	  commonCounter++;
   	}
      }
      
      // One then builds the merged list :
      
      tmpExpr.setCapacity(left.size+right.size);
      tmpExpr.weight=left.weight+right.weight;
      Double leftFactor=left.weight/tmpExpr.weight;
      Double rightFactor=right.weight/tmpExpr.weight;
      tmpExpr.index=left.index;
      tmpExpr.independentValue=leftFactor*left.independentValue
   	+rightFactor*right.independentValue;
      if (abs(tmpExpr.independentValue)
 	  <zero*(abs(leftFactor*left.independentValue)
 	  	 +abs(rightFactor*right.independentValue)))
   	tmpExpr.independentValue=0.0;
      leftPos=0;
      rightPos=0;
      while (leftPos<left.size&&rightPos<right.size) {
   	if (left.indices[leftPos]<right.indices[rightPos]) {
 	  tmpExpr.addTerm(left.indices[leftPos],leftFactor*left.factors[leftPos]);
 	  leftPos++;
   	}
   	else if (left.indices[leftPos]>right.indices[rightPos]) {
 	  tmpExpr.addTerm(right.indices[rightPos],rightFactor*right.factors[rightPos]);
 	  rightPos++;
   	}
   	else {
 	  tmpExpr.addTerm(left.indices[leftPos],leftFactor*left.factors[leftPos]+
 	  		  rightFactor*right.factors[rightPos]);
 	  leftPos++;
 	  rightPos++;
   	}
      }
      while (leftPos<left.size) {
   	tmpExpr.addTerm(left.indices[leftPos],leftFactor*left.factors[leftPos]);
   	leftPos++;
      }
      while (rightPos<right.size) {
   	tmpExpr.addTerm(right.indices[rightPos],rightFactor*right.factors[rightPos]);
   	rightPos++;
      }
      
      tmpExpr.simplify();
      operator=(tmpExpr);
    }
    else if (left.layer<right.layer)
      operator=(right);
    else
      operator=(left);

  }
  catchAll("void expression::merge(const expression &left,const expression &right)","")
}



void expression::substitute(const expression &left,const expression &right) {
  try {    
    // One substitutes left expression in right expression.
    static expression tmpExpr;
    static Int i,leftIndex,posLeftIndexInRightIndices;
    static Double multFact;
    
    tmpExpr.setCapacity(left.size+right.size);
    if (left.index==right.index) {
      tmpExpr.index=-1;
      tmpExpr.independentValue=right.independentValue-left.independentValue;
      if (abs(tmpExpr.independentValue)
	  <zero*(abs(left.independentValue)+abs(right.independentValue)))
	tmpExpr.independentValue=0.0;
      for (i=0;i<left.size;i++)
	tmpExpr.addTerm(left.indices[i],-left.factors[i]);
      for (i=0;i<right.size;i++)
	tmpExpr.addTerm(right.indices[i],right.factors[i]);
      tmpExpr.simplify();
      operator=(tmpExpr);
      return;
    }
    
    leftIndex=left.index;
    posLeftIndexInRightIndices=-1;
    for (i=0;i<right.size;i++) {
      if (right.indices[i]==leftIndex) {
	posLeftIndexInRightIndices=i;
	multFact=right.factors[i];
	break;
      }
    }
    
    if (posLeftIndexInRightIndices==-1) {
      // No substitution is done :
      operator=(right);
      return;
    }
    
    // One then builds the merged list :
    
    if (tmpExpr.capacity<left.size+right.size-1)
      tmpExpr.setCapacity(left.size+right.size-1);
    tmpExpr.independentValue=(right.independentValue
			      +multFact*left.independentValue);
    if (abs(tmpExpr.independentValue)<
	zero*(abs(right.independentValue)+multFact*abs(left.independentValue)))
      tmpExpr.independentValue=0.0;
    tmpExpr.size=0;
    tmpExpr.weight=right.weight;
    tmpExpr.index=right.index;
    
    for (i=0;i<right.size;i++)
      if (right.indices[i]!=left.index)
	tmpExpr.addTerm(right.indices[i],right.factors[i]);
    
    for (i=0;i<left.size;i++)
      tmpExpr.addTerm(left.indices[i],multFact*left.factors[i]);
    
    //One simplifies the merged expression :
    
    tmpExpr.simplify();
    
    operator=(tmpExpr);
    setLayer(right.layer);
    
    return; 
  }
  catchAll("void expression::substitute(const expression &left,const expression &right)","")    
}



void expression::simplify(void) {
  try {    
    Int pos,i;
    Double tmpDouble,somme,factor,moyenne;
    Int tmpLayer;
    
    if (size==0)
      return;
      
    tmpLayer=layer;
    
    // One stores the terms in a list :
    
    static list<IntDouble> IDList;
    static IntDouble ID;
    static list<IntDouble>::iterator IDIt;
    
    IDList.clear();
    for (pos=0;pos<size;pos++)
      IDList.push_back(IntDouble(indices[pos],factors[pos]));
    
    if (size>1) {    
      IDList.sort();    
      // One merges duplicates elements (if any) :    
      static list<IntDouble> tmpIDList;
      static Int previousIndex,currentIndex;
      
      tmpIDList.clear();
      IDIt=IDList.begin();
      ID=*IDIt;
      previousIndex=ID.first;
      IDIt++;
      for (pos=1;pos<size;pos++,IDIt++) {
	currentIndex=IDIt->first;
	if (currentIndex==previousIndex) {
	  tmpDouble=ID.second+IDIt->second;
	  if (abs(tmpDouble)<zero*(abs(ID.second)+abs(IDIt->second)))
	    ID.second=0.0;
	  else
	    ID.second=tmpDouble;
	}
	else {
	  tmpIDList.push_back(ID);
	  ID=*IDIt;
	  previousIndex=currentIndex;
	}
      }
      tmpIDList.push_back(ID); // pushes the last pair.    
      IDList=tmpIDList;
    } // if (size>1)
    
    // This part checks the left hand side for terms with indices=index :
    
    static bool indexFound;
    
    indexFound=false;
    for (IDIt=IDList.begin();IDIt!=IDList.end();IDIt++) {
      if (IDIt->first==index) {
	indexFound=true;
	ID=*IDIt;
	break;
      }
    }
    
    if (!indexFound) {
      setCapacity(IDList.size());
      for (IDIt=IDList.begin();IDIt!=IDList.end();IDIt++)
	addTerm(IDIt->first,IDIt->second);
      goto finalSimplification ;
    }
    else {
      factor=1.0-ID.second;
      if (abs(factor)<zero*(1.0+abs(ID.second))) {
	// Problem in the equation :
	somme=abs(independentValue);
	for (IDIt=IDList.begin();IDIt!=IDList.end();IDIt++)
	  if (IDIt->first!=index)
	    somme+=abs(IDIt->second);
	if (somme<10.0*zero) {
	  // undeterminate equation (all terms vanish) :
	  setCapacity(0);
	  weight=0.0;
	  independentValue=0.0;
	  index=-1;
	  goto finalSimplification ;
	}
	else {
	  // we think it is an incompatible equation :
	  write(cerr);
//	  goto finalSimplification ;
	  errThrow("expression::simplify","incompatible equation.");
	  goto finalSimplification ;
	}	
      }
      // The bad term is displaced in left-hand-side of equation :
      setCapacity(IDList.size()-1);
      weight*=factor;
      independentValue/=factor;
      for (IDIt=IDList.begin();IDIt!=IDList.end();IDIt++)
	if (IDIt->first!=index)
	  addTerm(IDIt->first,IDIt->second/factor);
      // Returns :
      goto finalSimplification ;
    } // if (!indexFound)
    
    // Terms with too small a factor are cancelled :
    finalSimplification :
      
    bool isToBeSimplified=false;
    somme=abs(independentValue);
    for (i=0;i<size;i++)
      somme+=abs(factors[i]);
    moyenne=somme/(size+1);
    for (i=0;i<size;i++) {
      if (abs(factors[i])<zero*moyenne) {
	isToBeSimplified=true;
	break;
      }
    }
    if (isToBeSimplified) {
      static expression tmpExpr;
      tmpExpr.clear();
      tmpExpr.weight=weight;
      tmpExpr.index=index;
      tmpExpr.independentValue=independentValue;
      if (tmpExpr.capacity<capacity)
	tmpExpr.setCapacity(capacity);
      for (i=0;i<size;i++)
	if (abs(factors[i])>=zero*moyenne)
	  tmpExpr.addTerm(indices[i],factors[i]);
      operator=(tmpExpr);
    }
    
    setLayer(tmpLayer);
    
    return; 
  }
  catchAll("void expression::simplify(void)","")    
}



void expression::write(ostream &os) const {  
  try {
    os << "weight = " << weight ;
    
    if (layer!=0)
      os << ", layer = " << layer << " : " ;
    else
      os << " : " ;
      
    if (index!=-1)
      os << "val(" << index << ") = " << independentValue ;
    else
      os << "0 = " << independentValue ;
    for (Int i=0;i<size;i++) {
      if (factors[i]>=0.0)
	os << "+" ;
      os << factors[i] << "*val(" << indices[i] << ")" ;
    }
    os << endl ;
  }
  catchAll("void expression::write(ostream &os) const","")
}



