
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <stdio.h>

#include "filter_base.h"
#include "array/array_base.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "param/param.h"
#include "param/paramList.h"
#include "print/printTree.h"
#include "step/step.h"

using namespace std ;
using namespace keyFE2 ;



filter::filter(void) : name(), parents() {
  lineType=true;
  A=NULL;
  dom=NULL;
}



filter::filter(paramList PL) : name() {
  try {
    array *tmpArray=NULL;
    domain *tmpDomain=NULL;
    string tmpName;
    bool tmpLType=true;
    
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plDomain=PL.getListOfType(DOM);
    PL=PL.getListNotOfType(DOM);
    paramList plArray=PL.getListOfType(ARRAY);
    PL=PL.getListNotOfType(ARRAY);
    paramList plLineType=PL.getListOfType(LINETYPE);
    PL=PL.getListNotOfType(LINETYPE);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString;
      tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++)
	tmpString=tmpString+" "+PL[i].getName();
      tmpString=tmpString+".";
      errWrite("filter construction",tmpString);
    }
    
    // Checks valid parameters :
    
    if (plName.getNbrParams()>1) {
      errWrite("filter construction","only one NAME argument is allowed.");
    }
    else if (plName.getNbrParams()==1) {
      tmpName=plName[0].getStringData();
      if (tmpName=="")
	errWrite("filter construction",
		 "NAME= has to be followed by a string argument.");
    }
    
    if (plArray.getNbrParams()!=1) {
      errWrite("filter construction",
	       "exactly one ARRAY parameter is needed.");
    }
    else {
      parents *tmpParents=plArray[0].getParentData();
      tmpArray=dynamic_cast<array*>(tmpParents);
      if (tmpArray==NULL)
	errWrite("filter construction",
		 "Array= must be followed by an Array.");
    }
    
    if (plDomain.getNbrParams()!=1)  {
      errWrite("filter construction",
	       "exactly one DOM parameter is needed.");
    }
    else {
      parents *tmpParents=plDomain[0].getParentData();
      tmpDomain=dynamic_cast<domain*>(tmpParents);
      if (tmpDomain==NULL)
	errWrite("filter construction",
		 "DOM= must be followed by a Domain.");
    }
    
    if (plLineType.getNbrParams()>1) {
      errWrite("filter construction","only one LINETYPE argument is allowed.");
    }
    else if (plLineType.getNbrParams()==1)
      tmpLType=plName[0].getBoolData();
    
    if (!err::getOk())
      errThrow("filter construction","bailing out due to preceding errors.");
    
    A=NULL;
    dom=NULL;
    
    setName(tmpName);
    setDom(tmpDomain);
    setArray(tmpArray);
    setLineType(tmpLType);
  }
  catchAll("filter::filter(paramList PL)","")
}



filter::filter(const filter &in) : name(in), parents() {
  try {
    lineType=in.lineType;
    A=NULL;
    setArray(in.A);
    dom=NULL;
    setDom(in.dom);
  }
  catchAll("filter::filter(const filter &in)","")
}



filter::~filter(void) {
  try {
    blacks.clear();
    lineType=true;
    setArray(NULL);
    setDom(NULL);
  }
  catchAll("filter::~filter(void)","")
}



void filter::operator=(const filter &in) {
  if (this==&in)
    return;
  
  try {
    blacks.clear();
    lineType=in.lineType;
    setArray(in.A);
    setDom(in.dom);
  }
  catchAll("void filter::operator=(const filter &in)","")
}



void filter::setLineType(bool b) {
  lineType=b;
}



bool filter::getLineType(void) const {
  return(lineType);
}



void filter::setArray(array *a) {
  try {
    delinkAndDelete(A,&A);
    A=a;
    if (A!=NULL)
      A->addParent(&A);
  }
  catchAll("void filter::setArray(array *a)","")
}  



array* filter::getArray(void) const {
  return(A);
}



void filter::setDom(domain *d) {
  try {
    delinkAndDelete(dom,&dom);
    dom=d;
    if (dom!=NULL)
      dom->addParent(&dom);  
  }
  catchAll("void filter::setDom(domain *d)","")
}



domain* filter::getDom(void) const {
  return dom;
}



void filter::generateBlacks(void) {
  try {
    step currentStep;
    Int i,j,k,*map,mapSize,mapIndex,nbrValues;
    bool *mapBool;
    scal *CV,*CVs;
    
    blacks.clear();
    
    currentStep.setEquation(A);
    for (i=0;i<dom->getObjectsNbr();i++) {
      currentStep.setElement(dom->getObject(i));
      currentStep.setDirection(dom->getObjectDirect(i));
      currentStep.exec();
      CVs=A->getComponentValues();
      nbrValues=A->getNbrVal();
      for (k=0;k<nbrValues;k++) {
	CV=CVs+k;
	map=CV->getMap();
	if (lineType) {
	  mapSize=CV->getLigSize();
	  mapIndex=CV->getLigIndex();
	  mapBool=CV->getLigBool();
	}
	else {
	  mapSize=CV->getColSize();
	  mapIndex=CV->getColIndex();
	  mapBool=CV->getColBool();
	}
	for (j=0;j<mapSize;j++)
	  if (mapBool[j])
	    blacks.insert(map[mapIndex+j]);
      } // for k
    } // for i
  }
  catchAll("void filter::generateBlacks(void)","")
}



void filter::cleanBlacks(void) {
  try {
    blacks.clear();
  }
  catchAll("void filter::cleanBlacks(void)","")
}



printTree filter::getPrintTreeData(void) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt;
  
  try {
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    if (lineType)
      sbuf="   type = line";
    else
      sbuf="   type = column";
    pt+=sbuf;
    
    return pt;
  }
  catchAll("filter::getPrintTreeData","")  
}



printTree filter::getPrintTreeValues(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    Int counter;
    set<Int>::iterator sit;
    
    sbuf="   nodal values to be shadowed:";
    pt+=sbuf;
    
    sbuf="   ";
    for (sit=blacks.begin(),counter=0;sit!=blacks.end();sit++,counter++) {
      if (counter!=0&&counter%10==0) {
	pt+=sbuf;
	sbuf="   ";
      }
      sprintf(cbuf,"%6d",*sit);
      sbuf+=cbuf;
    }
    if (counter>0) 
      pt+=sbuf;
    else
      pt="   no nodal values to be shadowed!";
    
    return(pt);
  }
  catchAll("filter::getPrintTreeValues","")
} 



printTree filter::getPrintTree(bool values) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt,ptArg;
  
  try {
    sprintf(cbuf,"filter(A,dom) %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();
    
    ptArg=A->getFirstPrintTree("A",false);
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    ptArg=dom->getFirstPrintTree("dom",false);
    ptArg.prepareForAddLast();
    pt+=ptArg;
    
    return(pt);
  }
  catchAll("filter::getPrintTree","")
}



