
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <map>
#include <set>

#include "mesh_base.h"
#include "meshTypes.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



inline void swap(Int &a,Int &b) {
  Int c=a;
  a=b;
  b=c;
}



inline void swap(Double &a,Double &b) {
  Double c=a;
  a=b;
  b=c;
}



void mesh::read_rsMesh(istream &inFile) {
  try {
    Int k;
    
    type=rsMesh;
    
    // Reading the mesh in input file :
    
    Int nbrNodes,nbrElements,nbrAretes;
    inFile >> setw(6) >> nbrNodes >> setw(6) >> nbrElements
      >> setw(6) >> nbrAretes ;
    if (!inFile)
      errThrow("mesh::read_rsMesh","   error in reading first line.");
    nodesNbr=nbrNodes;
    if (nbrNodes!=0)
      coord=new Double[nbrNodes][3];
    else
      coord=NULL;
    for (Int i=0;i<nbrNodes;i++) {
      Int node;
      inFile >> setw(6) >> node >> setw(16) >> coord[i][0]
	>> setw(16) >> coord[i][1] >> setw(16) >> coord[i][2] ;
      if (!inFile||i!=node)
	errThrow("mesh::read_rsMesh",
		 string("   error in reading node ")+strFromInt(i)+".");
    }      
    Int** elements=NULL;
    Int* elementsType=NULL;
    Int* elementsDomain=NULL;
    {
       elementsType=new Int[nbrElements];
       elementsDomain=new Int[nbrElements];
       Int* tmpInts=new Int[4*nbrElements];
       elements=new Int* [nbrElements];
       for (Int i=0;i<nbrElements;i++)
          elements[i]=tmpInts+4*i;
    }
    for (Int i=0;i<nbrElements;i++) {
      Int element;
      inFile >> setw(6) >> element >> setw(6) >> elementsType[i] ;
      if (!inFile||i!=element)
	errThrow("mesh::read_rsMesh",
		 string("   error in reading element ")+strFromInt(i)+".");
      for (Int j=0;j<4;j++) {
	inFile >> setw(6) >> elements[i][j] ;
	if (!inFile)
	  errThrow("mesh::read_rsMesh",
		   string("   error in reading element ")+strFromInt(i)+".");
      }
      inFile >> setw(6) >> elementsDomain[i] ;
      if (!inFile)
	errThrow("mesh::read_rsMesh",
		 string("   error in reading element ")+strFromInt(i)+".");
    }
    Int** aretes=NULL;
    Int* aretesBoundary=NULL;
    {
       Int* tmpInts=new Int[2*nbrAretes];
       aretes=new Int* [nbrAretes];
       for (Int i=0;i<nbrAretes;i++)
          aretes[i]=tmpInts+2*i;
       aretesBoundary=new Int[nbrAretes];
    }
    for (Int i=0;i<nbrAretes;i++) {
      inFile >> setw(6) >> aretes[i][0] >> setw(6) >> aretes[i][1]
	>> setw(6) >> aretesBoundary[i] ;
      if (!inFile)
	errThrow("mesh::read_rsMesh",
		 string("   error in reading boundary edge ")+strFromInt(i)+".");
    }
    
    // A list of edges is created :
    set<Int>* edges=NULL;
    edges=new set<Int> [nbrNodes];
    for (Int i=0;i<nbrElements;i++) {
      for (Int j=0;j<elementsType[i];j++) {
	edges[elements[i][j]].insert(elements[i][(j+1)%elementsType[i]]);
	edges[elements[i][(j+1)%elementsType[i]]].insert(elements[i][j]);
      }
    }
    
    Int nbrEdges=0;
    for (Int i=0;i<nbrNodes;i++) {
      set<Int>::iterator j;
      for(j=edges[i].begin();j!=edges[i].end();j++)
	if (*j>i)
	  nbrEdges++;
    }
    
    // The objects of the domain are created :
    
    objectsNbr=nbrNodes+nbrEdges+nbrElements;
    if (objectsNbr!=0)
      objectsLst=new object[objectsNbr];
    else
      objectsLst=NULL;
    
    object* nodes=objectsLst;
    for (Int i=0;i<nbrNodes;i++)
      nodes[i].setType("node");
    
    object* segment=objectsLst+nbrNodes;
    for (Int i=0;i<nbrEdges;i++) {
      segment[i].setType("segment");
      segment[i].setMinusNbr(2);
    }
    
    object* surfElt=objectsLst+nbrNodes+nbrEdges;
    for (Int i=0;i<nbrElements;i++) {
      if (elementsType[i]==3) 
	surfElt[i].setType("triangle");
      else if (elementsType[i]==4) 
	surfElt[i].setType("quadrangle");
      surfElt[i].setMinusNbr(elementsType[i]);
    }
    
    // The segment are down connected to the nodes :
    
    map<Int,object*>* edgeIndexPointer=NULL;
    edgeIndexPointer=new map<Int,object*>[nbrNodes];
    Int edgeIndex=0;
    for (Int i=0;i<nbrNodes;i++) {
      set<Int>::iterator j;
      for (j=edges[i].begin();j!=edges[i].end();j++) {
	if (*j>i) {
	  segment[edgeIndex].setMinus(0,nodes+i,0);
	  segment[edgeIndex].setMinus(1,nodes+(*j),0);
	  edgeIndexPointer[i][*j]=segment+edgeIndex;
	  edgeIndexPointer[*j][i]=segment+edgeIndex;
	  ++edgeIndex;
	}
      }
    }
    
    // The nodes are up connected to the segments :
    
    {
      Int* nodeIndex=NULL;
      nodeIndex=new Int[nbrNodes];
      for (Int i=0;i<nbrNodes;i++) {
	nodeIndex[i]=0;
	nodes[i].setPlusNbr(edges[i].size());
      }
      Int edgeIndex=0;
      for (Int i=0;i<nbrNodes;i++) {
	set<Int>::iterator j;
	for(j=edges[i].begin();j!=edges[i].end();j++) {
	  if (*j>i) {
	    nodes[i].setPlus(nodeIndex[i],segment+edgeIndex);
	    nodeIndex[i]++;
	    nodes[*j].setPlus(nodeIndex[*j],segment+edgeIndex);
	    nodeIndex[*j]++;
	    ++edgeIndex;
	  }
	}
      }
      delete[] nodeIndex;
    }
    
    // The triangles and quadrangles are created and down connected to the 
    // segments (and the segments are up connected to triangles and quadrangles) :
    
    {
      Int node1,node2;
      Int* nbrUp=NULL;
      nbrUp=new Int[nbrEdges];
      for (Int i=0;i<nbrEdges;i++)
	nbrUp[i]=0;
      for (Int i=0;i<nbrElements;i++) {
	surfElt[i].setMinusNbr(elementsType[i]);
	for (Int j=0;j<elementsType[i];j++) {
	  node1=elements[i][j];
	  node2=elements[i][(j+1)%elementsType[i]];
	  if (node2>node1)
	    surfElt[i].setMinus(j,edgeIndexPointer[node1][node2],2);
	  else
	    surfElt[i].setMinus(j,edgeIndexPointer[node2][node1],-2);
	  nbrUp[(edgeIndexPointer[node1][node2]-segment)]++;
	}
      }
      for (Int i=0;i<nbrEdges;i++)
	segment[i].setPlusNbr(nbrUp[i]);
      for (Int i=0;i<nbrEdges;i++)
	nbrUp[i]=0;
      for (Int i=0;i<nbrElements;i++) {
	for (Int j=0;j<elementsType[i];j++) {
	  Int edgeIndex=surfElt[i].getMinus(j)-segment;
	  surfElt[i].getMinus(j)->setPlus(nbrUp[edgeIndex],surfElt+i);
	  nbrUp[edgeIndex]++;
	}
      }
      delete[] nbrUp;
    }
    
    // Definition of subdomains and boundaries :      
    
    {
      set<Int> subDomains,boundaries;
      set<Int>::iterator j;
      for (Int i=0;i<nbrElements;i++)
	subDomains.insert(elementsDomain[i]);
      Int nbrSubDomains=subDomains.size();
      for (Int i=0;i<nbrAretes;i++)
	boundaries.insert(aretesBoundary[i]);
      Int nbrBoundaries=boundaries.size();
      
      setDomainsNbr(nbrSubDomains+nbrBoundaries);
      
      Int domainIndex=0;
      for (j=subDomains.begin();j!=subDomains.end();j++) {
	Int nbrES=0;
	for (Int i=0;i<nbrElements;i++)
	  if (elementsDomain[i]==(*j))
	    nbrES++;
	domainsLst[domainIndex].setObjectsNbr(nbrES);
	Int indexES=0;
	for (Int i=0;i<nbrElements;i++) {
	  if (elementsDomain[i]==(*j)) {
	    // One assumes that the orientation definition is conserved
	    // in the domain description.
	    domainsLst[domainIndex].setObject(indexES,surfElt+i,
					      surfElt[i].getMinusNbr());
	    indexES++;
	  }
	}
	domainsLst[domainIndex].setDim();
	domainsIntId[domainIndex]=(*j);
	domainIndex++;
      }
      
      Int boundaryIndex=0;
      Int indexAB;
      for (j=boundaries.begin();j!=boundaries.end();j++) {
	Int nbrAB=0;
	domain tmp;
	for (Int i=0;i<nbrAretes;i++)
	  if (aretesBoundary[i]==(*j))
	    nbrAB++;
	domainsLst[nbrSubDomains+boundaryIndex].setObjectsNbr(nbrAB);
	object *theEdge,*theMaille;
	Int direct;
	indexAB=0;
	for (Int i=0;i<nbrAretes;i++) {
	  if (aretesBoundary[i]==(*j)) {
	    theEdge=edgeIndexPointer[aretes[i][0]][aretes[i][1]];
	    if (theEdge->plusNbr==1) {
	      // The edge belongs to a boundary.
	      direct=0;
	      theMaille=theEdge->plusLst[0];
	      for (k=0;k<theMaille->minusNbr;k++)
		if (theMaille->minusLst[k]==theEdge)
		  direct=theMaille->minusDirect[k];
	      if (direct==0)
		errThrow("mesh::read_rsMesh",
			 string("   invalid edge for boundary ")
			 +strFromInt(aretes[i][0])
			 +" "+strFromInt(aretes[i][1])+".");
	    }
	    else
	      direct=2;
	    domainsLst[nbrSubDomains+boundaryIndex]
	      .setObject(indexAB,theEdge,direct);
	    indexAB++;
	  }
	}
	domainsLst[nbrSubDomains+boundaryIndex].setDim();
	domainsIntId[nbrSubDomains+boundaryIndex]=(*j);
	tmp.reorder1DDomain(domainsLst[nbrSubDomains+boundaryIndex]);
	domainsLst[nbrSubDomains+boundaryIndex]=tmp;
	boundaryIndex++;
      }
    }
    if (nbrElements!=0) {
      delete[] elements[0];
      delete[] elements;
      delete[] elementsType;
      delete[] elementsDomain;
    }
    if (nbrAretes!=0) {
      delete[] aretes;
      delete[] aretesBoundary;
    }
    delete[] edges;
    delete[] edgeIndexPointer;
  }
  catchAll("void mesh::read_rsMesh(istream &inFile)","")
}



void mesh::write_rsMesh(ostream &outFile) const {
  try {
    map<object*,Int> elemsDoms;
    domain *currentDom;
    Int objsNbr,domainId,segmentsNbr=0,elementsNbr=0;
    object *obj;
    for (Int i=0;i<domainsNbr;i++) {
      currentDom=domainsLst+i;
      domainId=domainsIntId[i];
      objsNbr=currentDom->getObjectsNbr();
      for (Int j=0;j<domainsLst[i].getObjectsNbr();j++) {
	obj=currentDom->getObject(j);
	if (obj->getType()==3||obj->getType()==4) {
	  elementsNbr++;
	  elemsDoms[obj]=domainId;
	}
	else if (obj->getType()==2)
	  segmentsNbr++;
      }
    }
    
    // Writing of preliminary data :
    
    outFile << setw(6) << nodesNbr << setw(6) << elementsNbr 
      << setw(6) << segmentsNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_rsMesh","   error in writing first line.");
    
    // Writing of nodes :
    
    outFile.setf(ios::scientific);
    outFile << setprecision(7);
    for (Int i=0;i<nodesNbr;i++) {
      outFile << setw(6) << i ;
      if (!outFile)
	errThrow("mesh::write_rsMesh",
		 string("   error in writing node ")+strFromInt(i+1)+".");
      for (Int j=0;j<3;j++) {
	outFile << setw(16) << coord[i][j] ;
	if (!outFile)
	  errThrow("mesh::write_rsMesh",
		   string("   error in writing node ")+strFromInt(i+1)+".");
      }
      outFile << endl ;
      if (!outFile)
	errThrow("mesh::write_rsMesh",
		 string("   error in writing node ")+strFromInt(i+1)+".");
    }
    
    // Writing of other elements (triangles, quadrangles) :
    
    Int elemsCounter=0;
    object* objList[50];
    Int objDirect[50],objNbr;
    for (Int i=0;i<objectsNbr;i++) {
      obj=objectsLst+i;
      if (obj->getType()==3||obj->getType()==4) {
	outFile << setw(6) << elemsCounter << setw(6) << obj->getType() ;
	if (!outFile)
	  errThrow("mesh::write_rsMesh",
		   string("   error in writing element ")+strFromInt(i+1)+".");
	objNbr=obj->split(obj->getType(),objList,objDirect);
	for (Int j=0;j<objNbr;j++) {
	  if (objList[j]->getType()==1) {
	    outFile << setw(6) << objList[j]-objectsLst ;
	    if (!outFile)
	      errThrow("mesh::write_rsMesh",
		       string("   error in writing element ")
		       +strFromInt(i+1)+".");
	  }
	}
	if (obj->getType()==3) {
	  outFile << setw(6) << -1 ;
	  if (!outFile)
	    errThrow("mesh::write_rsMesh",
		     string("   error in writing element ")
		     +strFromInt(i+1)+".");
	}
	if (elemsDoms.find(obj)!=elemsDoms.end()) {
	  outFile << setw(6) << elemsDoms[obj] ;
	  if (!outFile)
	    errThrow("mesh::write_rsMesh",
		     string("   error in writing element ")
		     +strFromInt(i+1)+".");
	}
	else {
	  outFile << setw(6) << -1 ;
	  if (!outFile)
	    errThrow("mesh::write_rsMesh",
		     string("   error in writing element ")
		     +strFromInt(i+1)+".");
	}
	outFile << endl ;
	if (!outFile)
	  errThrow("mesh::write_rsMesh",
		   string("   error in writing element ")+strFromInt(i+1)+".");
	elemsCounter++;
      }
    }
    
    for (Int i=0;i<domainsNbr;i++) {
      currentDom=domainsLst+i;
      domainId=domainsIntId[i];
      objsNbr=currentDom->getObjectsNbr();
      for (Int j=0;j<domainsLst[i].getObjectsNbr();j++) {
	obj=currentDom->getObject(j);
	if (obj->getType()==2) {
	  objNbr=obj->split(currentDom->getObjectDirect(j),objList,objDirect);
	  for (Int k=0;k<objNbr;k++) {
	    if (objList[k]->getType()==1) {
	      outFile << setw(6) << objList[k]-objectsLst ;
	      if (!outFile)
		errThrow("mesh::write_rsMesh",
			 string("   error in writing domain ")
			 +strFromInt(i+1)+".");
	    }
	  }
	  outFile << setw(6) << domainId << endl ;
	  if (!outFile)
	    errThrow("mesh::write_rsMesh",
		     string("   error in writing domain ")+strFromInt(i+1)+".");
	}
      }
    }
  }
  catchAll("void mesh::write_rsMesh(ostream &outFile) const","")
}



