
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iostream>
#include <iomanip>
#include <map>
#include <stdio.h>
#include <string>

#include "plot_base.h"
#include "array/array_base.h"
#include "array/arrayComp.h"
#include "err/err.h"
#include "domain/domain_base.h"
#include "field/field_base.h"
#include "mesh/mesh_base.h"
#include "object/object_base.h"
#include "object/objectTypes.h"
#include "print/print.h"
#include "print/printTree.h"
#include "result/result_base.h"

using namespace std ;
using namespace keyFE2 ;



plot::plot(void) {
  try {
    dom=NULL;
    coordinates=NULL;
    
    resultsCapacity=10;
    resultsSize=0;
    results=new result* [resultsCapacity];
  }
  catchAll("plot::plot(void)","")
}




plot::plot(domain *d,array *c) {
  try {
    dom=NULL;
    setDomain(d);
    coordinates=NULL;
    setCoordinates(c);
    
    resultsCapacity=10;
    resultsSize=0;
    results=new result* [resultsCapacity];
  }
  catchAll("plot::plot(domain *d,array *c","")
}



plot::~plot(void) {
  try {
    setDomain(NULL);
    setCoordinates(NULL);
    
    for (Int i=0;i<resultsSize;i++) {
      delinkAndDelete(results[i],results+i);
      results[i]=NULL;
    }
    delete[] results;
  }
  catchAll("plot::~plot(void)","")
}



void plot::setType(string s) {
  try {
    if (s!="vigie2D"&&s!="openDX"&&s!="gmsh") {
      errThrow("void plot::setType(string s)",
               "invalid TYPE: "+s+".");
    }
    outputType=s;
  }
  catchAll("void plot::setType(string s)","")
}



void plot::setDomain(domain *d) {
  try {
    delinkAndDelete(dom,&dom);
    dom=d;
    if (dom!=NULL)
      dom->addParent(&dom);
  }
  catchAll("void plot::setDomain(domain *d)","")
}



void plot::setCoordinates(array *c) {
  try {
    delinkAndDelete(coordinates,&coordinates);
    coordinates=c;
    if (coordinates!=NULL)
      coordinates->addParent(&coordinates);
  }
  catchAll("void plot::setCoordinates(array *c)","")
}



void plot::addResult(array *A) {
  try {
    if (A==NULL)
      errThrow("plot::addResult(array *A)","cannot add NULL result.");
    result *tmpR=new result(A,dom,"linear");
    tmpR->setName(A->getName());
    addResult(tmpR);
  }
  catchAll("void plot::addResult(array *A)","")
}



void plot::addResult(result *r) {
  try {
    if (r==NULL)
      errThrow("plot::addResult(result *r)","cannot add NULL result.");
    if (resultsSize==resultsCapacity) {
      result **tmpResults=new result* [2*resultsCapacity+1];
      for (Int i=0;i<resultsCapacity;i++) {
	tmpResults[i]=results[i];
	tmpResults[i]->addParent(tmpResults+i);
	delinkAndDelete(results[i],results+i);
	results[i]=NULL;
      }
      delete[] results;
      results=tmpResults;
      resultsCapacity=2*resultsCapacity+1;
    } 
    results[resultsSize]=r;
    results[resultsSize]->addParent(results+resultsSize);
    resultsSize++; 
  }
  catchAll("void plot::addResult(result *r)","")
}



void plot::exec(string baseFileName) {
  try {
    if (outputType=="vigie2D")
      plot::exec_vigie2D(baseFileName);
    else if (outputType=="openDX")
      plot::exec_openDX(baseFileName);
    else if (outputType=="gmsh")
      plot::exec_gmsh(baseFileName);
    else
      errThrow("plot::exec","invalid output type :  "+outputType+".");
  }
  catchAll("plot::exec(string outputType,string baseFileName)","")
}



void plot::exec_vigie2D(string baseFileName) {
  try {
    Int i,j,k,index,direct,nbr,nbrComponents;
    string resultFileName,descFileName;
    object *obj;
    Int nbrNodes,nbrTriangles,nbrQuadrangles;
    Int allObjectsNbr,objectsNbr;
    Int meshIndex;
    object *oList[50];
    Int dList[50];
    Int indices[MAX_ORDER_ARRAY];
    
    for (i=0;i<resultsSize;i++) {
      if (results[i]->dom!=dom)
        errThrow("plot::exec_vigie2D","invalid Domain for result.");
      if (results[i]->interpolationTypeName!="linear")
        errThrow("plot::exec_vigie2D",
	         "invalid result interpolation for vigie: "
		 +results[i]->interpolationTypeName+".");
    }
    
    resultFileName=baseFileName+".ascii2d";
    ofstream resFile(resultFileName.c_str());
    
    mesh *BM=dom->getBaseMesh();
    
    Int maxObj=BM->getObjectsNbr();
    Int* iNodes=NULL;
    iNodes=new Int[maxObj];
    for (i=0;i<maxObj;i++)
      iNodes[i]=-1;
    
    nbrNodes=0;
    nbrTriangles=0;
    nbrQuadrangles=0;
    allObjectsNbr=dom->getAllObjectsNbr();
    for (i=0;i<allObjectsNbr;i++) {
      obj=dom->getAllObject(i);
      if (obj->getType()==1) {
	nbrNodes++;
	meshIndex=BM->getObjectIndex(obj);
	iNodes[meshIndex]=i;
      }
      else if (obj->getType()==3)
	nbrTriangles++;
      else if (obj->getType()==4)
	nbrQuadrangles++;
    }
    resFile << "points " << nbrNodes << endl ;
    
    field fCoord(dom,2,"linear");
    fCoord.interpolate(coordinates,dom,coordinates);
    fCoord.printNodalValues(resFile);
    
    objectsNbr=dom->getObjectsNbr();
    
    if (nbrTriangles) {
      resFile << "triangles " << nbrTriangles << endl ;
      for (i=0;i<objectsNbr;i++) {
	obj=dom->getObject(i);
	direct=dom->getObjectDirect(i);
	if (obj->getType()==3) {
	  nbr=obj->split(direct,oList,dList);
	  for (j=0;j<nbr;j++) {
	    if (oList[j]->getType()==1) {
	      index=BM->getObjectIndex(oList[j]);
	      resFile << setw(6) << iNodes[index] ;
	    }
	  }
	  resFile << endl ;
	}
      }
    }
    
    if (nbrQuadrangles) {
      resFile << "quadrangles " << nbrQuadrangles << endl ;
      for (i=0;i<objectsNbr;i++) {
	obj=dom->getObject(i);
	direct=dom->getObjectDirect(i);
	if (obj->getType()==4) {
	  nbr=obj->split(direct,oList,dList);
	  for (j=0;j<nbr;j++) {
	    if (oList[j]->getType()==1) {
	      index=BM->getObjectIndex(oList[j]);
	      resFile << setw(6) << iNodes[index] ;
	    }
	  }
	  resFile << endl ;
	}
      }
    }
    
    descFileName=baseFileName+".desc";
    ofstream descFile(descFileName.c_str());
    descFile << "ascii2d" << endl ;
    descFile << "./" << resultFileName << endl ;
    for (j=0;j<resultsSize;j++) {
      string resName=results[j]->getName();
      char *str=const_cast<char *>(resName.c_str());
      Int strlength=strlen(str);
      char blanc=' ';
      for (i=0;i<strlength;i++)
	if (str[i]==blanc)
	  str[i]='_';
	  
      results[j]->generateField(coordinates);
      array *resArray=results[j]->A;
      field *resField=results[j]->F;
      
      if (resArray->getOrder()==1&&resArray->getDim(0)==2) {
	resFile << "vectors " << resName << " " 
	  << resName << ".1" << " " << resName << ".2" 
	    << " .1" << endl ;
	for (i=0;i<resField->nodalValuesNbrPC;i++) {
	  for (k=0;k<resField->nbrComponents;k++)
	    resFile << " " 
	      << resField->nodalValues[k*resField->nodalValuesNbrPC+i] ;
	  resFile << endl ;
	}
	descFile << resName << endl ;
      }
      else if (resArray->getOrder()==1&&resArray->getDim(0)>2) {	
	{	  
	  // The two first components are plotted as a vector :	
	  resFile << "vectors " << resName << " " 
	    << resName << ".1" << " " << resName << ".2" 
	      << " .1" << endl ;
	  for (i=0;i<resField->nodalValuesNbrPC;i++) {
	    for (k=0;k<2;k++)
	      resFile << " " 
	        << resField->nodalValues[k*resField->nodalValuesNbrPC+i] ;
	    resFile << endl ;
	  }
	  descFile << resName << endl ;
	}	
	{	  
	  // Other components :	
	  nbrComponents=resArray->getNbrVal();
	  for (i=2;i<nbrComponents;i++) {
	    descFile << resName << i+1 << endl ;
	    resFile << "scalars " << resName << i+1 << endl ;
	    for (k=0;k<resField->nodalValuesNbrPC;k++)
	      resFile << " " 
	        << resField->nodalValues[i*resField->nodalValuesNbrPC+k]
	        << endl ;
	  }	
	}
      }
      else {
	nbrComponents=resArray->getNbrVal();
	for (i=0;i<nbrComponents;i++) {
	  descFile << resName ;
	  resFile << "scalars " << resName ;
	  resArray->getIndicesFromIndex(i,indices);
	  for (k=0;k<resArray->getOrder();k++) {
	    descFile << indices[k]+1 ;
	    resFile << indices[k]+1 ;
	  }
	  descFile << endl ;
	  resFile << endl ;
	  for (k=0;k<resField->nodalValuesNbrPC;k++)
	    resFile << " " 
	      << resField->nodalValues[i*resField->nodalValuesNbrPC+k]
	      << endl ;
	}
      }
      results[j]->cleanField();
    }
    delete[] iNodes;
  }
  catchAll("void plot::exec_vigie2D(string baseFileName)","")
}



void plot::exec_openDX(string baseFileName) {
  try {
    map<domain*,int> mapDom;
    Int i,iVal,domCounter;
    
    domCounter=0;
    for (i=0;i<resultsSize;i++) {
      iVal=mapDom[results[i]->dom];
      if (iVal==0)
        mapDom[results[i]->dom]=(++domCounter);
      if (results[i]->interpolationTypeName!="linear"&&
          results[i]->interpolationTypeName!="constantPerElement")
        errThrow("plot::exec_openDX",
	         "invalid result interpolation for openDX: "
		 +results[i]->interpolationTypeName+".");
    }
    
    string resultFileName=baseFileName+".dx";
    ofstream resFile(resultFileName.c_str());
    
    map<domain*,Int>::iterator mapDomIt;
    for (mapDomIt=mapDom.begin();mapDomIt!=mapDom.end();mapDomIt++) {
      object *obj;
      Int nbrNodes,nbrTriangles,nbrQuadrangles;
      Int allObjectsNbr,objectsNbr;
      Int meshIndex;
      object *oList[50];
      Int dList[50];
      Int direct,j,nbr;
    
      domain *dom=mapDomIt->first;
      iVal=mapDomIt->second;
      mesh *BM=dom->getBaseMesh();
      
      Int maxObj=BM->getObjectsNbr();
      Int* iNodes=NULL;
      iNodes=new Int[maxObj];
      for (i=0;i<maxObj;i++)
  	iNodes[i]=-1;
      
      nbrNodes=0;
      nbrTriangles=0;
      nbrQuadrangles=0;
      allObjectsNbr=dom->getAllObjectsNbr();
      for (i=0;i<allObjectsNbr;i++) {
  	obj=dom->getAllObject(i);
  	if (obj->getType()==1) {
  	  nbrNodes++;
  	  meshIndex=BM->getObjectIndex(obj);
  	  iNodes[meshIndex]=i;
  	}
  	else if (obj->getType()==3)
  	  nbrTriangles++;
  	else if (obj->getType()==4)
  	  nbrQuadrangles++;
      }
      
      resFile << "object " << 2*iVal-1 
              << " class array type float rank 1 shape 3 items "
	      << nbrNodes << " data follows" << endl ;
      field fCoord(dom,2,"linear");
      fCoord.interpolate(coordinates,dom,coordinates);
      for (Int k=0;k<fCoord.nodalValuesNbrPC;k++)
        resFile 
          << " " << fCoord.nodalValues[0*fCoord.nodalValuesNbrPC+k]
          << " " << fCoord.nodalValues[1*fCoord.nodalValuesNbrPC+k]
          << " " << 0.0
          << endl ;
      resFile << endl << endl ;
      
      objectsNbr=dom->getObjectsNbr();
      
      Int maxNbr=-1;
      if (nbrTriangles)
        maxNbr=3;
      if (nbrQuadrangles)
        maxNbr=4;	
      resFile << "object " << 2*iVal
              << " class array type int rank 1 shape " << maxNbr << " items "
	      << objectsNbr << " data follows" << endl ;
      
      Int nodeIndices[4],nodeCounter;
      for (i=0;i<objectsNbr;i++) {
        obj=dom->getObject(i);
        direct=dom->getObjectDirect(i);
        nbr=obj->split(direct,oList,dList);
	nodeCounter=0;
        for (j=0;j<nbr;j++) {
          if (oList[j]->getType()==1)
            nodeIndices[nodeCounter++]=iNodes[BM->getObjectIndex(oList[j])] ;
        }
	if (obj->getType()==3) {
	  if (maxNbr==3)
	    resFile << " " << nodeIndices[0] 
	            << " " << nodeIndices[1] 
		    << " " << nodeIndices[2] 
		    << endl ;
          else if (maxNbr==4)
	    resFile << " " << nodeIndices[0] 
	            << " " << nodeIndices[1] 
		    << " " << nodeIndices[2] 
		    << " " << nodeIndices[2] 
		    << endl ;
	  else 
	    errThrow("void plot::exec_openDX(string baseFileName)",
	             "error in plotting triangular result.")
        }
	else if (obj->getType()==4)
	 resFile << " " << nodeIndices[0] 
	 	 << " " << nodeIndices[1] 
	     	 << " " << nodeIndices[3] 
	     	 << " " << nodeIndices[2] 
	     	 << endl ;
	else 
	  errThrow("void plot::exec_openDX(string baseFileName)",
		   "unknown object type.")
      }

      resFile << "attribute \"element type\" string \"";
      if (maxNbr==4)
        resFile << "quads\"" << endl ;
      else if (maxNbr==3)
        resFile << "triangles\"" << endl ;
      else
        errThrow("void plot::exec_openDX(string baseFileName)",
	         "invalid element type.")
      resFile << "attribute \"ref\" string \"positions\"" << endl ;
      resFile << endl << endl ;
      
      delete[] iNodes;
    } // for (mapDomIt=mapDom.begin();mapDomIt!=mapDom.end();mapDomIt++)
    
    for (Int j=0;j<resultsSize;j++) {
      results[j]->generateField(coordinates);
      array *resArray=results[j]->A;
      field *resField=results[j]->F;
      
      if (resArray->getOrder()==0) {
        resFile << "object " << 2*domCounter+2*j+1
	        << " class array type float rank 0 items "
		<< resField->nodalValuesNbrPC << " data follows" << endl ;
	for (i=0;i<resField->nodalValuesNbrPC;i++)
	  resFile << " " << resField->nodalValues[i] << endl ;
      }
      else if (resArray->getOrder()==1) {
        resFile << "object " << 2*domCounter+2*j+1
	        << " class array type float rank 1 shape 3 items "
		<< resField->nodalValuesNbrPC << " data follows" << endl ;
        Int nbrComponents=resArray->getNbrVal();
	for (i=0;i<resField->nodalValuesNbrPC;i++)
	  if (nbrComponents==2)
	    resFile << " "
	            << resField->nodalValues[0*resField->nodalValuesNbrPC+i]
	            << " "
	            << resField->nodalValues[1*resField->nodalValuesNbrPC+i]
		    << " 0"  << endl ;
	  else if (nbrComponents==3)
	    resFile << " "
	            << resField->nodalValues[0*resField->nodalValuesNbrPC+i]
	            << " "
	            << resField->nodalValues[1*resField->nodalValuesNbrPC+i]
	            << " "
	            << resField->nodalValues[2*resField->nodalValuesNbrPC+i]
		    << endl ;
	  else
            errThrow("void plot::exec_openDX(string baseFileName)",
	             "invalid number of components for vector "
		     +results[j]->getName()+".");
      }
      else if (resArray->getOrder()==2) {
        resFile << "object " << 2*domCounter+2*j+1
	        << " class array type float rank 2 shape 3 3 items "
		<< resField->nodalValuesNbrPC << " data follows" << endl ;
        Int nbrComponents=resArray->getNbrVal();
	for (i=0;i<resField->nodalValuesNbrPC;i++)
	  if (nbrComponents==4)
	    resFile << " "
	            << resField->nodalValues[0*resField->nodalValuesNbrPC+i]
	            << " "
	            << resField->nodalValues[1*resField->nodalValuesNbrPC+i]
		    << " 0 "
	            << resField->nodalValues[2*resField->nodalValuesNbrPC+i]
	            << " "
	            << resField->nodalValues[3*resField->nodalValuesNbrPC+i]
		    << " 0 0 0 0"  << endl ;
	  else if (nbrComponents==9) {
	    for (Int k=0;k<9;k++)
	      resFile << " "
	              << resField->nodalValues[k*resField->nodalValuesNbrPC+i] ;
	      resFile << endl ;
          }
	  else
            errThrow("void plot::exec_openDX(string baseFileName)",
	             "invalid number of components for tensor "
		     +results[j]->getName()+".");
      }
      else
        errThrow("void plot::exec_openDX(string baseFileName)",
	         "maximum order of output Array is 2.");

      resFile << "attribute \"dep\" string " ;
      if (results[j]->interpolationTypeName=="linear")
        resFile << "\"positions\"" << endl ;
      else if (results[j]->interpolationTypeName=="constantPerElement")
        resFile << "\"connections\"" << endl ;
      else
        errThrow("void plot::exec_openDX(string baseFileName)",
	         "invalid interpolation: "+results[j]->interpolationTypeName
		 +".");
      resFile << endl << endl ;
      
      resFile << "object " 
              << " \"" << results[j]->getName() << "\"" 
	      << " class field" << endl ;
      Int domId=mapDom[results[j]->dom];
      resFile << "component \"positions\" value " << 2*domId-1 << endl ;
      resFile << "component \"connections\" value " << 2*domId << endl ;
      resFile << "component \"data\" value " << 2*domCounter+2*j+1 << endl ;
      resFile << endl << endl ;
      
      results[j]->cleanField();
    }
    resFile << "end" << endl ;
  }
  catchAll("void plot::exec_openDX(string baseFileName)","")
}



void plot::exec_gmsh(string baseFileName) {
  try {
  
    Int i,j,k,l,m,n;
  
    for (i=0;i<resultsSize;i++) {
      if (results[i]->interpolationTypeName!="linear"&&
          results[i]->interpolationTypeName!="linearPerElement"&&
          results[i]->interpolationTypeName!="constantPerElement")
        errThrow("plot::exec_gmsh",
	         "invalid result interpolation for gmsh: "
		 +results[i]->interpolationTypeName+".");
    }
    
    string resultFileName=baseFileName+".gmsh";
    ofstream resFile(resultFileName.c_str());
    
    resFile << "$PostFormat" << endl ;
    resFile << 1.2 << " " << 0 << " " << sizeof(double) << endl ; // 0 = ASCII
    resFile << "$EndPostFormat" << endl ;
    
    Int dimDom,objType,objDirect,nbrCmp,nbrValPerCmp,nbrMaxCmp;
    double *values;
    domain *currentDom;
    object *obj;
    Int parentBidon;
    field *tmpF=NULL,*tmpC=NULL;
    Int nbrNodes,nbrEdges,nbrTriangles,nbrQuadrangles;
    string tmpString;

    for (n=0;n<resultsSize;n++) {
      resFile << "$View" << endl ;
      tmpString=results[n]->getName();
      while (tmpString.find(" ")!=string::npos)
        tmpString.replace(tmpString.find(" "),1,"_");
      resFile << tmpString << " " << 1 << endl ;
      results[n]->generateField(coordinates);
      array *resArray=results[n]->A;
      field *resField=results[n]->F;

      currentDom=results[n]->dom;
      dimDom=currentDom->getDim();
      if (dimDom<0||dimDom>2)
        errThrow("plot::exec_gmsh",
		 "invalid domain dimension for gmsh:"
	         +strFromInt(dimDom)+".");
      
      nbrNodes=nbrEdges=nbrTriangles=nbrQuadrangles=0;
      for (j=0;j<currentDom->getObjectsNbr();j++) {
        obj=currentDom->getObject(j);
	objType=obj->getType();
	switch (objType) {
	  case 1:
	    nbrNodes++;
	    break;
	  case 2:
	    nbrEdges++;
	    break;
	  case 3:
	    nbrTriangles++;
	    break;
	  case 4:
	    nbrQuadrangles++;
	    break;
	  default:
            errThrow("plot::exec_gmsh",
		     "invalid object type for gmsh:"
	             +strFromInt(obj->getType())+".");
	} // switch (obj->getType())
      } // for (Int j=0;j<currentDom->getObjectsNbr();j++)
      
      switch(dimDom) {
        case 0:
	  if (nbrEdges||nbrTriangles||nbrQuadrangles)
            errThrow("plot::exec_gmsh",
		     "inconsistant domain and objects dimensions.");
	  break;
	case 1:
	  if (nbrNodes||nbrTriangles||nbrQuadrangles)
            errThrow("plot::exec_gmsh",
		     "inconsistant domain and objects dimensions.");
	  break;
	case 2:
	  if (nbrNodes||nbrEdges)
            errThrow("plot::exec_gmsh",
		     "inconsistant domain and objects dimensions.");
	  break;
      }
      
      Int nbrResults[3];
      nbrResults[0]=nbrResults[1]=nbrResults[2]=0;
      nbrResults[resArray->getOrder()]=1;
      
      resFile << nbrNodes*nbrResults[0] << " "
              << nbrNodes*nbrResults[1] << " "
              << nbrNodes*nbrResults[2] << " " ;
      resFile << nbrEdges*nbrResults[0] << " "
              << nbrEdges*nbrResults[1] << " "
              << nbrEdges*nbrResults[2] << " " ;
      resFile << nbrTriangles*nbrResults[0] << " "
              << nbrTriangles*nbrResults[1] << " "
              << nbrTriangles*nbrResults[2] << " " ;
      resFile << nbrQuadrangles*nbrResults[0] << " "
              << nbrQuadrangles*nbrResults[1] << " "
              << nbrQuadrangles*nbrResults[2] << " " ;
      resFile << 0 << " "
              << 0 << " "
              << 0 << " " ; // tetrahedra
      resFile << 0 << " "
              << 0 << " "
              << 0 << " " ; // hexahedra
      resFile << 0 << " "
              << 0 << " "
              << 0 << " " ; // prisms
      resFile << 0 << " "
              << 0 << " "
              << 0 << " " ; // pyramids
      resFile << 0 << " "
              << 0 << " "
              << 0 << " "
              << 0 << endl ; // texts
	      
      resFile << 0.0 << endl ; // time steps values

      tmpC=new field(currentDom,coordinates->getNbrVal(),"linearPerElement");
      tmpC->addParent(&parentBidon);
      tmpC->interpolate(coordinates,currentDom,coordinates);
      tmpF=new field(currentDom,resField->getNbrComponents(),"linearPerElement");
      tmpF->interpolate(resField,currentDom,tmpC);
      tmpF->addParent(&parentBidon);
      
      objType=-1;
      for (i=0;i<4;i++) {
        switch (i) {
	  case 0:
	    objType=node;
	    break;
	  case 1:
	    objType=segment;
	    break;
	  case 2:
	    objType=triangle;
	    break;
	  case 3:
	    objType=quadrangle;
	    break;
	}
	if ((objType==node&&dimDom==0)
	    ||(objType==segment&&dimDom==1)
	    ||(objType==triangle&&dimDom==2)
	    ||(objType==quadrangle&&dimDom==2)) {
          for (j=0;j<currentDom->getObjectsNbr();j++) {
            obj=currentDom->getObject(j);
            objDirect=currentDom->getObjectDirect(j);
	    if (obj->getType()==objType) {
	      tmpC->fillLocalNodalValues(obj,objDirect,false,true);
	      nbrCmp=tmpC->nbrComponents;
	      nbrValPerCmp=tmpC->localNodalValuesSize/nbrCmp;
	      values=tmpC->localNodalValues;
	      for (k=0;k<nbrCmp;k++) {
	        for (l=0;l<nbrValPerCmp;l++)
	          resFile << values[k*nbrValPerCmp+l] << " " ;
		resFile << endl ;
	      }
	      for (k=nbrCmp;k<3;k++) {
	        for (l=0;l<nbrValPerCmp;l++)
	          resFile << 0.0 << " " ;
		resFile << endl ;
	      }
	      
	      tmpF->fillLocalNodalValues(obj,objDirect,false,true);
	      nbrCmp=tmpF->nbrComponents;
	      nbrValPerCmp=tmpF->localNodalValuesSize/nbrCmp;
	      values=tmpF->localNodalValues;
	      switch (resArray->getOrder()) {
	        case 0:
		  if (nbrCmp>1)
		    errThrow("plot::exec_gmsh",
			     "invalid number of components for order 0 result: "
			     +strFromInt(nbrCmp)+".");
		  nbrMaxCmp=1;
		  break;
		case 1:
		  if (nbrCmp!=2&&nbrCmp!=3)
		    errThrow("plot::exec_gmsh",
			     "invalid number of components for order 1 result: "
			     +strFromInt(nbrCmp)+".");
		  nbrMaxCmp=3;
		  break;
		case 2:
		  if (nbrCmp!=4&&nbrCmp!=9)
		    errThrow("plot::exec_gmsh",
			     "invalid number of components for order 0 result: "
			     +strFromInt(nbrCmp)+".");
		  nbrMaxCmp=9;
		  break;
		default:
		  errThrow("plot::exec_gmsh",
			   "invalid result order: "
			   +strFromInt(resArray->getOrder())+".");
	      } // switch (resArray->getOrder())
	      switch (nbrCmp) {
	        case 1:
	        case 3:
	        case 9:
		  for (l=0;l<nbrValPerCmp;l++) {
		    for (k=0;k<nbrCmp;k++)
		      resFile << values[k*nbrValPerCmp+l] << " " ;
		    resFile << endl ;
		  }
		  break;
	        case 2:
		  for (l=0;l<nbrValPerCmp;l++) {
		    for (k=0;k<nbrCmp;k++)
		      resFile << values[k*nbrValPerCmp+l] << " " ;
		    resFile << 0.0 ;
		    resFile << endl ;
		  }
		  break;
	        case 4:
		  for (l=0;l<nbrValPerCmp;l++) {
		    for (k=0;k<nbrCmp;k++) {
		      for (m=0;m<nbrCmp;m++)
		        resFile << values[(3*k+m)*nbrValPerCmp+l] << " " ;
		      resFile << 0.0 << " " ;
		    }
		    resFile << 0.0 << " " << 0.0 << " " << 0.0 << endl ;
		  }
		  break;
	      } // switch (nbrCmp)
	    } // if (obj->getType()==objType)
          } // for (Int j=0;j<currentDom->getObjectsNbr();j++)
        } // if ((objType==...)
      } // for (Int i=0;i<4;i++)

      tmpC->supParent(&parentBidon);
      delete tmpC;
      tmpC=NULL;
      tmpF->supParent(&parentBidon);
      delete tmpF;
      tmpF=NULL;
      results[n]->cleanField();
      
      resFile << "$EndView" << endl ;
    } // for (n=0;n<resultsSize;n++)
  }
  catchAll("void plot::exec_gmsh(string baseFileName)","")
}



void plot::setNewPoint(bool b) {
  if (dom!=NULL)
    dom->setNewPoint(b);
  if (coordinates!=NULL)
    coordinates->setNewPoint(b);
  for (Int i=0;i<resultsSize;i++)
    if (results[i]!=NULL) 
      results[i]->setNewPoint(b);
  newPoint::setNewPoint(b);
}



void plot::accessNbrInit(void) {
  if (dom!=NULL)
    dom->accessNbrInit();
  if (coordinates!=NULL)
    coordinates->accessNbrInit();
  for (Int i=0;i<resultsSize;i++)
    if (results[i]!=NULL) 
      results[i]->accessNbrInit();
  newPoint::accessNbrInit();
}



printTree plot::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("plot::getPrintTreeData","")
}



printTree plot::getPrintTreeValues(void) const {
  try {
    printTree pt;
    return(pt);
  }
  catchAll("plot::getPrintTreeValues","")
}



printTree plot::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int i;
    
    sprintf(cbuf,"plot %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();

    if (dom!=NULL)
      ptArg=dom->print::getPrintTreeSummary("dom");
    else
      ptArg="dom = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    if (coordinates!=NULL)
      ptArg=coordinates->print::getPrintTreeSummary("coordinates");
    else
      ptArg="coordinates = NULL";
    ptArg.prepareForAdd();
    pt+=ptArg;
    
    sprintf(cbuf," | resultsSize = %d",resultsSize);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<resultsSize;i++) {
      sprintf(cbuf,"results[%d]",i+1);
      ptArg=results[i]->getPrintTree(cbuf);
      if (i<resultsSize-1)
	ptArg.prepareForAdd();
      else
	ptArg.prepareForAddLast();
      pt+=ptArg;
    }
    
    return(pt);
  }
  catchAll("plot::getPrintTree","")
}




