
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>

#include "Problem.h"
#include "problem_base.h"
#include "array/array_base.h"
#include "array/Array.h"
#include "constr/constr_base.h"
#include "constr/Constr.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "param/param.h"
#include "param/paramList.h"

using namespace std ;
using namespace keyFE2 ;



Problem::Problem(void) {
  try {
    problem *R=new problem;
    data=NULL;
    setData(R);
  }
  catchAll("Problem::Problem(void)","")
}



Problem::Problem(const Problem &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("Problem::Problem(const Problem &in)","")
}



Problem::Problem(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Problem R(PL);
    data=NULL;
    setData(R.getData());
  }
  catchAll("Problem::Problem(param P)","")
}



Problem::Problem(paramList PL) {
  try {
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plArray=PL.getListOfType(ARRAY);
    PL=PL.getListNotOfType(ARRAY);
    paramList plConstr=PL.getListOfType(CONSTR);
    PL=PL.getListNotOfType(CONSTR);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("Problem construction",tmpString);
    }
    
    param P;
    problem *outputP=new problem;
        
    array *tmpA;
    for (Int i=0;i<plArray.getNbrParams();i++) {
      P=plArray[i];
      tmpA=dynamic_cast<array*>(P.getParentData());
      if (tmpA==NULL) {
	errWrite("Problem construction",
		 "ARRAY= must be followed by an Array argument.");
      }
      else
        outputP->addArray(tmpA);
    }
    
    constr *tmpC;
    for (Int i=0;i<plConstr.getNbrParams();i++) {
      P=plConstr[i];
      tmpC=dynamic_cast<constr*>(P.getParentData());
      if (tmpC==NULL) {
	errWrite("Problem construction",
		 "Constr= must be followed by a Constr argument.");
      }
      else
        outputP->addConstr(tmpC);
    }
    
    string sid;
    if (plName.getNbrParams()>1) {
      errWrite("Problem construction",
               "maximum one NAME argument expected.");
    }
    else if (plName.getNbrParams()==0)
      ;
    else {
      param P=plName[0];
      sid=P.getStringData();
      outputP->setName(sid);
    }
         
    if (!err::getOk())
      errThrow("Problem construction","bailing out due to preceding errors.");
    
    data=NULL;
    setData(outputP);  
  }
  catchAll("Problem::Problem(paramList PL)","")
}



Problem::~Problem(void) {
  try {
    setData(NULL);
  }
  catchAll("Problem::~Problem(void)","")
}



void Problem::setData(problem *newProblemData) {
  try {
    problem *P=getData();
    delinkAndDelete(P,&data);
    if (newProblemData!=NULL)
      newProblemData->addParent(&data);
    data=static_cast<void*>(newProblemData);
  }
  catchAll("void Problem::setData(problem *newProblemData)","")
}



problem* Problem::getData(void) const {
  return(static_cast<problem*>(data));
}



void Problem::setName(string s) {
  try {
    problem *problemData=getData();
    problemData->setName(s);
  }
  catchAll("void Problem::setName(string s)","")
}



string Problem::getName(void) const {
  try {
    problem *problemData=getData();
    return(problemData->getName());
  }
  catchAll("string Problem::getName(void) const","")
}



void Problem::addArray(Array &inArray) {
  try {
    problem *problemData=getData();
    array *arrayData=inArray.getData();
    problemData->addArray(arrayData);
  }
  catchAll("string Problem::addArray(Array &inArray)","")
}



void Problem::addConstr(Constr &inConstr) {
  try {
    problem *problemData=getData();
    constr *constrData=inConstr.getData();
    problemData->addConstr(constrData);
  }
  catchAll("string Problem::addConstr(Constr &inConstr)","")
}



void Problem::write(void) const {
  try {
    problem *problemData=getData();
    problemData->write();
  }
  catchAll("Problem::write","");
}



void Problem::writeValues(void) const {
  try {
    problem *problemData=getData();
    problemData->writeValues();
  }
  catchAll("Problem::writeValues","");
}



void Problem::write(ostream &out) const {
  try {
    problem *problemData=getData();
    problemData->write(out);
  }
  catchAll("Problem::write","");
}



void Problem::writeValues(ostream &out) const {
  try {
    problem *problemData=getData();
    problemData->writeValues(out);
  }
  catchAll("Problem::writeValues","");
}



void Problem::operator=(const Problem &Rin) {
  if (this==&Rin)
    return;
  
  try {
    setData(Rin.getData());
    return;
  }
  catchAll("void Problem::operator=(const Problem &Rin)","")
}

