
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>

#include "Result.h"
#include "result_base.h"
#include "array/array_base.h"
#include "err/err.h"
#include "domain/domain_base.h"
#include "field/field_base.h"
#include "param/param.h"
#include "param/paramList.h"

using namespace std ;
using namespace keyFE2 ;



Result::Result(void) {
  try {
    result *R=new result;
    data=NULL;
    setData(R);
  }
  catchAll("Result::Result(void)","")
}



Result::Result(const Result &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("Result::Result(const Result &in)","")
}



Result::Result(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Result R(PL);
    data=NULL;
    setData(R.getData());
  }
  catchAll("Result::Result(param P)","")
}



Result::Result(paramList PL) {
  try {
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plDomain=PL.getListOfType(DOM);
    PL=PL.getListNotOfType(DOM);
    paramList plInterp=PL.getListOfType(INTERPOLATION);
    PL=PL.getListNotOfType(INTERPOLATION);
    paramList plArray=PL.getListOfType(ARRAY);
    PL=PL.getListNotOfType(ARRAY);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("Result construction",tmpString);
    }
    
    array *tmpA=NULL;
    if (plArray.getNbrParams()!=1) {
      errWrite("Result construction",
               "exactly one ARRAY argument expected.");
    }
    else {
      param P=plArray[0];
      tmpA=dynamic_cast<array*>(P.getParentData());
      if (tmpA==NULL)
	errWrite("Result construction",
		 "ARRAY= must be followed by an Array argument.");
    }
    
    domain *tmpD=NULL;
    if (plDomain.getNbrParams()!=1) {
      errWrite("Result construction",
               "exactly one DOM argument expected.");
    }
    else {
      param P=plDomain[0];
      tmpD=dynamic_cast<domain*>(P.getParentData());
      if (tmpD==NULL)
	errWrite("Result construction",
		 "DOM= must be followed by a Domain argument.");
    }
    
    string tmpS;
    if (plInterp.getNbrParams()!=1) {
      errWrite("Result construction",
               "exactly one Interp argument expected.");
    }
    else {
      param P=plInterp[0];
      tmpS=P.getStringData();
      if (tmpS=="")
	errWrite("Result construction",
		 "INTERPOLATION= must be followed by a string argument.");
    }
    
    string sid;
    if (plName.getNbrParams()>1) {
      errWrite("Result construction",
               "maximum one NAME argument expected.");
    }
    else if (plName.getNbrParams()==0)
      ;
    else {
      param P=plName[0];
      sid=P.getStringData();
    }
         
    if (!err::getOk())
      errThrow("Result construction","bailing out due to preceding errors.");
    
    result *outputP=new result(tmpA,tmpD,tmpS);
    outputP->setName(sid);
    
    data=NULL;
    setData(outputP);  
  }
  catchAll("Result::Result(paramList PL)","")
}



Result::~Result(void) {
  try {
    setData(NULL);
  }
  catchAll("Result::~Result(void)","")
}



void Result::setData(result *newResultData) {
  try {
    result *R=getData();
    delinkAndDelete(R,&data);
    if (newResultData!=NULL)
      newResultData->addParent(&data);
    data=static_cast<void*>(newResultData);
  }
  catchAll("void Result::setData(result *newResultData)","")
}



result* Result::getData(void) const {
  return(static_cast<result*>(data));
}



void Result::setName(string s) {
  try {
    result *resultData=getData();
    resultData->setName(s);
  }
  catchAll("void Result::setName(string s)","")
}



string Result::getName(void) const {
  try {
    result *resultData=getData();
    return(resultData->getName());
  }
  catchAll("string Result::getName(void) const","")
}



void Result::write(void) const {
  try {
    result *resultData=getData();
    resultData->write();
  }
  catchAll("Result::write","");
}



void Result::writeValues(void) const {
  try {
    result *resultData=getData();
    resultData->writeValues();
  }
  catchAll("Result::writeValues","");
}



void Result::write(ostream &out) const {
  try {
    result *resultData=getData();
    resultData->write(out);
  }
  catchAll("Result::write","");
}



void Result::writeValues(ostream &out) const {
  try {
    result *resultData=getData();
    resultData->writeValues(out);
  }
  catchAll("Result::writeValues","");
}



void Result::operator=(const Result &Rin) {
  if (this==&Rin)
    return;
  
  try {
    setData(Rin.getData());
    return;
  }
  catchAll("void Result::operator=(const Result &Rin)","")
}

