
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>

#include "shape.h"
#include "interpolationTypes.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "object/objectTypes.h"

using namespace std ;
using namespace keyFE2 ;



Int shape::interpolationType(string interpolationTypeName) {
  if (interpolationTypeName=="none")
    return(none);
  else if (interpolationTypeName=="constantOnDomain")
    return(constantOnDomain);
  else if (interpolationTypeName=="constantPerElement")
    return(constantPerElement);
  else if (interpolationTypeName=="linear")
    return(linear);
  else if (interpolationTypeName=="quadratic")
    return(quadratic);
  else if (interpolationTypeName=="Serendip")
    return(Serendip);
  else if (interpolationTypeName=="linearPerElement")
    return(linearPerElement);
  else if (interpolationTypeName=="quadraticPerElement")
    return(quadraticPerElement);
  else if (interpolationTypeName=="SerendipPerElement")
    return(SerendipPerElement);
  else
    return(-1);
}



string shape::interpolationTypeName(Int interpolationType) {
  if (interpolationType==none)
    return("none");
  else if (interpolationType==constantOnDomain)
    return("constantOnDomain");
  else if (interpolationType==constantPerElement)
    return("constantPerElement");
  else if (interpolationType==linear)
    return("linear");
  else if (interpolationType==quadratic)
    return("quadratic");
  else if (interpolationType==Serendip)
    return("Serendip");
  else if (interpolationType==linearPerElement)
    return("linearPerElement");
  else if (interpolationType==quadraticPerElement)
    return("quadraticPerElement");
  else if (interpolationType==SerendipPerElement)
    return("SerendipPerElement");
  else
    errThrow("shape::interpolationTypeName",
	     string("unknown interpolation type : ")
	     +strFromInt(interpolationType)+".");
}



Int shape::getSpecificNVnbr(object *theObject,Int interpolationType,
			    domain *dom) {
  try {
    switch (interpolationType) {
    case none :
      return(0);
    case constantOnDomain :
    case constantPerElement :
      if (dom!=NULL) {
	if (theObject->getDim()==dom->getDim())
	  return(1);
	else
	  return(0);
      }
      else
	errThrow("shape::getSpecificNVnbr",string("for ")
		 +interpolationTypeName(interpolationType)
		 +" interpolation type, we need a non NULL domain argument.");
    case linear :
      switch (theObject->getType()) {
      case node :
	return(1);
      case segment :
      case triangle :
      case quadrangle :
	return(0);
      default :
	errThrow("function shape::getSpecificNVnbr",
		 string("unknown object type : ")
		 +strFromInt(theObject->getType())+".");
      }
    case quadratic :
      switch (theObject->getType()) {
      case node :
	return(1);
      case segment :
	return(1);
      case triangle :
	return(0);
      case quadrangle :
	return(1);
      default :
	errThrow("shape::getSpecificNVnbr",string("unknown object type : ")
		 +strFromInt(theObject->getType())+".");
      }
    case Serendip :
      switch (theObject->getType()) {
      case node :
	return(1);
      case segment :
	return(1);
      case triangle :
      case quadrangle :
	return(0);
      default :
	errThrow("shape::getSpecificNVnbr",string("unknown object type : ")
		 +strFromInt(theObject->getType())+".");
      }
    case linearPerElement :
      if (theObject->getDim()==dom->getDim()) {
        switch (theObject->getType()) {
        case node :
      	  return(1);
        case segment :
      	  return(2);
        case triangle :
	  return(3);
        case quadrangle :
      	  return(4);
        default :
      	  errThrow("shape::getSpecificNVnbr",string("unknown object type : ")
      	  	   +strFromInt(theObject->getType())+".");
        }
      }
      else
        return(0);
    case quadraticPerElement :
      if (theObject->getDim()==dom->getDim()) {
        switch (theObject->getType()) {
        case node :
      	  return(1);
        case segment :
      	  return(3);
        case triangle :
	  return(6);
        case quadrangle :
      	  return(9);
        default :
      	  errThrow("shape::getSpecificNVnbr",string("unknown object type : ")
      	  	   +strFromInt(theObject->getType())+".");
        }
      }
      else
        return(0);
    case SerendipPerElement :
      if (theObject->getDim()==dom->getDim()) {
        switch (theObject->getType()) {
        case node :
      	  return(1);
        case segment :
      	  return(3);
        case triangle :
	  return(6);
        case quadrangle :
      	  return(8);
        default :
      	  errThrow("shape::getSpecificNVnbr",string("unknown object type : ")
      	  	   +strFromInt(theObject->getType())+".");
        }
      }
      else
        return(0);
    default :
      errThrow("shape::getSpecificNVnbr",string("unknown interpolation type : ")
	       +strFromInt(interpolationType)+".");
    }
  }
  catchAll("shape::getSpecificNVnbr","")
}



Int shape::getSpecificNVPermutation(object *theObject,Int interpolationType,
				    Int thisDirect,Int *permutation,domain *dom) {
  try {
    
    // Cette fonction sert  construire la liste contenant l'ordre dans lequel
    // les valeurs nodales spcifiques  l'objet considr apparaissent. Par
    // exemple, permutation[0] donne le dcalage dans la liste globale pour
    // l'objet du premier objet de la liste locale.
    
    // C'est surtout utile lorsque plusieurs valeurs nodales sont retournes
    // sur un object (retour diffrent de 0 ou de 1).
    
    Int i;
  
    switch (interpolationType) {
      case none :
        return(0);
      case constantOnDomain :
        if (dom!=NULL) {
    	  if (theObject->getDim()==dom->getDim()) {
    	    permutation[0]=0;
    	    return(1);
    	  }
    	  else
    	    return(0);
        }
        else
    	  errThrow("shape::getSpecificNVPermutation",string("for ")
    	  	   +interpolationTypeName(interpolationType)
    	  	   +" interpolation type, we need a non NULL domain argument.");
      case linear:
        switch (theObject->getType()) {
          case node :
    	    permutation[0]=0;
    	    return(1);
          case segment:
          case triangle:
          case quadrangle:
    	    return(0);
          default :
    	    errThrow("shape::getSpecificNVPermutation",
    	  	     string("unknown object type : ")
    	  	     +strFromInt(theObject->getType())+".");
        }
      case quadratic:
        switch (theObject->getType()) {
          case node :
    	    permutation[0]=0;
    	    return(1);
          case segment :
    	    permutation[0]=0;
    	    return(1);
          case triangle :
    	    return(0);
          case quadrangle :
    	    permutation[0]=0;
    	    return(1);
          default :
    	    errThrow("shape::getSpecificNVPermutation",
    	  	     string("unknown object type : ")
    	  	     +strFromInt(theObject->getType())+".");
        } // switch (theObject->getType())
      case constantPerElement:
        if (dom!=NULL) {
    	  if (theObject->getDim()==dom->getDim()) {
    	    permutation[0]=0;
    	    return(1);
    	  }
    	  else
    	    return(0);
        }
        else
    	  errThrow("shape::getSpecificNVPermutation",string("for ")
    	  	   +interpolationTypeName(interpolationType)
    	  	   +" interpolation type, we need a non NULL domain argument.");
      case linearPerElement:
        if (dom!=NULL) {
    	  if (theObject->getDim()==dom->getDim()) {
	    switch (theObject->getType()) {
	      case node:
    	        permutation[0]=0;
    	        return(1);
	      case segment:
	        if (thisDirect==2) {
		  permutation[0]=0;
		  permutation[1]=1;
		}
		else {
		  permutation[0]=1;
		  permutation[1]=0;
		}
		return(2);
              case triangle:
	        for (i=0;i<3;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%3;
		  else
		    permutation[i]=(6+thisDirect-i-1)%3;
		}
		return(3);
              case quadrangle:
	        for (i=0;i<4;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%4;
		  else
		    permutation[i]=(8+thisDirect-i-1)%4;
		}
		return(4);
              default :
    	        errThrow("shape::getSpecificNVPermutation",
    	  	         string("unknown object type : ")
    	  	         +strFromInt(theObject->getType())+".");
              } // switch (theObject->getType())
    	  }
    	  else
    	    return(0);
        }
        else
    	  errThrow("shape::getSpecificNVPermutation",string("for ")
    	  	   +interpolationTypeName(interpolationType)
    	  	   +" interpolation type, we need a non NULL domain argument.");
      case quadraticPerElement:
        if (dom!=NULL) {
    	  if (theObject->getDim()==dom->getDim()) {
	    switch (theObject->getType()) {
	      case node:
    	        permutation[0]=0;
    	        return(1);
	      case segment:
	        if (thisDirect==2) {
		  permutation[0]=0;
		  permutation[1]=1;
		}
		else {
		  permutation[0]=1;
		  permutation[1]=0;
		}
		permutation[2]=2;
		return(3);
              case triangle:
	        for (i=0;i<3;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%3;
		  else
		    permutation[i]=(6+thisDirect-i-1)%3;
		  permutation[3+i]=permutation[i]+3;
		}
		return(6);
              case quadrangle:
	        for (i=0;i<4;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%4;
		  else
		    permutation[i]=(8+thisDirect-i-1)%4;
		  permutation[4+i]=permutation[i]+4;
		}
		permutation[8]=8;
		return(9);
              default :
    	        errThrow("shape::getSpecificNVPermutation",
    	  	         string("unknown object type : ")
    	  	         +strFromInt(theObject->getType())+".");
              } // switch (theObject->getType())
    	  }
    	  else
    	    return(0);
        }
        else
    	  errThrow("shape::getSpecificNVPermutation",string("for ")
    	  	   +interpolationTypeName(interpolationType)
    	  	   +" interpolation type, we need a non NULL domain argument.");
      case SerendipPerElement:
        if (dom!=NULL) {
    	  if (theObject->getDim()==dom->getDim()) {
	    switch (theObject->getType()) {
	      case node:
    	        permutation[0]=0;
    	        return(1);
	      case segment:
	        if (thisDirect==2) {
		  permutation[0]=0;
		  permutation[1]=1;
		}
		else {
		  permutation[0]=1;
		  permutation[1]=0;
		}
		permutation[2]=2;
		return(3);
              case triangle:
	        for (i=0;i<3;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%3;
		  else
		    permutation[i]=(6+thisDirect-i-1)%3;
		  permutation[3+i]=permutation[i]+3;
		}
		return(6);
              case quadrangle:
	        for (i=0;i<4;i++) {
		  if (thisDirect>0)
		    permutation[i]=(thisDirect+i)%4;
		  else
		    permutation[i]=(8+thisDirect-i-1)%4;
		  permutation[4+i]=permutation[i]+4;
		}
		return(8);
              default :
    	        errThrow("shape::getSpecificNVPermutation",
    	  	         string("unknown object type : ")
    	  	         +strFromInt(theObject->getType())+".");
              } // switch (theObject->getType())
    	  }
    	  else
    	    return(0);
        }
        else
    	  errThrow("shape::getSpecificNVPermutation",string("for ")
    	  	   +interpolationTypeName(interpolationType)
    	  	   +" interpolation type, we need a non NULL domain argument.");
      default :
        errThrow("shape::getSpecificNVPermutation",
    	  	 string("unknown interpolation type : ")
    	  	 +strFromInt(interpolationType)+".");
      }
  }
  catchAll("shape::getSpecificNVPermutation","")
}


