
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>

#include "Solver.h"
#include "solver_base.h"
#include "array/Array.h"
#include "array/array_base.h"
#include "constr/Constr.h"
#include "constr/constr_base.h"
#include "err/err.h"
#include "param/param.h"
#include "param/paramList.h"
#include "problem/problem_base.h"

using namespace std ;
using namespace keyFE2 ;



Solver::Solver(void) {
  try {
    solver *S=new solver;
    data=NULL;
    setData(S);
  }
  catchAll("Solver::Solver(void)","")
}



Solver::Solver(const Solver &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("Solver::Solver(const Solver &in)","")
}



Solver::Solver(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Solver S(PL);
    data=NULL;
    setData(S.getData());
  }
  catchAll("Solver::Solver(param P)","")
}



Solver::Solver(paramList PL) {
  try {
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plSilent=PL.getListOfType(SILENT);
    PL=PL.getListNotOfType(SILENT);
    paramList plProblem=PL.getListOfType(PROBLEM);
    PL=PL.getListNotOfType(PROBLEM);
    paramList plConstraints=PL.getListOfType(CONSTR);
    PL=PL.getListNotOfType(CONSTR);
    paramList plConvergence=PL.getListOfType(CONVERGENCE);
    PL=PL.getListNotOfType(CONVERGENCE);
    paramList plDivergence=PL.getListOfType(DIVERGENCE);
    PL=PL.getListNotOfType(DIVERGENCE);
    paramList plNitmax=PL.getListOfType(NITMAX);
    PL=PL.getListNotOfType(NITMAX);
    paramList plAutofix=PL.getListOfType(AUTOFIX);
    PL=PL.getListNotOfType(AUTOFIX);
    paramList plDelay=PL.getListOfType(DELAY);
    PL=PL.getListNotOfType(DELAY);
    paramList plPermut=PL.getListOfType(PERMUT);
    PL=PL.getListNotOfType(PERMUT);
    paramList plZero=PL.getListOfType(ZERO);
    PL=PL.getListNotOfType(ZERO);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("Solver construction",tmpString);
    }
    
    solver *outputS=new solver;
    
    // Sets the problems solved in the solver :
    
    if (plProblem.getNbrParams()==0) {
      errWrite("Solver construction",
	       "at least one PROBLEM parameter is needed.");
    }
    else {
      for (Int i=0;i<plProblem.getNbrParams();i++) {
	param PP=plProblem[i];
	array *tmpA=dynamic_cast<array*>(PP.getParentData());
	problem *tmpP=dynamic_cast<problem*>(PP.getParentData());
	if (tmpA==NULL&&tmpP==NULL) {
	  errWrite("Solver construction",
		   "PROBLEM= must be followed by an Array or a Problem argument.");
	}
	else {
	  if (tmpA!=NULL)
	    outputS->addProblem(tmpA);
	  if (tmpP!=NULL) {
	    for (Int j=0;j<tmpP->getNbrArrays();j++)
	      outputS->addProblem(tmpP->getArray(j));
	    for (Int j=0;j<tmpP->getNbrConstrs();j++)
	      outputS->addConstraint(tmpP->getConstr(j));
	  }
	}
      }
    }
    
    // Sets the constraints in the solver :
    
    for (Int i=0;i<plConstraints.getNbrParams();i++) {
      param PP=plConstraints[i];
      constr *tmpC=dynamic_cast<constr*>(PP.getParentData());
      if (tmpC==NULL) {
	errWrite("Solver construction",
		 "CONSTR= must be followed by a Constr argument.");
      }
      else
	outputS->addConstraint(tmpC);
    }
    
    // Sets various solver parameters :
    
    if (plConvergence.getNbrParams()!=1) {
      errWrite("Solver construction",
	       "exactly one CONVERGENCE parameter is needed.");
    }
    else {
      param PC=plConvergence[0];
      if (PC.getTypeName()!="Double") {
	errWrite("Solver construction",
		 "CONVERGENCE= must be followed by a double argument.");
      }
      else if (PC.getDoubleData()==0.0) {
	errWrite("Solver construction",
		 "CONVERGENCE= must be followed by a double argument >0.");
      }
      else
	outputS->converge=PC.getDoubleData();
    }
    
    if (plDivergence.getNbrParams()!=1) {
      errWrite("Solver construction",
	       "exactly one DIVERGENCE parameter is needed.");
    }
    else {
      param PD=plDivergence[0];
      if (PD.getTypeName()!="Double") {
	errWrite("Solver construction",
		 "DIVERGENCE= must be followed by a double argument.");
      }
      else if (PD.getDoubleData()==0.0) {
	errWrite("Solver construction",
		 "DIVERGENCE= must be followed by a double argument.");
      }
      else
	outputS->diverge=PD.getDoubleData();
    }
    
    if (plNitmax.getNbrParams()!=1) {
      errWrite("Solver construction",
	       "exactly one NITMAX parameter is needed.");
    }
    else {
      param PN=plNitmax[0];
      if (PN.getTypeName()!="Int") {
	errWrite("Solver construction",
		 "NITMAX= must be followed by an integer argument.");
      }
      else
	outputS->nitMax=PN.getIntData();
    }
    
    if (plSilent.getNbrParams()>1) {
      errWrite("Solver construction",
	       "maximum one SILENT parameter is needed.");
    }
    else if (plSilent.getNbrParams()==0) {
      // Nothing to do!
    }
    else {
      param PS=plSilent[0];
      if (PS.getTypeName()!="bool") {
	errWrite("Solver construction",
		 "SILENT= must be followed by a bool argument.");
      }
      else
        outputS->silent=PS.getBoolData();
    }
    
    if (plZero.getNbrParams()>1) {
      errWrite("Solver construction",
	       "maximum one ZERO parameter is needed.");
    }
    else if (plZero.getNbrParams()==0) {
      // Nothing to do!
    }
    else {
      param PZ=plZero[0];
      if (PZ.getTypeName()!="Double") {
	errWrite("Solver construction",
		 "SILENT= must be followed by a double argument.");
      }
      else
        outputS->setZero(PZ.getDoubleData());
    }
    
    if (plAutofix.getNbrParams()>1) {
      errWrite("Solver construction",
	       "maximum one AUTOFIX parameter is needed.");
    }
    else if (plAutofix.getNbrParams()==0) {
      // Nothing to do!
    }
    else {
      param PA=plAutofix[0];
      if (PA.getTypeName()!="Double") {
	errWrite("Solver construction",
		 "AUTOFIX= must be followed by a Double argument.");
      }
      else {
        outputS->autoFix=true;
        outputS->autoFixValue=PA.getDoubleData();
      }
    }
    
    if (plDelay.getNbrParams()>1) {
      errWrite("Solver construction",
	       "maximum one DELAY parameter is needed.");
    }
    else if (plDelay.getNbrParams()==0) {
      // Nothing to do!
    }
    else {
      param PD=plDelay[0];
      if (PD.getTypeName()!="Int") {
	errWrite("Solver construction",
		 "DELAY= must be followed by an integer argument.");
      }
      else
        outputS->maxNbrDelayed=PD.getIntData();
    }
    
    if (plPermut.getNbrParams()>1) {
      errWrite("Solver construction",
	       "maximum one PERMUT parameter is needed.");
    }
    else if (plPermut.getNbrParams()==0) {
      // Nothing to do!
    }
    else {
      param PP=plPermut[0];
      if (PP.getTypeName()!="bool") {
	errWrite("Solver construction",
		 "PERMUT= must be followed by a bool argument.");
      }
      else
        outputS->permut=PP.getBoolData();
    }
    
    if (!err::getOk())
      errThrow("Solver construction","bailing out due to preceding errors.");
    
    data=NULL;
    setData(outputS);  
  }
  catchAll("Solver::Solver(paramList PL)","")
}



Solver::~Solver(void) {
  try {
    setData(NULL);
  }
  catchAll("Solver::~Solver(void)","")
}



void Solver::setData(solver *newSolverData) {
  try {
    solver *S=getData();
    delinkAndDelete(S,&data);
    if (newSolverData!=NULL)
      newSolverData->addParent(&data);
    data=static_cast<void*>(newSolverData);
  }
  catchAll("void Solver::setData(solver *newSolverData)","")
}



solver* Solver::getData(void) const {
  return(static_cast<solver*>(data));
}



void Solver::setName(string s) {
  try {
    solver *solverData=getData();
    solverData->setName(s);
  }
  catchAll("void Solver::setName(string s)","")
}



string Solver::getName(void) const {
  try {
    solver *solverData=getData();
    return(solverData->getName());
  }
  catchAll("string Solver::getName(void) const","")
}



void Solver::write(void) const {
  try {
    solver *solverData=getData();
    solverData->write();
  }
  catchAll("Solver::write","");
}



void Solver::writeValues(void) const {
  try {
    solver *solverData=getData();
    solverData->writeValues();
  }
  catchAll("Solver::writeValues","");
}



void Solver::write(ostream &out) const {
  try {
    solver *solverData=getData();
    solverData->write(out);
  }
  catchAll("Solver::write","");
}



void Solver::writeValues(ostream &out) const {
  try {
    solver *solverData=getData();
    solverData->writeValues(out);
  }
  catchAll("Solver::writeValues","");
}



void Solver::operator=(const Solver &Sin) {
  if (this==&Sin)
    return;
  
  try {
//    setData(new solver(*(Sin.getData())));
    setData(Sin.getData());
  }
  catchAll("void Solver::operator=(const Solver &Sin)","")
}



void Solver::setZero(Double d) {
  try {
    solver *solverData=getData();
    solverData->setZero(d);
  }
  catchAll("void Solver::setZero(Double d)","")
}



void Solver::addProblem(Array &A) {
  try {
    solver *solverData=getData();
    solverData->addProblem(A.getData());
  }
  catchAll("void Solver::addProblem(Array &A)","")
}



void Solver::addConstraint(Constr &C) {
  try {
    solver *solverData=getData();
    solverData->addConstraint(C.getData());
  }
  catchAll("void Solver::addConstraint(Constr &C)","")
}



bool Solver::NewtonRaphson(void) {
  try {
    solver *solverData=getData();
    return(solverData->NewtonRaphson());
  }
  catchAll("bool Solver::NewtonRaphson(void)","")
}



Int Solver::getLastState(void) const {
  try {
    solver *solverData=getData();
    if (solverData!=NULL)
      return(solverData->getLastState());
    else {
      errThrow("Int Solver::getLastState(void) const",
               "unitialized Solver object");
    }
  }
  catchAll("Int Solver::getLastState(void) const","")
}



Int Solver::getLastNbrIterations(void) const {
  try {
    solver *solverData=getData();
    if (solverData!=NULL)
      return(solverData->nbrIterations);
    else {
      errThrow("Int Solver::getLastLastNbrIterations(void) const",
               "unitialized Solver object");
    }
  }
  catchAll("Int Solver::getLastLastNbrIterations(void) const","")
}



Int Solver::getLastNbrDelayed(void) const {
  try {
    solver *solverData=getData();
    if (solverData!=NULL)
      return(solverData->nbrDelayed);
    else {
      errThrow("Int Solver::getLastNbrDelayed(void) const",
               "unitialized Solver object");
    }
  }
  catchAll("Int Solver::getLastNbrDelayed(void) const","")
}



Int Solver::getLastNbrAutoFix(void) const {
  try {
    solver *solverData=getData();
    if (solverData!=NULL)
      return(solverData->nbrAutoFix);
    else {
      errThrow("Int Solver::getLastNbrAutoFix(void) const",
               "unitialized Solver object");
    }
  }
  catchAll("Int Solver::getLastNbrAutoFix(void) const","")
}



Int Solver::getLastNbrPermut(void) const {
  try {
    solver *solverData=getData();
    if (solverData!=NULL)
      return(solverData->nbrPermut);
    else {
      errThrow("Int Solver::getLastNbrPermut(void) const",
               "unitialized Solver object");
    }
  }
  catchAll("Int Solver::getLastNbrPermut(void) const","")
}



bool Solver::check(ostream &os) {
  try {
    solver* tmpS=getData();
    return(tmpS->check(os));
  }
  catchAll("bool Solver::check(ostream &os)","")
}
