#ifndef SOLVER_BASE_H
#define SOLVER_BASE_H

//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <list>
#include <iostream>
#include <set>

#include "expr/expression.h"
#include "expr/exprCollect.h"
#include "print/print.h"
#include "step/stepList.h"
#include "util/parents.h"
#include "util/utils.h"

namespace keyFE2 {

  class array ;
  class constr ;
  class arrayFilter ;
  
  class solver : public virtual name, public virtual parents,
		 public virtual print, public virtual newPoint {
					
    friend class Solver ;
  
    private :
      
      // Data for solver :
      
      Int problemsCapacity ;
      Int problemsSize ;
      array **problems ;
      
      Double zero ;
      
      Int constraintsCapacity ;
      Int constraintsSize ;
      constr **constraints ;
      
      // Data for frontal elimination :
      
      Int slSize ; // Number of steps of the solver.
      stepList SL ; // List of steps of the solver.
      std::set<arrayFilter*> SAF ; // List of all the arrayFilters of the solver.
      
      Int unknownsCapacity ; // Number of fields nodal values.
      bool *bUnknowns ; // Fields nodal values active in the solver.
      Int *lUnknowns ; // Layers for all nodal values active in the solver.
      Int nvSize ; // Number of fields nodal values.
      Double *NVBackup; // Backup of all the nodal values.
      Double *NVDelta; // Variation of all the nodal values.
      
      Int *appearance ; // Index of the step at which a nodal value appears.
      Int *disappearance ; // Index of the step at which a nodal value 
      // disappears.
      
      Int nbrActMax ; // Maximum number of active variables.
      Double **mat ; // Working matrix.
      Double *rhand ; // Working right hand vector.
      bool *active ; // This array indicates which variable of the working
      // matrix is active.
      Int *matPointers ; // Pointers indicating at which field nodal value a
      // position in the working matrix corresponds.
      Int *globalPointers ; // Pointers indicating to which field nodal value
      // a position in the working matrix corresponds.
      
      Int map2Max ; // Mappings used for assembling the local matrix and vector
      Int *ligMap2 ; // in the working matrix and vectors.
      Int *colMap2 ; //
      
      Int nbrVariables ; // Number of active variables in the solver.
      Int *varToNVPointers ; // Pointers from the active variables towards the
                             // field nodal values.
      expression *upperMat ; // Upper matrix for Gaussian elimination.
      std::list<Int> elimList ;
      
      // All what follows :  constraints in frontal elimination :  
      
      exprCollect EC ; // Constraints
      expression **NVConstrs ; // Constrained fields nodal values.
      Int tmpSize,tmpCapacity;
      bool *tmpBool;
      Double *tmpCol;
      Double **tmpMat,*tmpMatB;
      
      // Elimination strategy :
      
      Int  maxNbrDelayed ;
      bool autoFix ;
      Double autoFixValue ;
      bool permut ;
      
      // For diagnostic :
      
      Int nbrDelayed ;
      Int nbrAutoFix ;
      Int nbrPermut ;
      Int nbrIterations ;
      Int lastState ;
      
    public :
      
      Int nitMax ;
      Double converge ;
      Double diverge ;
    
      bool silent ;
    
    public :
      
      void initForConstruct(void) ;
      void clear(void) ;
    
      solver(void) ;
      solver(array *A) ;
      solver(const solver &in) ;
      ~solver(void) ;
      
      void addProblem(array *A) ;
      void setZero(Double d) ;
      void addConstraint(constr *C) ;
      
      void constraintsPrepare(void) ;
      void frontalePrepare(void) ;
      void frontaleIteration(void) ;
         void frontaleIteration_eliminateUnknown(Int index,Int &variableIndex) ;
	 void frontaleIteration_permutLines(Int index1,Int index2);
      void frontaleClear(void) ;
      
      bool NewtonRaphson(void) ;
      
      void setLastState(Int nbr) ;
      Int getLastState(void) const ;
      
      void setNewPoint(bool) ;
      void accessNbrInit(void) ;
      
      printTree getPrintTreeData(void) const ;
      printTree getPrintTreeValues(void) const ;
      printTree getPrintTree(bool values) const ;
      printTree getPrintTreeSummary(void) const ;
      
      bool check(std::ostream &os) ;
      
  } ;
  
}

#endif
