
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>

#include "step.h"
#include "array/array_base.h"
#include "array/arrayFilter.h"
#include "array/arrayInteg.h"
#include "err/err.h"
#include "field/field_base.h"
#include "object/object_base.h"
#include "scal/scal.h"

using namespace std ;
using namespace keyFE2 ;



scal* step::scalPtr=NULL;
scal step::S=0.0;



step::step(void) {
  factor=1.0;
  equation=NULL;
  element=NULL;
  direction=-1000;
}



step::step(array *eq,object *elem,Int direct) {
  factor=1.0;
  equation=eq;
  element=elem;
  direction=direct;
}



step::step(const step &in) {
  try {
    factor=in.factor;
    equation=in.equation;
    element=in.element;
    direction=in.direction;
    AFL=in.AFL;
  }
  catchAll("step::step(const step &in)","")
}



step::~step(void) {
  try {
    factor=1.0;
    equation=NULL;
    element=NULL;
    direction=-1000;
    AFL.clear();
  }
  catchAll("step::~step(void)","")
}



void step::write(ostream &myCout) const {
  try {
    myCout << "   " << equation << "   " << element
      << "   " << direction << "   AFL.size()=" << AFL.size() << endl ;
  }
  catchAll("void step::write(ostream &myCout) const","")
}



void step::setEquation(array *arg) {
  equation=arg;
}



void step::setElement(object *arg) {
  element=arg;
}



void step::setDirection(Int arg) {
  direction=arg;
}



void step::setFactor(Double x) {
  factor=x;
}



void step::multFactor(Double x) {
  factor*=x;
}



Double step::getFactor(void) const {
  return(factor);
}



array* step::getEquation(void) const {
  return(equation);
}



object* step::getElement(void) const {
  return(element);
}



Int step::getDirection(void) const {
  return(direction);
}



scal* step::getScalPtr(void) const {
  return(scalPtr);
}



void step::addArrayFilter(arrayFilter *A) {
  try {
    AFL.insert(A);
  }
  catchAll("void step::addArrayFilter(arrayFilter *A)","")
}



void step::buildActiveFieldList(void) const {
  try {
    equation->buildActiveFieldList();
  }
  catchAll("void step::buildActiveFieldList(void) const","")
}



void step::exec(Double *Coord) const {
  try {
    Double coord[3]={0.0,0.0,0.0};
    if (Coord) {
      coord[0]=Coord[0];
      coord[1]=Coord[1];
      coord[2]=Coord[2];
    }
    Int j,size;
    set<field*> *FL;
    set<field*>::iterator SFit;
    field *F;
    set<arrayFilter*>::iterator AFit;
    
    FL=equation->getFieldList();
    size=0;
    scal::mapSize=0;
    for (SFit=FL->begin();SFit!=FL->end();SFit++) {
      F=*SFit;
      F->fillLocalNodalValues(element,direction,true,true);
      
      size+=F->localNodalValuesSize;
      scal::mapSize=size;
    }
    if (size>scal::mapCapacity) {
      scal::mapCapacity=size;
      delete[] scal::map;
      if (scal::mapCapacity>0)
	scal::map=new Int[scal::mapCapacity];
      else
	scal::map=NULL;
    }
    size=0;
    scal::mapSize=0;
    for (SFit=FL->begin();SFit!=FL->end();SFit++) {
      F=*SFit;;
      for (j=0;j<F->localNodalValuesSize;j++)
	scal::map[size+j]=F->firstNodalValueIndex+F->mappingLocalGlobal[j];
      size+=F->localNodalValuesSize;
    }    
    scal::mapSize=size;
    
    arrayInteg *AI=dynamic_cast<arrayInteg*>(equation);
    if (AI!=NULL)
      AI->makeOnObject(element,direction,true,true);
    else {
      equation->setNewPoint(true);
      equation->makeAtPoint(element,direction,coord,true,true);
    }
    
    for (AFit=AFL.begin();AFit!=AFL.end();AFit++)
      (*AFit)->applyFiltersToScal(equation->getComponentValues());
    
    if (factor==1.0)
      scalPtr=equation->getComponentValues();
    else {
      S=factor*(*(equation->getComponentValues()));
      scalPtr=&S;
    }
  }
  catchAll("void step::exec(Double *Coord) const","")
}



step& step::operator=(const step &in) {
  if (this==&in)
    return(*this);
  
  try {
    factor=in.factor;
    equation=in.equation;
    element=in.element;
    direction=in.direction;
    AFL=in.AFL;
    return(*this);
  }
  catchAll("step& step::operator=(const step &in)","")
}



bool step::operator<(const step &arg2) const {
  if (element<arg2.element)
    return true;
  if (element>arg2.element)
    return false;
  if (equation<arg2.equation)
    return true;
  if (equation>arg2.equation)
    return false;
  if (direction<arg2.direction)
    return true;
  return false;
}



bool step::operator>(const step &arg2) const {
  if (element>arg2.element)
    return true;
  if (element<arg2.element)
    return false;
  if (equation>arg2.equation)
    return true;
  if (equation<arg2.equation)
    return false;
  if (direction>arg2.direction)
    return true;
  return false;
}



bool step::operator==(const step &arg2) const {
  return(element==arg2.element&&equation==arg2.equation
	 &&direction==arg2.direction);
}


