
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include "TimeDep.h"
#include "timeDep_base.h"
#include "err/err.h"
#include "field/Field.h"
#include "field/field_base.h"
#include "param/param.h"
#include "param/paramList.h"
#include "util/parents.h"
#include "solver/Solver.h"
#include "solver/solver_base.h"

using namespace std ;
using namespace keyFE2 ;



TimeDep::TimeDep(void) {
  try {
    timeDep *TD=new timeDep;
    data=NULL;
    setData(TD);
  }
  catchAll("TimeDep::TimeDep(void)","")
}



TimeDep::TimeDep(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    TimeDep TD(PL);
    data=NULL;
    setData(TD.getData());
  }
  catchAll("TimeDep::TimeDep(param P)","")
}



TimeDep::TimeDep(paramList PL) {
  try {
    string tdName="";
    string tdScheme="";
    
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plScheme=PL.getListOfType(SCHEME);
    PL=PL.getListNotOfType(SCHEME);
    paramList plMaxStepsNbr=PL.getListOfType(MAXSTEPSNBR);
    PL=PL.getListNotOfType(MAXSTEPSNBR);
    paramList plDtMin=PL.getListOfType(DTMIN);
    PL=PL.getListNotOfType(DTMIN);
    paramList plDtMax=PL.getListOfType(DTMAX);
    PL=PL.getListNotOfType(DTMAX);
    paramList plDtInit=PL.getListOfType(DTINIT);
    PL=PL.getListNotOfType(DTINIT);
    paramList plEpsilon=PL.getListOfType(EPSILON);
    PL=PL.getListNotOfType(EPSILON);
    paramList plTInit=PL.getListOfType(TINIT);
    PL=PL.getListNotOfType(TINIT);
    paramList plTFinal=PL.getListOfType(TFINAL);
    PL=PL.getListNotOfType(TFINAL);
    paramList plInflu=PL.getListOfType(INFLU);
    PL=PL.getListNotOfType(INFLU);
    paramList plActive=PL.getListOfType(ACTIVE);
    PL=PL.getListNotOfType(ACTIVE);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("TimeDep construction",tmpString);
    }
    
    // Sets the name of the timeDep :
    
    if (plName.getNbrParams()>1)  {
      errWrite("TimeDep construction","only one NAME argument is allowed.");
    }
    else if (plName.getNbrParams()==1) {
      tdName=plName[0].getStringData();
      if (tdName=="")
	errWrite("TimeDep construction",
		 "NAME= has to be followed by a string argument.");
    }
    
    // Sets the scheme of the timeDep :
    
    if (plScheme.getNbrParams()>1) {
      errWrite("TimeDep construction","only one SCHEME argument is allowed.");
    }
    else if (plScheme.getNbrParams()==1) {
      tdScheme=plScheme[0].getStringData();
      if (tdScheme=="") 
	errWrite("TimeDep construction",
		 "SCHEME= has to be followed by a string argument.");
    }
    
    // Checks the other unique parameters :
    
    Int maxStepsNbr=0;
    Double dtMin=0.0,dtMax=0.0,dtInit=0.0,tInit=0.0,tFinal=0.0,epsilon=0.0;
    
    if (plMaxStepsNbr.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one MAXSTEPSNBR argument is needed.");
    }
    else {
      maxStepsNbr=plMaxStepsNbr[0].getIntData();
      if (maxStepsNbr==0)
	errWrite("TimeDep construction",
		 "MAXSTEPSNBR= has to be followed by an integer argument.");
    }
    
    if (plDtMin.getNbrParams()!=1)  {
      errWrite("TimeDep construction",
	       "exactly one DTMIN argument is needed.");
    }
    else {
      dtMin=plDtMin[0].getDoubleData();
      if (dtMin<0.0) 
	errWrite("TimeDep construction",
		 "DTMIN= has to be followed by a positive Double argument.");
    }
    
    if (plDtMax.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one DTMAX argument is needed.");
    }
    else {
      dtMax=plDtMax[0].getDoubleData();
      if (dtMax<0.0)
	errWrite("TimeDep construction",
		 "DTMAX= has to be followed by a positive Double argument.");
    }
    
    if (plEpsilon.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one EPSILON argument is needed.");
    }
    else {
      epsilon=plEpsilon[0].getDoubleData();
      if (epsilon<=0.0)
	errWrite("TimeDep construction",
		 "EPSILON= has to be followed by a positive Double argument.");
    }
    
    if (plDtInit.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one DTINIT argument is needed.");
    }
    else {
      dtInit=plDtInit[0].getDoubleData();
      if (dtInit<0.0)
	errWrite("TimeDep construction",
		 "DTINIT= has to be followed by a positive Double argument.");
    }
    
    if (plTInit.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one TINIT argument is needed.");
    }
    else
      tInit=plTInit[0].getDoubleData();
    
    if (plTFinal.getNbrParams()!=1) {
      errWrite("TimeDep construction",
	       "exactly one TFINAL argument is needed.");
    }
    else
      tFinal=plTFinal[0].getDoubleData();
    
    if (!err::getOk())
      errThrow("TimeDep construction","bailing out due to preceding errors.");
    
    // Construction of timeDep :
    
    timeDep *currentTD=new timeDep;
    
    currentTD->setName(tdName);
    currentTD->setTimeScheme(tdScheme);
    currentTD->setMaxStepsNbr(maxStepsNbr);
    currentTD->setDtMin(dtMin);
    currentTD->setDtMax(dtMax);
    currentTD->setDtInit(dtInit);
    currentTD->setEpsilon(epsilon);
    currentTD->setTInit(tInit);
    currentTD->setTFinal(tFinal);
    
    for (Int i=0;i<plInflu.getNbrParams();i++) {
      param PP=plInflu[i];
      if (dynamic_cast<field*>(PP.getParentData())!=NULL) {
	field *tmpF=dynamic_cast<field*>(PP.getParentData());
	currentTD->addInfluField(tmpF);
      }
      else if (dynamic_cast<solver*>(PP.getParentData())!=NULL) {
	solver *tmpS=dynamic_cast<solver*>(PP.getParentData());
	currentTD->addInfluSolver(tmpS);
      }
      else 
	errWrite("TimeDep construction",
		 "INFLU= must be followed by a Field or Solver argument.");
    }
    
    for (Int i=0;i<plActive.getNbrParams();i++) {
      param PP=plActive[i];
      if (dynamic_cast<field*>(PP.getParentData())!=NULL) {
	field *tmpF=dynamic_cast<field*>(PP.getParentData());
	currentTD->addActiveField(tmpF);
      }
      else 
	errWrite("TimeDep construction",
		 "ACTIVE= must be followed by a Field argument.");
    }
    
    if (!err::getOk())
      errThrow("TimeDep construction","bailing out due to preceding errors.");
    
    data=NULL;
    setData(currentTD);
  }
  catchAll("TimeDep::TimeDep(paramList PL)","")
}



TimeDep::TimeDep(const TimeDep &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("TimeDep::TimeDep(const TimeDep &in)","")
}



TimeDep::~TimeDep(void) {
  try {
    setData(NULL);
  }
  catchAll("TimeDep::~TimeDep(void)","")
}



void TimeDep::setData(timeDep *newTD) {
  try {
    timeDep *oldTD=getData();
    if (oldTD!=NULL)
      delinkAndDelete(oldTD,&data);
    if (newTD!=NULL)
      newTD->addParent(&data);
    data=static_cast<void*>(newTD);
  }
  catchAll("void TimeDep::setData(timeDep *newTD)","")
}



timeDep* TimeDep::getData(void) const {
  return(static_cast<timeDep*>(data));
}



void TimeDep::setName(string s) {
  try {
    timeDep *TD=getData();
    TD->setName(s);
  }
  catchAll("void TimeDep::setName(string s)","")
}



string TimeDep::getName(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getName());
  }
  catchAll("string TimeDep::getName(void) const","")
}



void TimeDep::addActive(Field &F) {
  try {
    timeDep *TD=getData();
    field *f=F.getData();
    TD->addActiveField(f,0);
  }
  catchAll("void TimeDep::addActive(Field &F)","")
}



void TimeDep::addInflu(Field &F) {
  try {
    timeDep *TD=getData();
    field *f=F.getData();
    TD->addActiveField(f,1);
  }
  catchAll("void TimeDep::addInflu(Field &F)","")
}



void TimeDep::addInflu(Solver &S) {
  try {
    timeDep *TD=getData();
    solver *s=S.getData();
    TD->addInfluSolver(s);
  }
  catchAll("void TimeDep::addInflu(Solver &S)","")
}



void TimeDep::write(void) const {
  try {
    timeDep *timeDepData=getData();
    timeDepData->write();
  }
  catchAll("TimeDep::write","");
}



void TimeDep::writeValues(void) const {
  try {
    timeDep *timeDepData=getData();
    timeDepData->writeValues();
  }
  catchAll("TimeDep::writeValues","");
}



void TimeDep::write(ostream &out) const {
  try {
    timeDep *timeDepData=getData();
    timeDepData->write(out);
  }
  catchAll("TimeDep::write","");
}



void TimeDep::writeValues(ostream &out) const {
  try {
    timeDep *timeDepData=getData();
    timeDepData->writeValues(out);
  }
  catchAll("TimeDep::writeValues","");
}



TimeDep TimeDep::operator=(const TimeDep &arg) {
  if (this==&arg)
    return(*this);
  
  try {
    setData(arg.getData());
    return(*this);
  }
  catchAll("TimeDep TimeDep::operator=(const TimeDep &arg)","")
}



bool TimeDep::isToBeContinued(void) {
  try {
    timeDep *TD=getData();
    return(TD->isToBeContinued());
  }
  catchAll("bool TimeDep::isToBeContinued(void)","")
}



Double TimeDep::getThetaDT(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getThetaDT());
  }
  catchAll("Double TimeDep::getThetaDT(void) const","")
}



Int TimeDep::getCurrentStepIndex(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getCurrentStepIndex());
  }
  catchAll("Int TimeDep::getCurrentStepIndex(void) const","")
}



Double TimeDep::getCurrentTimeStep(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getCurrentTimeStep());
  }
  catchAll("Double TimeDep::getCurrentTimeStep(void) const","")
}



Double TimeDep::getCurrentTime(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getCurrentTime());
  }
  catchAll("Double TimeDep::getCurrentTime(void) const","")
}



Double TimeDep::getPreviousTime(void) const {
  try {
    timeDep *TD=getData();
    return(TD->getPreviousTime());
  }
  catchAll("Double TimeDep::getPreviousTime(void) const","")
}





